/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "EdgeConnector.h"

EdgeConnector::EdgeConnector() : m_inserted(false) {
	CONSTRUCTORTRACE(EdgeConnector);
}

EdgeConnector::~EdgeConnector() {
	DESTRUCTORTRACE(EdgeConnector);
}
// IDGMConnector
STDMETHODIMP EdgeConnector::put_LayoutObject(IUnknown *val) {
	METHTRACE(EdgeConnector,put_LayoutObject);
	if(m_edge)
		return S_FALSE;
	QISET(m_edge = val);
	return AtlAdvise(m_edge,(IDGEdgeShadow*)this,IID_IDGEdgeShadow,&m_dwEdgeCookie);
}
STDMETHODIMP EdgeConnector::get_LayoutObject(IUnknown **ret) {
	METHTRACE(EdgeConnector,get_LayoutObject);
	m_edge.CopyTo((IDGDrawnEdge**)ret);
	return S_OK;
}
STDMETHODIMP EdgeConnector::put_Site(DIMCCSite *val) {
	return SetSite(val);
}
// IDGEdgeShadow:
STDMETHODIMP EdgeConnector::Insert() {
	METHTRACE(EdgeConnector,Insert);
	if(!m_msite)
		ERR(E_PENDING);
	if(m_inserted)
		return S_FALSE;
	m_inserted = TRUE;
	// force it to position itself.
	CComPtr<IDGEdgeShadow> ear;
	RUN(m_msite->GetControl(riidret(&ear)));
	RUN(ear->Modify(DGMF_POS));
	// RUN(m_msite->PutBefore(0)); // edges at back
	RUN(m_msite->put_Intent(m_nextState));
	return S_OK;
}
STDMETHODIMP EdgeConnector::Delete() {
	METHTRACE(EdgeConnector,Delete);
	if(!m_msite)
		ERR(E_PENDING);
	if(!m_inserted)
		return S_FALSE;
	m_inserted = FALSE;
	CComPtr<IDGMEdgeControl> ctrl;
	RUN(m_msite->GetControl(riidret(&ctrl)));
	RUN(ctrl->Initialize(0));
	m_edge.Release();
	RUN(m_msite->put_Intent(MCCS_DELETED));
	return S_OK;
}
STDMETHODIMP EdgeConnector::Modify(int flags) {
    // SimpleEdge and LabelMode handle these
    return S_FALSE;
}
// IMCCSiteOwner
STDMETHODIMP EdgeConnector::Move(MCC_POS_INFO *mpi) {
	METHTRACE(EdgeConnector,Move);
	return S_OK; // presumably it will actually get moved by nodes around it being moved.
}
STDMETHODIMP EdgeConnector::PutBefore(IMCCSite *other) {
	METHTRACE(EdgeConnector,PutBefore);
	return S_FALSE; // default handling
}
STDMETHODIMP EdgeConnector::PutBehind(IMCCSite *other) {
	METHTRACE(EdgeConnector,PutBehind);
	return S_FALSE; // default handling
}
STDMETHODIMP EdgeConnector::Intend(long intent) {
	METHTRACE(EdgeConnector,Intend);
	if(!m_msite)
		ERR(E_PENDING);
	if(intent==MCCS_DELETED) {
		if(!m_inserted)
			return S_FALSE;
		return m_edge->Delete();
	}
	HRESULT hr;
	if(!m_inserted) {
		m_nextState = intent;
		CComPtr<IDGMEdgeControl> ctrl;
		RUN(m_msite->GetControl(riidret(&ctrl)));
		RUN(ctrl->Initialize(m_edge));
		//RUN(m_edge->put_Length(2.0));
		RUN(m_edge->Insert());
	}
	else
		hr = m_msite->put_Intent(intent); // allow edge not to UI activate
	return S_OK; //S_FALSE; // continue with default
}
// IMCCSiteEvents
STDMETHODIMP EdgeConnector::Moving(LPCRECT winRect) {
	METHTRACE(EdgeConnector,Moving);
	return S_FALSE;
}
STDMETHODIMP EdgeConnector::Shifting(long oldState, long newState) {
	METHTRACE(EdgeConnector,Shifting);
	// eliminate loops so system can shut down
	if((newState==MCCS_DELETED)&&m_edge) {
		RUN(AtlUnadvise(m_edge,IID_IDGEdgeShadow,m_dwEdgeCookie));
	}
	return S_OK;
}
STDMETHODIMP EdgeConnector::Layering(IMCCLayer *newLayer) {
	return S_FALSE;
}
STDMETHODIMP EdgeConnector::ZChanged() {
	return S_FALSE;
}
STDMETHODIMP EdgeConnector::Wakeup() {
	METHTRACE(EdgeConnector,Wakeup);
	return S_FALSE;
}
// IMCCExpandSelection
STDMETHODIMP EdgeConnector::Expand(IItemSet *selection) {
    METHTRACE(EdgeConnector,Expand);
	CComPtr<IMoniker> mk;
	CComPtr<IDGNode> n;
	RUN(m_edge->get_HeadNode(&n));
	CComPtr<IMCCSite> s = siteFromNode(CComQIPtr<IDGDrawnNode>(n));
	RUN(s->get_RelMoniker(&mk));
	HRESULT hr;
	RUN(hr = selection->Add(mk));
	// if news, recurse this event
	if(hr==S_OK)
		RUN(fireSiteEvent(s,fireOneParam<IMCCExpandSelection,IItemSet*>(&IMCCExpandSelection::Expand,selection)));
	mk.Release();
	n.Release();
	s.Release();
	RUN(m_edge->get_TailNode(&n));
	s = siteFromNode(CComQIPtr<IDGDrawnNode>(n));
	RUN(s->get_RelMoniker(&mk));
	RUN(hr = selection->Add(mk));
	if(hr==S_OK)
		RUN(fireSiteEvent(s,fireOneParam<IMCCExpandSelection,IItemSet*>(&IMCCExpandSelection::Expand,selection)));
	return S_OK;
}
// INeedSomePointers
STDMETHODIMP EdgeConnector::TakePointers(IEnumUnknown *os,IEnumMoniker *ms) {
	METHTRACE(EdgeConnector,TakePointers);
	if(m_edge)
		return S_FALSE;
	CComPtr<IUnknown> o;
	RUN(os->Next(1,&o,0));
	RUN(put_LayoutObject(o));
	o.Release();
	// realize end node's sites.  This is necessary so that the layout nodes have the
	// right information derived from the control, position, size, etc.  (The layout
	// edge only has monikers to the layout nodes, of course, so it has no way of forcing
	// this info to be entered
	BindCtx bctx;
	RUN(os->Next(1,&o,0));
	CComQIPtr<IGroupItem> gi;
	QISET(gi = o);
	RUN(gi->Realize(bctx)); // tail
	o.Release();
	RUN(os->Next(1,&o,0));
	QISET(gi = o);
	RUN(gi->Realize(bctx)); // head
	return S_OK;
}
// IObjectWithSite
STDMETHODIMP EdgeConnector::SetSite(IUnknown *pUnkSite) {
	METHTRACE(EdgeConnector,SetSite);
	QISET(m_msite = m_site = pUnkSite);
	DWORD milano;
	RUN(APCAdvise(m_msite,(IMCCSiteOwner*)this,IID_IMCCSiteOwner,&milano));
	RUN(APCAdvise(m_msite,(IMCCSiteEvents*)this,IID_IMCCSiteEvents,&milano));
	RUN(APCAdvise(m_msite,GetUnknown(),IID_IMCCExpandSelection,&milano));
	return S_OK;
}
STDMETHODIMP EdgeConnector::GetSite(REFIID riid, void **ret) {
	METHTRACE(EdgeConnector,GetSite);
	if(!m_site)
		ERR(E_FAIL);
	return m_site->QueryInterface(riid,ret);
}
