/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


// ShapeNode.cpp : Implementation of ShapeNode (duh)
#include "stdafx.h"
#include <math.h>
#include "ShapeNode.h"
#include <strstream>

#include "RegCategory.h"
#include <common/Geometry.h>
#include <common/useful.h>
#include "FakeBezier.h"

#include "hdcSupportsBezier.h"
#include "allocDGShapes.h"


inline POINT renderCoord(LPCRECTL other,SIZE extent, POINTD p) {
	RECT rc = *(const RECT*)other;
	bool cartesian = (rc.bottom<rc.top);
	POINT ret = {rc.left + ROUND((extent.cx/2.0+p.x*1000.0)*wid(rc)/extent.cx),
		rc.top +(cartesian?-1:1)*ROUND((extent.cy/2.0-p.y*1000.0)*hei(rc)/extent.cy)};
	return ret;
}
inline Coord internalCoord(LPCRECTL bounds,SIZE extent, POINT p) {
	Coord ret(p.x,p.y);
	ret -= Coord(bounds->left,bounds->top);
	ret.x *= extent.cx;
	ret.y *= extent.cy;
	ret.x /= wid(*bounds);
	ret.y /= hei(*bounds);
	if(ret.y<0)
		ret.y = extent.cy + ret.y;
	ret -= Coord(extent.cx,extent.cy);
	return ret;
}
ShapeNode::ShapeNode() : m_toDraw(0)
#ifdef _DEBUG
	,m_inserted(false)
#endif
{
	CONSTRUCTORTRACE(ShapeNode);
	// default size is 1cm square
	m_sizeExtent.cx = 1000;
	m_sizeExtent.cy = 1000;
	m_sizeNatural = m_sizeExtent;
	m_bDrawGetDataInHimetric = true;
}
ShapeNode::~ShapeNode() {
	DESTRUCTORTRACE(ShapeNode);
	if(m_toDraw)
		freeDGShapes(m_toDraw);
}
void ShapeNode::doDrawActions(LPCRECTL lpcrl,HDC hdc) {
    METHTRACE(ShapeNode,doDrawActions);
	if(!m_toDraw)
		return;
	if(m_toDraw->n==0)
		return;

	CHECK(SetPolyFillMode(hdc,WINDING));

	// temporary arrays for PolyPoly
	std::vector<POINT> pts; // all for PolyPolygon
	std::vector<int> counts; // polygon side counts


	bool isBezier = m_toDraw->sps[0]->isSpline;

	CComPtr<IMontageContainer> mc;
	RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
	CComPtr<IMCCTransform> tr;
	RUN(mc->get_Transform(&tr));

	POINTD p;
	RUN(m_node->get_Pos(&p));
	int j,npts=0;
	for(j=0; j<m_toDraw->n; ++j)
		npts += m_toDraw->sps[j]->n;
	pts.reserve(npts);
	for(j=0; j<m_toDraw->n; ++j) {
		DGSHAPEDESC *sh = m_toDraw->sps[j];
		int ct = sh->n;
		if(!sh->isSpline) // windows closes polygons but we close polybeziers
			ct--,npts--;
		for(int k = 0; k<ct; ++k) {
			POINTL cp = pointl(int(1000.0*(sh->pts[k].x+p.x)),int(1000.0*(sh->pts[k].y+p.y)));
			POINT vp;
			RUN(tr->C2VP(cp,&vp));
			pts.push_back(vp);
		}
		counts.push_back(ct);
	}
	_ASSERTE(npts==pts.size());
	if(isBezier) {
		bool toMetafile = !hdcSupportsBezier(hdc);
		if(toMetafile) {
			vector<int> counts2;
			vector<POINT> mega;
			int k = 0;
			for(int j = 0; j<m_toDraw->n; ++j) {
				int lastCount = mega.size();
				_ASSERTE(unsigned(k+counts[j])<=pts.size());
				CHECK(FakePolyBezier(&pts[k],counts[j],mega));
				counts2.push_back(mega.size()-lastCount);
				k += counts[j];
			}
			PolyPolygon(hdc,&*mega.begin(),&*counts2.begin(),m_toDraw->n);
		}
		else {
//#define SHOW_GUIDES
#ifndef SHOW_GUIDES
			int k = 0;
			for(int j = 0; j<m_toDraw->n; ++j) {
				_ASSERTE(unsigned(k+counts[j])<=pts.size());
				CHECK(PolyBezier(hdc,&pts[k],counts[j]));
#ifdef CATCH_BLANK
				for(int l = 1; l<counts[j]; ++l)
					_ASSERTE(pts[j+l-1]!=pts[j+l]);
#endif
				k += counts[j];
			}
			_ASSERTE(k==pts.size());
#else
			DCObject pen(hdc,CreatePen(PS_SOLID,0,RGB(200,0,0)));
			PolyPolyline(hdc,pts.begin(),(DWORD*)counts.begin(),counts.size());
#endif
		}
	}
	else {
		CHECK(PolyPolygon(hdc,&*pts.begin(),&*counts.begin(),counts.size()));
	}
}
HRESULT ShapeNode::OnDrawAdvanced(ATL_DRAWINFO& di)
{
	METHTRACE(ShapeNode,OnDraw);
	USES_CONVERSION;

	_ASSERTE(m_inserted);
	_ASSERTE(m_node);
	_ASSERTE(m_toDraw);
	_ASSERTE(m_toDraw->n);

	if(!m_node || !m_toDraw)
		ERR(E_FAIL);
	if(m_toDraw->n==0)
		return S_FALSE;

	HDC hdc = di.hdcDraw;
	bool toMetafile = !hdcSupportsBezier(hdc);
	COLORREF crFill,crBorder;
    {
        OLE_COLOR clr;
        CComQIPtr<DIDGDrawnNode> dgn = m_node;
        RUN(dgn->get_FillColor(&clr));
        RUN(OleTranslateColor(clr,0,&crFill));
        RUN(dgn->get_BorderColor(&clr));
        RUN(OleTranslateColor(clr,0,&crBorder));
    }
	DCObject brush(hdc,CreateSolidBrush(crFill));
	DCObject pen(hdc,CreatePen(m_nBorderStyle,m_nBorderWidth,crBorder));

	bool isBezier = m_toDraw->sps[0]->isSpline;
	if(isBezier && !toMetafile) {
		CHECK(BeginPath(hdc));
		doDrawActions(di.prcBounds,hdc);
		CHECK(EndPath(hdc));
		CHECK(StrokeAndFillPath(hdc));
	}
	else
		doDrawActions(di.prcBounds,hdc);

	return S_OK;
}

// IDGMNodeControl
STDMETHODIMP ShapeNode::get_Node(IDGDrawnNode **ret) {
    METHTRACE(ShapeNode,get_Node);
	m_node.CopyTo(ret);
	return S_OK;
}
STDMETHODIMP ShapeNode::put_Node(IDGDrawnNode *val) {
    METHTRACE(ShapeNode,put_Node);
	if(m_node)
		RUN(AtlUnadvise(m_node,IID_IDGNodeShadow,m_carobWheatGermBar));
	if(m_node = val) {
		RUN(AtlAdvise(m_node,GetUnknown(),IID_IDGNodeShadow,&m_carobWheatGermBar));
	}
	return S_OK;
}
// IDGNodeShadow
STDMETHODIMP ShapeNode::Insert(void) {
    METHTRACE(ShapeNode,Insert);
#ifdef _DEBUG
	_ASSERTE(!m_inserted);
	_ASSERTE(m_node);
	m_inserted = true;
#endif
    RUN(Modify(DGMF_SHAPE));
	_ASSERTE(m_toDraw&&m_toDraw->n);
	return S_OK;
}
STDMETHODIMP ShapeNode::Delete(void) {
    METHTRACE(ShapeNode,Delete);
#ifdef _DEBUG
	_ASSERTE(m_inserted);
	m_inserted = false;
#endif
    return S_FALSE; }
STDMETHODIMP ShapeNode::Modify(int flags) {
#ifdef _DEBUG
	_ASSERTE(m_inserted);
#endif
    if(flags&DGMF_SHAPE) {
	    if(m_toDraw)
		    freeDGShapes(m_toDraw);
	    RUN(m_node->get_Shape(&m_toDraw));
    }
    else if(flags&DGMF_COLORS)
        RUN(FireViewChange());
    return S_OK;
}
// IPersistStream
STDMETHODIMP ShapeNode::GetClassID(CLSID *ret) {
    METHTRACE(ShapeNode,GetClassID);
	*ret = GetObjectCLSID();
	return S_OK;
}
HRESULT ShapeNode::InitNew() {
	METHTRACE(ShapeNode,InitNew);

	m_bBorderVisible = TRUE;
	m_nBorderStyle = PS_SOLID;
	m_nBorderWidth = 0;

	return S_OK;
}
HRESULT ShapeNode::IsDirty() {
    METHTRACE(ShapeNode,IsDirty);
	return S_OK;
}
STDMETHODIMP ShapeNode::Load(IStream *stream) {
	METHTRACE(ShapeNode,Load);
	RUN(stream->Read(&m_bBorderVisible,sizeof(BOOL),0));
	RUN(stream->Read(&m_nBorderStyle,sizeof(long),0));
	RUN(stream->Read(&m_nBorderWidth,sizeof(long),0));
	return S_OK;
}
STDMETHODIMP ShapeNode::Save(IStream *stream, BOOL clean) {
	METHTRACE(ShapeNode,Save);
	RUN(stream->Write(&m_bBorderVisible,sizeof(BOOL),0));
	RUN(stream->Write(&m_nBorderStyle,sizeof(long),0));
	RUN(stream->Write(&m_nBorderWidth,sizeof(long),0));
	return S_OK;
}
STDMETHODIMP ShapeNode::GetSizeMax(ULARGE_INTEGER *ret) {
    METHTRACE(ShapeNode,GetSizeMax);
	ERR(E_NOTIMPL);
}
STDMETHODIMP ShapeNode::QueryHitPoint(DWORD dwAspect, LPCRECT pRectBounds, POINT ptlLoc, LONG, DWORD *pHitResult) {
	METHTRACE(ShapeNode,QueryHitPoint);
	if(dwAspect!=DVASPECT_CONTENT)
		ERR(E_FAIL);
	if(!m_toDraw)
		ERR(E_PENDING);
	if(m_toDraw->n==0)
		return *pHitResult = HITRESULT_OUTSIDE,S_OK;

	HDC hdc = CreateCompatibleDC(0);
	BeginPath(hdc);
	doDrawActions((LPCRECTL)pRectBounds,hdc);
	EndPath(hdc);
	HRGN hrgn = PathToRegion(hdc);
	if(PtInRegion(hrgn,ptlLoc.x,ptlLoc.y))
		*pHitResult = HITRESULT_HIT;
	else
		*pHitResult = HITRESULT_OUTSIDE;
	DeleteObject(hrgn);
	DeleteDC(hdc);
	return S_OK;
	if(PtInRect(pRectBounds, ptlLoc)) {
		// translate point to internal coords
		Coord loc2 = internalCoord((LPCRECTL)pRectBounds,m_sizeExtent,ptlLoc);
		Region r;
		r.shape.degree = m_toDraw->sps[0]->isSpline?3:1;
		r.shape.assign(reinterpret_cast<Coord*>(m_toDraw->sps[0]->pts),
			reinterpret_cast<Coord*>(m_toDraw->sps[0]->pts+m_toDraw->sps[0]->n));
		*pHitResult = r.Hit(loc2);
		/*
		// poly translated into POINTs
		std::vector<POINT> pts;
		pts.reserve(m_polydef.sides);

		// scale points to current rect
		for(int i=0;i<m_poly->n;i++)
			pts.push_back(renderCoord((RECTL*)pRectBounds, m_sizeExtent, m_poly->p[i]));

		// and check using handy Dynagraph function
		*pHitResult = pointInPoly(ctr(*pRectBounds),ptlLoc,pts.begin(),pts.size())?
			HITRESULT_HIT:HITRESULT_OUTSIDE;
		*/
	}
	else *pHitResult = HITRESULT_OUTSIDE;
	return S_OK;
}
STDMETHODIMP ShapeNode::QueryHitRect(DWORD dwAspect,LPCRECT pRectBounds,
		LPCRECT pRectLoc,LONG lCloseHint,DWORD *pHitResult) {
	METHTRACE(ShapeNode,QueryHitRect);
	if(dwAspect!=DVASPECT_CONTENT)
		ERR(E_FAIL);
	if(!m_toDraw)
		ERR(E_PENDING);
	if(m_toDraw->n==0) {
		*pHitResult = HITRESULT_OUTSIDE;
		return S_OK;
	}

	// copied from simpleedge
	HDC hdc = CreateCompatibleDC(0);
	BeginPath(hdc);
	doDrawActions((LPCRECTL)pRectBounds,hdc);
	EndPath(hdc);
	HRGN hrgn = PathToRegion(hdc);
	if(RectInRegion(hrgn,pRectLoc))
		*pHitResult = HITRESULT_HIT;
	else
		*pHitResult = HITRESULT_OUTSIDE;
	DeleteObject(hrgn);
	DeleteDC(hdc);
	return S_OK;

}
STDMETHODIMP ShapeNode::Activate(HMENU,OLEMENUGROUPWIDTHS*) {
    METHTRACE(ShapeNode,Activate);
	return S_FALSE;
}
STDMETHODIMP ShapeNode::Deactivate(HMENU,OLEMENUGROUPWIDTHS*) {
    METHTRACE(ShapeNode,Deactivate);
	return S_FALSE;
}
