/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


// implementation of the Simple Edge control

#include "stdafx.h"
#include "SimpleEdge.h"

#include <common/useful.h>
#include "FakeBezier.h"

#include "hdcSupportsBezier.h"

#include "pointd2coord.h"

const double LEN = 0.15,
	SPIKEX = 0.15,
	SPIKEY = 0.1;

SimpleEdge::SimpleEdge() : m_amSpline(false),m_amSelected(false),m_pulling(-1),m_coconut(0) {
	CONSTRUCTORTRACE(SimpleEdge);
}
SimpleEdge::~SimpleEdge() {
	DESTRUCTORTRACE(SimpleEdge);
}
// IDGMEdgeControl
STDMETHODIMP SimpleEdge::Initialize(IDGDrawnEdge *edge) {
    METHTRACE(SimpleEdge,Initialize);
	if(m_edge&&m_coconut) {
		RUN(AtlUnadvise(m_edge,IID_IDGEdgeShadow,m_coconut));
		m_coconut = 0;
		m_arrowNote.Release();
	}
	if(edge) {
		RUN(AtlAdvise(edge,(IDGEdgeShadow*)this,IID_IDGEdgeShadow,&m_coconut));
		RUN(edge->new_Label(&m_arrowNote));
		RUN(m_arrowNote->put_Where(1.0));
		RUN(m_arrowNote->put_Length(-LEN));
	}
	m_edge = edge;
	return S_OK;
}
// IDGEdgeShadow
STDMETHODIMP SimpleEdge::Insert() {
	METHTRACE(SimpleEdge,Insert);
	// EdgeConnector handles these
	return S_OK;
}
STDMETHODIMP SimpleEdge::Delete() {
	METHTRACE(SimpleEdge,Delete);
	if(m_coconut) { // avoid loops
		_ASSERTE(m_edge);
		RUN(AtlUnadvise(m_edge,IID_IDGEdgeShadow,m_coconut));
		m_coconut = 0;
	}
	return S_OK;
}
HRESULT SimpleEdge::fit() {
    METHTRACE(SimpleEdge,fit);
	RECTL rc;
	if(m_vp.empty()) {
		// probably error in engine but let's not make it worse
		// base a zero-size pos on tail's position
		CComPtr<IDGNode> tail;
		RUN(m_edge->get_TailNode(&tail));
		POINTD tailPos;
		RUN(CComQIPtr<IDGDrawnNode>(tail)->get_Pos(&tailPos));
		POINTL tailPosC = pointl(tailPos);
		rc.left = rc.right = tailPosC.x;
		rc.top = rc.bottom = tailPosC.y;
	}
	else {
		Bounds r;
		unsigned i;
		for(i = 0;i<m_vp.size();++i)
			r |= m_vp[i];
		for(i = 0; i<4; ++i)
			r |= m_arrowHead[i];
		POINTL ul = pointl(pointd(r.NW())),lr = pointl(pointd(r.SE()));
		rc.left = ul.x;
		rc.top = ul.y;
		rc.right = lr.x;
		rc.bottom = lr.y;
		// hack: rect's got to be a little bigger to let fuzzy hittesting happen,
		// and for grapples (even worse a hack!)
		rc.left -= 200;
		rc.right += 200;
		rc.bottom -= 200;
		rc.top += 200;
	}
	// ooh, control directly controlling site - antiOLE or the hidden promise?
	RUN(CComQIPtr<IMCCSite>(m_spClientSite)->put_VirtRect(&rc));
	return S_OK;
}
STDMETHODIMP SimpleEdge::Modify(int flags) {
    METHTRACE(SimpleEdge,Modify);

    if((flags&DGMF_POS)==0)
        return S_FALSE;

    CComPtr<IDGGraph> g;
	RUN(m_edge->get_Graph(&g));
	CComQIPtr<IDGLayout> eng = g;
	DG_SPLINELEVEL level;
	RUN(eng->get_SplineLevel(&level));
	if(level==DGSL_BOUNDS)
		m_transparency = 127; // 50%
	else
		m_transparency = 255; // 100%

	m_vp.clear();
	DGSHAPEDESC *pdgep;
	if(SUCCEEDED(m_edge->GetPoints(&pdgep))) {
		m_amSpline = pdgep->isSpline!=0;

		m_vp.resize(pdgep->n);
		m_vp.assign((Coord*)pdgep->pts,(Coord*)(pdgep->pts+pdgep->n));

		CoTaskMemFree(pdgep);
	}
	POINTD pos1,pos2;
	RUN(m_arrowNote->get_Pos1(&pos1));
	RUN(m_arrowNote->get_Pos2(&pos2));
	POINTD spike = {SPIKEX,SPIKEY};
	double theta = atan2(pos2.y-pos1.y,pos2.x-pos1.x);
	POINTD sharp1 = {pos2.x + spike.x*cos(theta) - spike.y*sin(theta), pos2.y + spike.y*cos(theta) + spike.x*sin(theta)};
	spike.y *= -1.0;
	POINTD sharp2 = {pos2.x + spike.x*cos(theta) - spike.y*sin(theta), pos2.y + spike.y*cos(theta) + spike.x*sin(theta)};
	m_arrowHead[0] = coord(pos1);
	m_arrowHead[1] = coord(sharp1);
	m_arrowHead[2] = coord(pos2);
	m_arrowHead[3] = coord(sharp2);

	RUN(FireViewChange());
	RUN(fit());
	return S_OK;
}
// IOleObject
STDMETHODIMP SimpleEdge::SetClientSite(IOleClientSite *site) {
	METHTRACE(SimpleEdge,SetSite);
	RUN(IOleObjectImpl<SimpleEdge>::SetClientSite(site));
	if(site) {
		DWORD nutterButter;
		RUN(APCAdvise(site,(IMCCSelectionDraw*)this,IID_IMCCSelectionDraw,&nutterButter));
	}
	else
		RUN(Initialize(0));
	return S_OK;
}
STDMETHODIMP SimpleEdge::DoVerb(LONG iVerb,LPMSG lpmsg,IOleClientSite *activeSite,LONG lindex,HWND hwndParent,LPCRECT pos) {
    METHTRACE(SimpleEdge,DoVerb);
	if(iVerb==OLEIVERB_UIACTIVATE || iVerb==OLEIVERB_PRIMARY)
		ERR(E_FAIL);
	return IOleObjectImpl<SimpleEdge>::DoVerb(iVerb,lpmsg,activeSite,lindex,hwndParent,pos);
}

// IMCCSelectionDraw
STDMETHODIMP SimpleEdge::Show() {
	METHTRACE(SimpleEdge,Show);
	m_amSelected = true;
	return FireViewChange();
}
STDMETHODIMP SimpleEdge::Hide() {
	METHTRACE(SimpleEdge,Hide);
	m_amSelected = false;
	return FireViewChange();
}
// ICoUIActiveObject
STDMETHODIMP SimpleEdge::Activate(HMENU,OLEMENUGROUPWIDTHS*) {
    METHTRACE(SimpleEdge,Activate);
	return S_FALSE;
}
STDMETHODIMP SimpleEdge::Deactivate(HMENU,OLEMENUGROUPWIDTHS*) {
    METHTRACE(SimpleEdge,Deactivate);
	return S_FALSE;
}

// IViewObject
void drawSegEnd(HDC hdc, POINT p) {
	Rectangle(hdc,p.x-2,p.y-2,p.x+3,p.y+3);
}
void drawDot(HDC hdc, POINT p) {
	Ellipse(hdc,p.x-2,p.y-2,p.x+3,p.y+3);
}
bool SimpleEdge::transformToBounds(vp_t &out) {
    METHTRACE(SimpleEdge,transformToBounds);
	CComPtr<IMontageContainer> mc;
	RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
	CComPtr<IMCCTransform> t;
	RUN(mc->get_Transform(&t));
	out.resize(m_vp.size());
	vp_t::iterator j = out.begin();
	for(vpd_t::iterator i = m_vp.begin();i!=m_vp.end();i++)
		RUN(t->C2VP(pointl(pointd(*i)),&*j++));
	VARIANT_BOOL hideDots;
	RUN(t->get_IsViewing(&hideDots));
	return hideDots!=VARIANT_FALSE; // okay stoopid thing to return but I need it
}
struct DrawSegs {
	HDC hdc;
	DrawSegs(HDC hdc) : hdc(hdc) {}
	bool operator()(POINTL pt) {
		LineTo(hdc,pt.x,pt.y);
		return false;
	}
};
STDMETHODIMP SimpleEdge::Draw(DWORD dwDrawAspect,LONG lindex,void *pvAspect,
     DVTARGETDEVICE *ptd,HDC hdcTargetDev,HDC hdc,LPCRECTL lprcBounds,
     LPCRECTL lprcWBounds,BOOL(STDMETHODCALLTYPE *pfnContinue )(DWORD dwContinue),
     DWORD dwContinue) {
	METHTRACE(SimpleEdge,Draw);
	if(!m_spClientSite)
		ERR(E_PENDING);
	if(m_vp.empty())
		return S_OK;
	vp_t cvt;
	bool hideDots = transformToBounds(cvt);
	bool toMetafile = !hdcSupportsBezier(hdc);
	COLORREF clr;
	if(m_transparency<128)  // use pseudo-color based on address as substitute for real transparency
		clr = COLORREF((unsigned(this)*1572869)&0x00ffffff);
    else {
        OLE_COLOR clr0;
        RUN(CComQIPtr<DIDGDrawnEdge>(m_edge)->get_BorderColor(&clr0));
        RUN(OleTranslateColor(clr0,0,&clr));
    }
	DCObject pen(hdc,CreatePen(PS_SOLID,(m_amSelected&&!hideDots)?3:0,clr)),
		fill(hdc,CreateSolidBrush(clr));
	if(m_amSpline) {
		if(toMetafile) {
			vector<POINT> muchos;
			FakePolyBezier(&*cvt.begin(),cvt.size(),muchos);
			Polyline(hdc,&*muchos.begin(),muchos.size());
		}
		else
			PolyBezier(hdc,&*cvt.begin(),cvt.size());
	}
	else
		Polyline(hdc,&*cvt.begin(),cvt.size());
	{
        COLORREF crFill;
        {
            OLE_COLOR clr0;
            RUN(CComQIPtr<DIDGDrawnEdge>(m_edge)->get_FillColor(&clr0));
            RUN(OleTranslateColor(clr0,0,&crFill));
        }
		DCObject fill(hdc,CreateSolidBrush(crFill));
		POINT headV[4];
		CComPtr<IMontageContainer> mc;
		RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
		CComPtr<IMCCTransform> t;
		RUN(mc->get_Transform(&t));
		for(int i = 0;i<4; ++i)
			RUN(t->C2VP(pointl(pointd(m_arrowHead[i])),&headV[i]));
		Polygon(hdc,headV,4);
	}
	if(m_amSelected && !hideDots)
		for(unsigned j = 0;j<m_vp.size();j++) {
			DCObject pen(hdc);
			if(j==m_vp.size()-1)
				pen = CreatePen(PS_SOLID,0,RGB(255,0,0));
			if(!m_amSpline||(j%3)==0)
				drawSegEnd(hdc,cvt[j]);
			/*
			else if(m_bInPlaceActive)
				drawDot(hdc,cvt[j]);
			*/
		}
	return S_OK;
}
STDMETHODIMP SimpleEdge::QueryHitPoint(DWORD aspect,LPCRECT bounds,POINT pt,
		LONG closeness,DWORD *ret) {
	METHTRACE(SimpleEdge,QueryHitPoint);
	*ret = HITRESULT_OUTSIDE;
	if(m_vp.empty())
		return S_OK;
    if(closeness<=0)
        closeness = 200;
	CComPtr<IMontageContainer> mc;
	RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
	CComPtr<IMCCTransform> t;
	RUN(mc->get_Transform(&t));
	POINTL ptl;
	RUN(t->W2CP(pt,&ptl)); // would misfire if QueryHitPoint were ever used for non-screen calculations
	Coord c = coord(pointd(ptl));
	pair<Coord,double> res = close(&*m_vp.begin(),m_vp.size(),m_amSpline?3:1,c,closeness/1000.0);
	if(dist(res.first,c)<closeness/1000.0)
		*ret = HITRESULT_HIT;
	else {
		// hit test the arrowhead; again this relies on pt in window coords
		POINT headV[4];
		for(int i = 0;i<4; ++i)
			RUN(t->C2VP(pointl(pointd(m_arrowHead[i])),&headV[i]));
		// hit test a circle that contains all arrowhead points
		POINT ctr = headV[0];
		for(int i = 1;i<4; ++i)
			ctr += headV[i];
		ctr /= 4;
		int maxr2 = dsqr(headV[0],ctr);
		for(int i = 1;i<4; ++i)
			maxr2 = max(maxr2,dsqr(headV[i],ctr));
		if(dsqr(ctr,pt)<maxr2)
			*ret = HITRESULT_HIT;
		/*
		HDC info;
		RUN(m_spInPlaceSite->GetDC(0,OLEDC_NODRAW,&info));
		BeginPath(info);
		Polygon(info,headV,4);
		EndPath(info);
		HRGN rgn = PathToRegion(info);
		if(PtInRegion(rgn,pt.x,pt.y))
			*ret = HITRESULT_HIT;
		DeleteObject(rgn);
		*/
	}
#ifdef DEBUG_HITTEST
	if(*ret==HITRESULT_HIT) {
		HDC hdc;
		CComQIPtr<IOleInPlaceSiteWindowless> wls(m_spClientSite);
		RUN(wls->GetDC(0,0,&hdc));
		SetPixel(hdc,pt.x,pt.y,RGB(255,0,0));
		RUN(wls->ReleaseDC(hdc));
	}
#endif
	return S_OK;
}
struct InRectCallback {
	Rect r;
	InRectCallback(Rect r) : r(r) {}
	bool operator()(Coord c) {
		return r.Contains(c);
	}
};
STDMETHODIMP SimpleEdge::QueryHitRect(DWORD dwAspect,LPCRECT pRectBounds,
		LPCRECT pRectLoc,LONG lCloseHint,DWORD *pHitResult) {
	METHTRACE(SimpleEdge,QueryHitRect);
	if(dwAspect!=DVASPECT_CONTENT)
		ERR(E_FAIL);
	if(m_vp.empty()) {
		*pHitResult = HITRESULT_OUTSIDE;
		return S_OK;
	}
	if(!lCloseHint)
		lCloseHint = 200;
	CComPtr<IMontageContainer> mc;
	RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
	CComPtr<IMCCTransform> t;
	RUN(mc->get_Transform(&t));
	RECTL loc;
	RUN(t->W2CR(pRectLoc,&loc));
	InRectCallback cb(rect(rectd(loc)));
	if(enumerate(&*m_vp.begin(),m_vp.size(),m_amSpline?3:1,lCloseHint/1000.0,cb))
		*pHitResult = HITRESULT_HIT;
	else
		*pHitResult = HITRESULT_OUTSIDE;
	return S_OK;
	/*
	// if edge is pretty much a straight line, the region approach
	// below won't work
	float m;
	bool vert = false, horz = false, found = false;
	for(int i=1;i<m_vp.size();++i) {
		long dx = m_vp[i].x-m_vp[i-1].x;
		if(!dx) {
			if(found&&!vert)
				break;
			vert = true;
			continue;
		}
		float m2 = float(m_vp[i].y-m_vp[i-1].y)/float(dx);
		if(found)
			if(vert || abs(m2-m)>0.1f)
				break;
		m = m2;
		found = true;
	}
	if(i==m_vp.size()) {// it's a line (or close)
		CComPtr<IMontageContainer> mc;
		RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
		CComPtr<IMCCTransform> t;
		RUN(mc->get_Transform(&t));
		// see if any point is within (this simple test only works for lines)
		for(vpl_t::iterator j = m_vp.begin();j!=m_vp.end();++j) {
			POINT pt;
			RUN(t->C2VP(*j,&pt));
			if(smartPtInRect(pRectLoc,pt)) {
				*pHitResult = HITRESULT_HIT;
				return S_OK;
			}
		}
		POINT front,back;
		RUN(t->C2VP(m_vp.front(),&front));
		RUN(t->C2VP(m_vp.back(),&back));
		// special case vert & horz to avoid /0
		if(back.x==front.x) {
			// test top and bottom sides
			if(between(front.x,pRectLoc->left,pRectLoc->right)
					&& (between(pRectLoc->top,front.y,back.y)
					||between(pRectLoc->bottom,front.y,back.y)))
				*pHitResult = HITRESULT_HIT;
			else
				*pHitResult = HITRESULT_OUTSIDE;
			return S_OK;
		}
		else if(back.y==front.y) {
			// test left and right sides
			if(between(front.y,pRectLoc->top,pRectLoc->bottom) &&
				(between(pRectLoc->left,front.x,back.x)
				||between(pRectLoc->right,front.x,back.x)))
				*pHitResult = HITRESULT_HIT;
			else
				*pHitResult = HITRESULT_OUTSIDE;
			return S_OK;
		}
		else {
			// the more complicated test: intersection with one of three sides
			long dx = back.x - front.x,
				dy = back.y - front.y;
			long leftYint = dy*(pRectLoc->left - back.x)/dx + back.y;
			if(between(leftYint,back.y,front.y)&&between(leftYint,pRectLoc->top,pRectLoc->bottom)) {
				*pHitResult = HITRESULT_HIT;
				return S_OK;
			}
			long rightYint = dy*(pRectLoc->right - back.x)/dx + back.y;
			if(between(rightYint,back.y,front.y)&&between(rightYint,pRectLoc->top,pRectLoc->bottom)) {
				*pHitResult = HITRESULT_HIT;
				return S_OK;
			}
			long topXint = dx*(pRectLoc->top - back.y)/dy + back.x;
			if(between(topXint,back.x,front.x)&&between(topXint,pRectLoc->right,pRectLoc->left)) {
				*pHitResult = HITRESULT_HIT;
				return S_OK;
			}
		}
		*pHitResult = HITRESULT_OUTSIDE;
		return S_OK;
	}
	// different, probably more expensive api-based approach (how would
	// you do this mathematically?
	HDC hdc = CreateCompatibleDC(0);
	BeginPath(hdc);
	RUN(Draw(dwAspect,-1,0,0,0,hdc,(LPCRECTL)pRectBounds,0,0,0));
	EndPath(hdc);
	HRGN hrgn = PathToRegion(hdc);
	if(RectInRegion(hrgn,pRectLoc))
		*pHitResult = HITRESULT_HIT;
	else
		*pHitResult = HITRESULT_OUTSIDE;
	DeleteObject(hrgn);
	DeleteDC(hdc);
	return S_OK;
	*/
}
// MESSAGES
LRESULT SimpleEdge::ButtonDown(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
    METHTRACE(SimpleEdge,ButtonDown);
	POINT pt = MAKEPOINT(lParam);
	vp_t cvt;
	transformToBounds(cvt);
	m_pulling = -1;
	for(unsigned i = 0; i<cvt.size(); ++i)
		if(dsqr(cvt[i],pt)<=9)
			break;
	if(i<cvt.size())
		m_pulling = i;
	else
		return 0;
	m_downpt = pt;
	RUN(m_spInPlaceSite->SetCapture(true));
	return 0;
}
LRESULT SimpleEdge::MouseMove(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
    METHTRACE(SimpleEdge,MouseMove);
	POINT pt = MAKEPOINT(lParam);
	if(m_pulling>0) {
		CComPtr<IMontageContainer> mc;
		RUN(CComQIPtr<IMCCSite>(m_spClientSite)->GetContainer(riidret(&mc)));
		CComPtr<IMCCTransform> t;
		RUN(mc->get_Transform(&t));
		POINTL ptl;
		RUN(t->W2CP(pt,&ptl));
		m_vp[m_pulling] = coord(pointd(ptl));
		RUN(m_spInPlaceSite->InvalidateRect(0,true));
	}
	return 0;
}
LRESULT SimpleEdge::ButtonUp(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
    METHTRACE(SimpleEdge,ButtonUp);
	if(m_pulling<0)
		return 0;
	RUN(m_spInPlaceSite->SetCapture(false));
	m_pulling = -1;
	return 0;
}
