/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


// GroupClass.h : Declaration of the Group

#ifndef __Group_H_
#define __Group_H_

#include "oleFuncs.h"
inline HRESULT open(IStorage *parent,LPOLESTR name,IStorage **ret) {
	if(FAILED(parent->OpenStorage(name,0,STGM_TRANSACTED|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,ret)))
		return parent->CreateStorage(name,STGM_TRANSACTED|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,ret);
	return S_OK;
}
inline HRESULT open(IStorage *parent,LPOLESTR name,IStream **ret) {
	if(FAILED(parent->OpenStream(name,0,STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,ret)))
		return parent->CreateStream(name,STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,ret);
	return S_OK;
}

// persistPasting means the object is being added - Paste is calling IGroupItem::Copy
// and doesn't yet have a pointer.  Thus the moniker is valid but the object is not - Paste
// returns S_FALSE with *retObj==0; GetObject returns E_PENDING
enum persistMethod {persistNone,persistPasting,persist,persistStream,persistStreamInit,persistStorage};

struct Item {
	CLSID clsid;
	long count;
	enum persistMethod method;
	union { // type determined by "method"; refcounted.
		IPersist *p;
		IPersistStream *pStream;
		IPersistStreamInit *pStreamInit;
		IPersistStorage *pStorage;
	};
	bool pending, // entry created but no data, e.g. CreatePassive. reset on load, bind
		element; // has its own storage element rather than being saved into the item stream
	ULONGLONG pos,size;
	CComPtr<IStorage> storage; // only for storages of course. cached because can't re-open with STGM_SHARE_EXCLUSIVE
	bool isMainStorage; // whether this is the storage that the object is holding onto too
	CComPtr<IStream> stream; // or...
	Item() :
		clsid(CLSID_NULL),
		count(0),
		method(persistNone),
		p(0),
		pending(false),
		element(false),
		pos(-1),size(-1),
		isMainStorage(false)
	{}
	~Item() {
		if(p)
			p->Release();
	}
};
typedef hash_map<BSTR,Item,oleHashTraits> itemMap;
// slow version:
// typedef map<BSTR,Item,oleOrder> itemMap;

/////////////////////////////////////////////////////////////////////////////
// Group
class ATL_NO_VTABLE Group :
	public CComObjectRoot,
	public CComCoClass<Group, &CLSID_Group>,
	public IPersistStorage,
	public IOleItemContainer,
	public IHaveIdentity,
	public IGroup,
	public IDispatchImpl<DIGroup>,
	public IGroupItem,
	public IClassActivator
{
	CComPtr<IGroup> m_parent;
	CComPtr<IUnknown> m_site;
	CComPtr<IStorage> m_storage;
	CComPtr<IStream> m_itemStream,m_indexStream;
	CComPtr<IKeepSets> m_idServer;
	DWORD m_serverID;
	itemMap m_data;
	DWORD m_anonID;

	bool m_noScribble; // after Save, can't use storage until SaveCompleted
	enum  sEvent {se_handsOff,se_saveCompleted, se_saveCompletedNew};
	HRESULT takeStorage(IStorage *pStg);
	HRESULT storageEvent(enum sEvent event);
	HRESULT fetch(IMoniker *mk, itemMap::iterator *iret);
	HRESULT findPersistType(IUnknown *o, Item&item);
	HRESULT unload(itemMap::iterator i);
	HRESULT realize(IBindCtx *bctx, itemMap::iterator i);
	HRESULT getName(LPOLESTR in, bool hasSource,DWORD serverID,BSTR *ret);
	pair<itemMap::iterator,bool> getItem(bool create, LPOLESTR name);
public:
	Group();
	~Group();
	REGISTER_GENERIC(L"Group")

//DEBUG_REFCOUNT_IMPL(Group);

BEGIN_COM_MAP(Group)
	COM_INTERFACE_ENTRY(IGroup)
	COM_INTERFACE_ENTRY(DIGroup)
	COM_INTERFACE_ENTRY(IPersist)
	COM_INTERFACE_ENTRY(IPersistStorage)
	COM_INTERFACE_ENTRY(IOleItemContainer)
	COM_INTERFACE_ENTRY(IHaveIdentity)
	COM_INTERFACE_ENTRY(IGroupItem)
	// COM_INTERFACE_ENTRY(IClassActivator)
END_COM_MAP()

// IPersistStorage
    STDMETHODIMP GetClassID(CLSID *pClassID);
    STDMETHODIMP IsDirty(void);
    STDMETHODIMP InitNew(IStorage *pStg);
    STDMETHODIMP Load(IStorage *pStg);
    STDMETHODIMP Save(IStorage *pStgSave,BOOL fSameAsLoad);
    STDMETHODIMP SaveCompleted(IStorage *pStgNew);
    STDMETHODIMP HandsOffStorage(void);
// IGroupItem
	STDMETHODIMP Connect(IGroup *group);
	STDMETHODIMP Copy(IBindCtx *bctx, IGroup *dest, IUnknown *aggr, REFIID riid, void **ret);
	STDMETHODIMP Realize(IBindCtx *bctx);
	STDMETHODIMP Remove();
// IOleItemContainer
// -IParseDisplayName
	STDMETHODIMP ParseDisplayName(IBindCtx *, LPOLESTR name, ULONG *retEat, IMoniker **retMon);
// -IOleContainer
	STDMETHODIMP EnumObjects(DWORD flags,IEnumUnknown **ret);
	STDMETHODIMP LockContainer(BOOL bLock);
// .
	STDMETHODIMP GetObject(LPOLESTR name,DWORD speed,IBindCtx *,REFIID riid,void **retObj);
	STDMETHODIMP GetObjectStorage(LPOLESTR name, IBindCtx *,REFIID riid,void **ret);
	STDMETHODIMP IsRunning(LPOLESTR name);
// IHaveIdentity
	STDMETHODIMP get_ServerID(DWORD *ret);
// IGroup
	STDMETHODIMP Paste(IBindCtx *bctx,long flags,IUnknown *src,IMoniker *mkSrc,REFIID riid,void **retObj,IMoniker **retMk);
	STDMETHODIMP CreatePassive(BOOL isStorage,IBindCtx *bctx, IMoniker *mk, IMoniker **retMk, REFIID riid,void **retStg);
	STDMETHODIMP SetMainStorage(IBindCtx *bctx, IMoniker *mk, BOOL isMain);
	STDMETHODIMP AddStgRef(IMoniker *mk);
	STDMETHODIMP ReleaseStgRef(IMoniker *mk);
	STDMETHODIMP GetStgRef(IMoniker *mk, long *ret);
	STDMETHODIMP Unload(IMoniker *mk);
	STDMETHODIMP Up(IMoniker **ret);
	STDMETHODIMP Enum(IEnumITEMINF **ret);
// DIGroup
	STDMETHODIMP Paste(IUnknown *src,  IUnknown *mkSrc,  long flags,  IUnknown **retMk);
	STDMETHODIMP AddStgRef(IUnknown *mk);
	STDMETHODIMP ReleaseStgRef(IUnknown *mk);
	STDMETHODIMP GetStgRef(IUnknown *mk,  long *ret);
	STDMETHODIMP Unload(IUnknown *mk);
	STDMETHODIMP Up(IUnknown **ret);
	STDMETHODIMP get_Items(IDispatch **ret);
	STDMETHODIMP ItemMoniker(BSTR name,  IUnknown **ret);
	STDMETHODIMP ComposeMonikers(IUnknown *mk1,  IUnknown *mk2,  IUnknown **ret);
	STDMETHODIMP Bind(IUnknown *mk,  IUnknown **ret);
// IClassActivator
	STDMETHODIMP GetClassObject(REFCLSID rclsid,DWORD dwClassContext,LCID locale,REFIID riid,void **ppv);
public:
};

#endif //__Group_H_
