/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


/////////////////////////////////////////////////////////////////////////////
// Group internals and IGroup
//
#include "stdafx.h"
#include "GroupClass.h"
#include <queue>
#include "copyITEMINF.h"
#include "ItemInfo.h"
#include "Collection.h"
#include "notFound.h"

Group::Group() : m_noScribble(false), m_anonID(0) {
	CONSTRUCTORTRACE(Group);
	HRESULT hr;
	hr = CREATE(m_idServer,CLSID_IdentityServer);
	_ASSERT(SUCCEEDED(hr));
	CComQIPtr<IGiveIdentity> gi = m_idServer;
	hr = gi->new_ServerID(&m_serverID);
	_ASSERT(SUCCEEDED(hr));
}
Group::~Group() {
	DESTRUCTORTRACE(Group);
	for(itemMap::iterator i = m_data.begin();i!=m_data.end();) {
		if(i->second.p)
			i->second.p->Release();
		// must increment before destroying string and of course
		// before removing the element
		BSTR s = i->first;
		itemMap::iterator j = i++;
		m_data.erase(j);
		SysFreeString(s);
	}
}
HRESULT Group::getName(LPOLESTR in,bool hasSource,DWORD serverID,BSTR *ret) {
	METHTRACE(Group,getName);
	wchar_t buf[50];
	LPOLESTR name;

	if(in && !hasSource) // no source server: name must be OK.
		name = in;
	else if(!in || !wcsncmp(in,L"item",4)) {// anonymous
		DWORD itemID;
		if(in&&hasSource&&serverID) { // name specified and has source ID
			IDENTITY id;
			id.server = serverID;
			id.item = _wtol(in+4);
			if(FAILED(m_idServer->Find(id,m_serverID,&itemID))) {
				IDENTITY id2 = {m_serverID,m_anonID};
				RUN(m_idServer->Equals(id,id2));
				itemID = m_anonID++;
			}
		}
		else itemID = m_anonID++; // if no name was specified, or source doesn't use the identity server, the contribution must have new identity
		swprintf(buf,L"item%d",itemID);
		name = buf;
	}
	else name = in;
	*ret = SysAllocString(name);
	MTrace(GP)("name \"%S\" becomes \"%S\"\n",in,*ret);
	return S_OK;
}
pair<itemMap::iterator,bool> Group::getItem(bool create, LPOLESTR name) {
	METHTRACE(itemMap,iterator);
	itemMap::iterator i = m_data.find(name);
	if(i!=m_data.end() || !create)
        return make_pair(i,false);
	pair<OLECHAR *const,Item> x(SysAllocString(name),Item());
	i = m_data.insert(x).first;
	return make_pair(i,true);
}
HRESULT Group::fetch(IMoniker *mk, itemMap::iterator *iret) {
	METHTRACE(Group,fetch);
	if(!iret)
		ERR(E_POINTER);
	DWORD type;
	RUN(mk->IsSystemMoniker(&type));
	if(type!=MKSYS_ITEMMONIKER)
		ERR(E_FAIL); // should recurse on composite
	CComPtr<IBindCtx> bctx;
	RUN(CreateBindCtx(0,&bctx));
	LPOLESTR s;
	RUN(mk->GetDisplayName(bctx,0,&s));
	if(!wcscmp(s+1,L".."))
		return GROUP_S_PARENT;
	*iret = m_data.find(s+1);
	CoTaskMemFree(s);
	if(*iret==m_data.end())
		ERR(MK_E_NOOBJECT);
	else
		return S_OK;
}
HRESULT Group::findPersistType(IUnknown *o, Item&item) {
	METHTRACE(Group,findPersistType);
	if(SUCCEEDED(o->QueryInterface(riidret(&item.pStream))))
		item.method = persistStream;
	else if(SUCCEEDED(o->QueryInterface(riidret(&item.pStreamInit))))
		item.method = persistStreamInit;
	else if(SUCCEEDED(o->QueryInterface(riidret(&item.pStorage))))
		item.method = persistStorage;
	else if(SUCCEEDED(o->QueryInterface(riidret(&item.p))))
		item.method = persist;
	else ERR(E_FAIL);
	return S_OK;
}
// IHaveIdentity
STDMETHODIMP Group::get_ServerID(DWORD *ret) {
	METHTRACE(Group,get_ServerID);
	*ret = m_serverID;
	return S_OK;
}
// IGroup
struct mkQ : public queue<IMoniker*> {
	~mkQ() { // in case there's an error
		while(!empty()) {
			front()->Release();
			pop();
		}
	}
};
HRESULT pasteCompound(IBindCtx *bctx,DWORD flags,IUnknown *dest,IUnknown *src, IEnumMoniker *em,REFIID riid,void **retObj,IMoniker **retMk) {
    FUNTRACE(pasteCompound);
	mkQ todo,sinceSrc;
	for(mk_iterator i(em);i!=mk_iterator();++i) {
		(*i)->AddRef();
		todo.push(*i);
	}
	CComPtr<IUnknown> s(src), // current source object
		d(dest); // current dest object
	while(!todo.empty()) {
		IMoniker *curr = todo.front();
		todo.pop();
		sinceSrc.push(curr);
		CComPtr<IMoniker> mkRes;
		if(CComQIPtr<IGroup> g = d) { // first try to get the group dest to paste
			// bind on source *up until* the moniker being pasted
			// the last link will be bound inside the dest's Paste.
			while(sinceSrc.size()>1) {
				IMoniker *catchup = sinceSrc.front();
				sinceSrc.pop();
				CComPtr<IMoniker> ms;
				SRUN(FoolishCreatePointerMoniker(s,&ms));
				s.Release();
				SRUN(catchup->BindToObject(bctx,ms,riidret(&s)));
				catchup->Release();
			}
			d.Release();
			SRUN(g->Paste(bctx,flags,s,curr,riidret(&d),&mkRes));
		}
		else { // okay, d is not a group: it had better already support that mk
			// HACK: we only need to bind if the object is needed or the mk string
			// continues.  And binding that last moniker (e.g. GetSubject) can have
			// very bad consequences!
			if(todo.size()||retObj) {
				CComPtr<IMoniker> md;
				SRUN(FoolishCreatePointerMoniker(d,&md));
				d.Release();
				SRUN(curr->BindToObject(bctx,md,riidret(&d)));
			}
			mkRes = curr;
		}
		if(!mkRes)
			SERR(E_FAIL);
		if(retMk) {
			if(*retMk) {
				CComPtr<IMoniker> mkComp;
				SRUN((*retMk)->ComposeWith(mkRes,false,&mkComp));
				(*retMk)->Release();
				(*retMk = mkComp)->AddRef();
			}
			else (*retMk = mkRes)->AddRef();
		}
	}
	return retObj?d->QueryInterface(riid,retObj):S_OK;
}
STDMETHODIMP Group::Paste(IBindCtx *bctx,long flags,IUnknown *src,IMoniker *mk,REFIID riid,void **retObj,IMoniker **retMk) {
	METHTRACE(Group,Paste);
	DWORD serverID(0);
	CComQIPtr<IHaveIdentity> hi(src);
	if(retMk)
		*retMk = 0;
	if(retObj)
		*retObj = 0;
	if(!(flags&(GP_SrcIsObj|GP_SrcIsCont))) // one of these two flags must be set
		ERR(E_INVALIDARG);
	if((flags&GP_SrcIsObj)&&(flags&GP_SrcIsCont)) // but not both!
		ERR(E_INVALIDARG);
	if(!src)
		ERR(E_POINTER);
	if(flags&GP_SrcIsCont) { // simple checks for container-to-container paste
		// if you give a NULL moniker when pasting from another container,
		// that must be because you don't actually have a moniker.  So let's
		// respect that and return a null moniker!
		if(!mk)
			return S_FALSE;
		// check if self is also the source
		if(hi) {
			RUN(hi->get_ServerID(&serverID));
			if(serverID==m_serverID) {
				if(retMk)
					(*retMk = mk)->AddRef();
				if(retObj) {
					CComPtr<IMoniker> mkS;
					RUN(FoolishCreatePointerMoniker(src,&mkS));
					RUN(mk->BindToObject(bctx,mkS,riid,retObj));
				}
				return S_OK;
			}
		}
	}
	LPOLESTR s = 0;
	if(mk) { // mk==0 means anonymous
		CComPtr<IEnumMoniker> em;
		RUN(mk->Enum(true,&em));
		if(em)
			return pasteCompound(bctx,flags,(IGroup*)this,src,em,riid,retObj,retMk);
		DWORD type;
		RUN(mk->IsSystemMoniker(&type));
		if(type==MKSYS_CLASSMONIKER) {
			// easy
			(*retMk = mk)->AddRef();
			return S_OK;
		}
		else if(type!=MKSYS_ITEMMONIKER)
			ERR(E_FAIL);
		RUN(mk->GetDisplayName(bctx,0,&s)); // starts with separator character
		if(!wcscmp(s+1,L"..")) {
			CoTaskMemFree(s);
			if(!m_parent)
				ERR(E_FAIL);
			if(retObj)
				RUN(m_parent->QueryInterface(riid,retObj));
			(*retMk = mk)->AddRef();
			return S_OK;
		}
	}

	// translate the name
	CComBSTR bstrName;
	RUN(getName(s?s+1:0,flags&GP_SrcIsCont,serverID,&bstrName));
	if(s)
		CoTaskMemFree(s);

	// set up return moniker
	if(retMk)
		RUN(CreateGroupMoniker(m_serverID,bstrName,retMk));

	// look item up
    pair<itemMap::iterator,bool> add = getItem(true,bstrName);
	pair<OLECHAR*const,Item> &p = *add.first;
	Item &item = p.second;
    if(!add.second) { // there was already something there
        if((flags&GP_Replace)==0) {
		    MTrace(GP)("item \"%S\" was already there\n",(BSTR)bstrName);
            if(retObj) {
                if(item.p)
			        return item.p->QueryInterface(riid,retObj);
                *retObj = 0; // odd case.  could wake it instead?
            }
		    return S_OK;
        }
        if(item.p)
            item.p->Release();
        item.p = 0;
    }

	if(item.method==persistPasting) { // in the process of creating
		MTrace(GP)("item \"%S\" is being created (lower on the stack)\n",(BSTR)bstrName);
		return GROUP_S_PASTING;
	}
	if(item.method!=persistNone && !item.pending) { // it has already been pasted, but it's passive
		MTrace(GP)("item \"%S\" already there & passive\n",(BSTR)bstrName);
		if(retObj)
			return GetObject(p.first,BINDSPEED_INDEFINITE,bctx,riid,retObj);
		return S_OK;
	}
	MTrace(GP)("adding your \"%S\"\n",(BSTR)bstrName);

	// start with stgref==0

	item.method = persistPasting;
	bool copied = false;
	// bind to source
	CComPtr<IUnknown> o;
	CComPtr<IMoniker> mkS;
	if(flags&GP_SrcIsObj) {
		o = src;
		if(CComQIPtr<IGroupItem> pg = o)
			RUN(pg->Connect(this));
	}
	else {
		CComPtr<IUnknown> o2;
		RUN(FoolishCreatePointerMoniker(src,&mkS));
		RUN(mk->BindToObject(bctx,mkS,riidret(&o2)));
		// ask it to transfer itself.  Copy should call Connect (before it initializes)
		if(CComQIPtr<IGroupItem> gi = o2) {
			RUN(gi->Copy(bctx,(IGroup*)this,0,riidret(&o)));
			copied = true;
		}
		else
			o = o2;
	}

	if(item.method==persistStorage) {  // storage may be set up by CreatePassive
		RUN(o->QueryInterface(riidret(&item.pStorage)));
	}
	else {
		if(FAILED(findPersistType(o,item))) {
			m_data.erase(p.first);
			ERR(E_INVALIDARG); // (no persistence interface)
		}
	}
	if(item.method==persistStorage) {
		if(!item.storage)
			RUN(open(m_storage,bstrName,&item.storage));
		// transfering an IPersistStorage object from container to container means
		// copying the data as well.  The group client is responsible for calling
		// HandsOffStorage and SaveCompleted as necessary to let the object
		// know which storage to hold onto.  (As there is no exclusive ownership
		// model (yet?) in group persistence.)
		if(flags&GP_SrcIsCont) {
			CComPtr<IStorage> stg;
			RUN(mk->BindToStorage(bctx,mkS,riidret(&stg)));
			LOOK(stg->CopyTo(0,0,0,item.storage)); // sometimes fails - why???
		}

		// at the moment there is only one class that implements both IPersistStorage
		// and IGroupItem, and that is Group.  A peculiarity here: although Copy
		// can be considered an initialization of the returned object, it doesn't
		// have a way to pass the storage of IPersistStorage::InitNew.  Group definitely
		// needs that, and InitNew doesn't do anything harmful after Copy, so it makes
		// sense to call InitNew on any Copied object.
		if(flags&GP_IsMainStg || copied) {
			item.isMainStorage = true;
			if(copied) {
				RUN(item.pStorage->InitNew(item.storage));
			}
		}
	}
	RUN(item.p->GetClassID(&item.clsid));

	return retObj?o->QueryInterface(riid,retObj):S_OK;
}
HRESULT bindBut(IBindCtx *bctx,IGroup *start, IMoniker *mk, IMoniker **retBody, IMoniker **retLast, IGroup **retGroup) {
    FUNTRACE(bindBut);
	CComPtr<IEnumMoniker> em;
	SRUN(mk->Enum(false,&em));
	CComPtr<IMoniker> last,inverse,cut,left;
	SRUN(em->Next(1,&last,0));
	SRUN(last->Inverse(&inverse));
	SRUN(mk->ComposeWith(inverse,false,&cut));
	SRUN(FoolishCreatePointerMoniker(start,&left));
	if(retBody)
		(*retBody = cut)->AddRef();
	if(retLast)
		(*retLast = last)->AddRef();
	return cut->BindToObject(bctx,left,IID_IGroup,(void**)retGroup);
}
STDMETHODIMP Group::CreatePassive(BOOL isStorage,IBindCtx *bctx, IMoniker *mk, IMoniker **retMk, REFIID riid,void **retElem) {
	METHTRACE(Group,CreatePassive);
	LPOLESTR s=0;
	if(mk) {
		DWORD type;
		RUN(mk->IsSystemMoniker(&type));
		switch(type) {
		case MKSYS_GENERICCOMPOSITE: {
			CComPtr<IGroup> dest;
			CComPtr<IMoniker> body, last, newLast;
			RUN(bindBut(bctx,this,mk,&body,&last,&dest));
			RUN(dest->CreatePassive(isStorage,bctx,last,&newLast,riid,retElem));
			RUN(body->ComposeWith(newLast,false,retMk));
			return S_OK;
		}
		case MKSYS_ITEMMONIKER:
			RUN(mk->GetDisplayName(bctx,0,&s)); // starts with separator character
			break;
		default:
			ERR(E_FAIL);
		}
	}
	// translate the name
	CComBSTR bstrName;
	RUN(getName(s?s+1:0,false,0,&bstrName));
	if(s)
		CoTaskMemFree(s);

	// look item up
    pair<itemMap::iterator,bool> add = getItem(true,bstrName);
    if(!add.second)
        ERR(E_FAIL);
	pair<OLECHAR *const, Item> &p = *add.first;
	Item &item = p.second;

	// set up return moniker
	if(retMk)
		RUN(CreateGroupMoniker(m_serverID,bstrName,retMk));
	item.pending = item.element = true;
	if(isStorage) {
		if(!item.storage)
			RUN(open(m_storage,p.first,&item.storage));
		item.method = persistStorage;
		if(retElem)
			RUN(item.storage->QueryInterface(riid,retElem));
	}
	else {
		if(!item.stream)
			RUN(open(m_storage,p.first,&item.stream));
		item.method = persistStream;
		if(retElem)
			RUN(item.stream->QueryInterface(riid,retElem));
	}
	return S_OK;
}
STDMETHODIMP Group::SetMainStorage(IBindCtx *bctx, IMoniker *mk, BOOL isMain) {
	METHTRACE(Group,SetMainStorage);
	HRESULT hr;
	if(!mk)
		ERR(E_INVALIDARG);
	DWORD type;
	RUN(mk->IsSystemMoniker(&type));
	switch(type) {
	case MKSYS_GENERICCOMPOSITE: {
		CComPtr<IMoniker> last;
		CComPtr<IGroup> dest;
		RUN(bindBut(bctx,this,mk,0,&last,&dest));
		return dest->SetMainStorage(bctx,last,isMain);
	}
	case MKSYS_ITEMMONIKER:
		break;
	default:
		ERR(E_FAIL);
	}
	itemMap::iterator i;
	RUN(hr = fetch(mk,&i));
	if(hr==GROUP_S_PARENT)
		ERR(E_INVALIDARG);
	if(i->second.method==persistStorage)
		i->second.isMainStorage = isMain!=0;
	return S_OK;
}
// applies a functor to any groups in the path of a moniker.
// terminates when the moniker is up or f return S_FALSE.
template<class F>
HRESULT apply(IBindCtx *bctx, IGroup *start, IMoniker *mk,F &f) {
    FUNTRACE(apply);
	if(!mk)
		SERR(E_INVALIDARG);
	DWORD type;
	SRUN(mk->IsSystemMoniker(&type));
	if(type==MKSYS_GENERICCOMPOSITE) {
		CComPtr<IEnumMoniker> mks;
		SRUN(mk->Enum(true,&mks));
		CComPtr<IUnknown> u = start;
		for(mk_iterator i(mks);i!=mk_iterator();) {
			if(CComQIPtr<IGroup> g = u) {
				HRESULT hr = f(g,*i);
				SRUN(hr);
				if(hr==S_FALSE)
					break;
			}
			CComPtr<IMoniker> mk = *i;
			++i;
			// hackish: it causes trouble if you bind GetSubjectMonikers needlessly,
			// so if this is the last moniker there's no need to bind it.  But this solution
			// doesn't help where there's more than one non-group moniker.  Maybe a
			// better solution would use bind speeds to specify "don't DO anything!"
			if(i!=mk_iterator()) {
				CComPtr<IMoniker> mkU;
				SRUN(FoolishCreatePointerMoniker(u,&mkU));
				u.Release();
				// HACK: work around fact you can't bind monikers to things that are being pasted,
				// although StgRef functions work OK
				HRESULT hr = mk->BindToObject(bctx,mkU,riidret(&u));
				if(hr==E_PENDING)
					return S_OK;
				SRUN(hr);
			}
		}
		return S_OK;
	}
	else
		return f(start,mk);
}
// runs AddStgRef or ReleaseStgRef on groups in a moniker path.
struct refFunc {
	typedef HRESULT (__stdcall IGroup::*func)(IMoniker *);
	func f;
	refFunc(func f) : f(f) {};
	HRESULT operator ()(IGroup *g,IMoniker *mk) {
		return (g->*f)(mk);
	}
};
STDMETHODIMP Group::AddStgRef(IMoniker *mk) {
	METHTRACE(Group,AddStgRef);
	HRESULT hr;
	if(!mk)
		return S_FALSE;
	DWORD type;
	RUN(mk->IsSystemMoniker(&type));
	if(type==MKSYS_GENERICCOMPOSITE) {
		CComPtr<IBindCtx> bctx;
		RUN(CreateBindCtx(0,&bctx));
		refFunc f(&IGroup::AddStgRef);
		RUN(apply(bctx,this,mk,f));
		return S_OK;
	}
	else if(type==MKSYS_CLASSMONIKER)
		return S_FALSE;
	itemMap::iterator i;
	hr = fetch(mk,&i);
	if(hr==MK_E_NOOBJECT) {
		GNOTFOUND(IGroup,AddStgRef,mk);
		ERR(MK_E_NOOBJECT);
	}
	if(hr==GROUP_S_PARENT)
		return S_FALSE;
	RUN(hr);
	ULONG ct = ++i->second.count;
	MTrace(SR)("Group\tAddStgRef\t%p\t%S\t%d\n",this,i->first,ct);
	return S_OK;
}
STDMETHODIMP Group::ReleaseStgRef(IMoniker *mk) {
	METHTRACE(Group,ReleaseStgRef);
	HRESULT hr;
	if(!mk)
		return S_FALSE;
	DWORD type;
	RUN(mk->IsSystemMoniker(&type));
	if(type==MKSYS_GENERICCOMPOSITE) {
		CComPtr<IBindCtx> bctx;
		RUN(CreateBindCtx(0,&bctx));
		refFunc f(&IGroup::ReleaseStgRef);
		RUN(apply(bctx,this,mk,f));
		return S_OK;
	}
	else if(type==MKSYS_CLASSMONIKER)
		return S_FALSE;
	itemMap::iterator i;
	hr = fetch(mk,&i);
	if(hr==MK_E_NOOBJECT) {
		GNOTFOUND(IGroup,ReleaseStgRef,mk);
		ERR(MK_E_NOOBJECT);
	}
	if(hr==GROUP_S_PARENT)
		return S_FALSE;
	RUN(hr);
	long ct = --i->second.count;
	MTrace(SR)("Group\tReleaseStgRef\t%p\t%S\t%d\n",this,i->first,ct);
	if(ct<1) { // Release from 0 or 1 deletes.
		BSTR name = i->first;
	    MTrace(GR)("Group\tremove\t%p\t%S\t%d\n",this,name);
		if(CComQIPtr<IGroupItem> gi = i->second.p) {
			RUN(gi->Remove());
			RUN(unload(i));
		}
		m_data.erase(i);
		if(!wcsncmp(name,L"item",4)) {
			IDENTITY id = {m_serverID,_wtol(name+4)};
			RUN(m_idServer->Remove(id));
		}
		SysFreeString(name);
		return S_FALSE;
	}
	return S_OK;
}
STDMETHODIMP Group::GetStgRef(IMoniker *mk, long *ret) {
	METHTRACE(Group,GetStgRef);
	itemMap::iterator i;
	RUN(fetch(mk,&i));
	*ret = i->second.count;
	return S_OK;
}
HRESULT Group::unload(itemMap::iterator i) {
	METHTRACE(Group,unload);
	itemMap::value_type &debug = *i;
	if(i->second.p) {
		if(CComQIPtr<IGroupItem> gi = i->second.p)
			RUN(gi->Connect(0)); // get it to release me
		if(i->second.method==persistStorage && i->second.isMainStorage)
			RUN(i->second.pStorage->HandsOffStorage()); // get it to release storage
		i->second.p->Release();
		i->second.p = 0;
	}
	if(i->second.storage)
		i->second.storage.Release();
	return S_OK;
}
STDMETHODIMP Group::Unload(IMoniker *mk) {
	METHTRACE(Group,Unload);
	// sad but true: there's nothing incremental in this implementation.
	// So we do a complete save and then unload.
	if(m_storage && !m_noScribble)
		RUN(Save(m_storage,true));
	if(mk) {
		DWORD type;
		RUN(mk->IsSystemMoniker(&type));
		if(type==MKSYS_GENERICCOMPOSITE) {
			CComPtr<IBindCtx> bctx;
			RUN(CreateBindCtx(0,&bctx));
			CComPtr<IMoniker> last;
			CComPtr<IGroup> dest;
			RUN(bindBut(bctx,this,mk,0,&last,&dest));
			return dest->Unload(last);
		}
		itemMap::iterator i;
		HRESULT hr;
		RUN(hr = fetch(mk,&i));
		if(hr==GROUP_S_PARENT)
			ERR(E_INVALIDARG);
		RUN(unload(i));
	}
	else
		for(itemMap::iterator i = m_data.begin();i!=m_data.end();i++)
			RUN(unload(i));
	return S_OK;
}
STDMETHODIMP Group::Up(IMoniker **ret) {
	METHTRACE(Group,Up);
	return CreateGroupMoniker(m_serverID,L"..",ret);
}
struct enumITEMINF : CComObject<CComEnum<IEnumITEMINF,&IID_IEnumITEMINF,ITEMINF,copyITEMINF> > {
	enumITEMINF(ITEMINF *begin,ITEMINF *end) {
		Init(begin,end,0,AtlFlagTakeOwnership); // cannot fail since no memory alloc.
	}
};

STDMETHODIMP Group::Enum(IEnumITEMINF **ret) {
	METHTRACE(Group,Enum);
	ITEMINF *ray = new ITEMINF[m_data.size()];
	int head = 0;
	for(itemMap::iterator i = m_data.begin();i != m_data.end(); i++) {
		ray[head].clsid = i->second.clsid;
		if(ray[head].o = i->second.p)
			ray[head].o->AddRef();
		RUN(CreateGroupMoniker(m_serverID,i->first,&ray[head].mk));
		ray[head].count = i->second.count;
		head++;
	}
	(*ret = new enumITEMINF(ray,ray+m_data.size()))->AddRef();
	return S_OK;
}
// DIGroup
STDMETHODIMP Group::Paste(IUnknown *src,  IUnknown *mkSrc,  long flags,  IUnknown **retMk) {
    METHTRACE(Group,Paste);
	CComPtr<IBindCtx> bctx;
	RUN(CreateBindCtx(0,&bctx));
	CComPtr<IMoniker> mk2;
	if(mkSrc)
		RUNRET(mkSrc->QueryInterface(riidret(&mk2)));
	return Paste(bctx,flags,src,mk2,IID_NULL,0,(IMoniker**)retMk);
}
STDMETHODIMP Group::AddStgRef(IUnknown *mk) {
    METHTRACE(Group,AddStgRef);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return AddStgRef(mk2);
}
STDMETHODIMP Group::ReleaseStgRef(IUnknown *mk) {
    METHTRACE(Group,ReleaseStgRef);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return ReleaseStgRef(mk2);
}
STDMETHODIMP Group::GetStgRef(IUnknown *mk,  long *ret) {
    METHTRACE(Group,GetStgRef);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return GetStgRef(mk2,ret);
}
STDMETHODIMP Group::Unload(IUnknown *mk) {
    METHTRACE(Group,Unload);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return Unload(mk2);
}
STDMETHODIMP Group::Up(IUnknown **ret) {
    METHTRACE(Group,Up);
	return Up((IMoniker**)ret);
}
STDMETHODIMP Group::get_Items(IDispatch **ret) {
    METHTRACE(Group,get_Items);
	// fantastically inefficient to copy all data into an enumerator
	// just to take it all back out and put it into another one, but
	// I think this is the safest (simplest) way to write this.
	CComPtr<IEnumITEMINF> eii;
	RUN(Enum(&eii));
	ITEMINF *iiray = new ITEMINF[m_data.size()];
	DWORD nub;
	RUN(eii->Next(m_data.size(),iiray,&nub));
	CComVariant *ray = new CComVariant[m_data.size()];
	int j = 0;
	for(itemMap::iterator i = m_data.begin();i!=m_data.end();++i,++j) {
		itemInfo *ii = new itemInfo;
		ii->Init(iiray[j]);
		ray[j] = ii;
	}
	while(--j>=0)
		copyITEMINF::destroy(&iiray[j]);
	(*ret = new collection(m_data.size(),ray))->AddRef();
	return S_OK;
}
STDMETHODIMP Group::ItemMoniker(BSTR name,  IUnknown **ret) {
    METHTRACE(Group,ItemMoniker);
	return CreateItemMoniker(L"!",name,(IMoniker**)ret);
}
STDMETHODIMP Group::ComposeMonikers(IUnknown *mk1,  IUnknown *mk2,  IUnknown **ret) {
    METHTRACE(Group,ComposeMonikers);
	if(!mk1 || !mk2)
		ERR(E_POINTER);
	CComPtr<IMoniker> mk1b,mk2b;
	RUNRET(mk1->QueryInterface(riidret(&mk1b)));
	RUNRET(mk2->QueryInterface(riidret(&mk2b)));
	return mk1b->ComposeWith(mk2b,false,(IMoniker**)ret);
}
STDMETHODIMP Group::Bind(IUnknown *mk,  IUnknown **ret) {
    METHTRACE(Group,Bind);
	CComPtr<IBindCtx> bctx;
	RUN(CreateBindCtx(0,&bctx));
	CComPtr<IMoniker> mkThis, mk2;
	RUN(FoolishCreatePointerMoniker(GetUnknown(),&mkThis));
	RUN(mk->QueryInterface(riidret(&mk2)));
	return mk2->BindToObject(bctx,mkThis,riidret(ret));
}
