/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "ItemSet.h"
#include "iterEnum.h"
#include "Collection.h"
#include "releaseDUA.h"

ItemSet::ItemSet() : m_realized(false) {
	CONSTRUCTORTRACE(ItemSet);
}
ItemSet::~ItemSet() {
	DESTRUCTORTRACE(ItemSet);
	RUN(clear(false));
}
HRESULT ItemSet::add(IMoniker *mk, bool addStgRef) {
	METHTRACE(ItemSet,add);
	if(!m_group&&addStgRef)
		ERR(E_PENDING);
	if(!mk)
		ERR(E_POINTER);
	if(m_set.insert(mk).second) {
		{
			LPOLESTR name;
			RUN(mk->GetDisplayName(BindCtx(),0,&name));
			MTrace(SA)("ItemSet\tadd\t%p\t\"%S\"\n",this,name+1);
			CoTaskMemFree(name);
		}
		mk->AddRef();
		if(addStgRef)
			RUNRET1(m_group->AddStgRef(mk),MK_E_NOOBJECT);
		for(IUnknown **iu = m_vec.begin();iu!=m_vec.end();++iu)
			if(*iu)
				((IItemSetEvents*)*iu)->Added(mk);
		return S_OK;
	}
	return S_FALSE;
}
HRESULT ItemSet::clear(bool releaseStgRef) {
    METHTRACE(ItemSet,clear);
	// first clear
	mkSet backup = m_set;
	m_set.clear();
	// then fire event
	for(mkSet::iterator i = backup.begin();i!=backup.end();)  {
		for(int ii = 0; ii<m_vec.GetSize(); ++ii) //IUnknown **iu = m_vec.begin();iu!=m_vec.end();++iu)
			if(IUnknown *u = m_vec.GetAt(ii))
				((IItemSetEvents*)u)->Removed(*i);
		// release stg ref
		if(releaseStgRef)
			RUN(m_group->ReleaseStgRef(*i));
		(*i++)->Release(); // must increment before release because op++ relies on hash
	}
	return S_OK;
}
// IItemSet
STDMETHODIMP ItemSet::get_Count(long *ret) {
	METHTRACE(ItemSet,get_Count);
	*ret = m_set.size();
	return S_OK;
}
STDMETHODIMP ItemSet::Add(IMoniker *mk) {
	METHTRACE(ItemSet,Add);
	return add(mk,true);
}
STDMETHODIMP ItemSet::Remove(IMoniker *mk) {
	METHTRACE(ItemSet,Remove);
	if(!mk)
		ERR(E_POINTER);
	mkSet::iterator i = m_set.find(mk);
	if(i==m_set.end())
		ERR(E_FAIL);
	// release the moniker we've been holding (not mk!)
	IMoniker *storedMk = *i;
	m_set.erase(i);
	storedMk->Release();
	// fire event
	for(IUnknown **iu = m_vec.begin();iu!=m_vec.end();++iu)
		if(*iu)
			((IItemSetEvents*)*iu)->Removed(mk);
	// then release storage ref (after event because may destroy effectiveness of moniker)
	RUN(m_group->ReleaseStgRef(mk));
	return S_OK;
}
STDMETHODIMP ItemSet::Check(IMoniker *mk) {
	METHTRACE(ItemSet,Check);
	if(!mk)
		ERR(E_POINTER);
	mkSet::iterator i = m_set.find(mk);
	return i==m_set.end()?S_FALSE:S_OK;
}
STDMETHODIMP ItemSet::Clear() {
	METHTRACE(ItemSet,Clear);
	return clear(true);
}
STDMETHODIMP ItemSet::Intersect(IItemSet *is) {
    METHTRACE(ItemSet,Intersect);
	for(mkSet::iterator im0 = m_set.begin(); im0!=m_set.end();) {
		mkSet::iterator next = im0;
		++next;
		if(is->Check(*im0)==S_FALSE)
			RUN(Remove(*im0));
		im0 = next;
	}
	return S_OK;
}
STDMETHODIMP ItemSet::Union(IItemSet *is) {
    METHTRACE(ItemSet,Union);
	CComPtr<IEnumMoniker> em;
	RUN(is->Enum(&em));
	for(mk_iterator im(em); im!=mk_iterator(); ++im)
		RUN(Add(*im));
	return S_OK;
}
STDMETHODIMP ItemSet::Assign(IItemSet *is) {
    METHTRACE(ItemSet,Assign);
	RUN(Intersect(is));
	RUN(Union(is));
	return S_OK;
}
STDMETHODIMP ItemSet::Enum(IEnumMoniker **ret) {
	METHTRACE(ItemSet,Enum);
	(*ret = new IterEnum<IEnumMoniker,&IID_IEnumMoniker,IMoniker*,mkSet::iterator,
		_CopyInterface<IMoniker> >(m_set.begin(),m_set.end(),m_set.begin()))->AddRef();
	return S_OK;
}
// DIItemSet
STDMETHODIMP ItemSet::Add(IUnknown *mk) {
    METHTRACE(ItemSet,Add);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return Add(mk2);
}
STDMETHODIMP ItemSet::Remove(IUnknown *mk) {
    METHTRACE(ItemSet,Remove);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return Remove(mk2);
}
STDMETHODIMP ItemSet::Check(IUnknown *mk, VARIANT_BOOL *ret) {
    METHTRACE(ItemSet,Check);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	*ret = Check(mk2)==S_OK?VARIANT_TRUE:VARIANT_FALSE;
	return S_OK;
}
STDMETHODIMP ItemSet::get_Items(IDispatch **ret) {
    METHTRACE(ItemSet,get_Items);
	(*ret = new collection(make_colldata(m_set)))->AddRef();
	return S_OK;
}
// IGroupItem
STDMETHODIMP ItemSet::Connect(IGroup *group) {
	METHTRACE(ItemSet,Connect);
	m_group = group;
	/*
	if(!m_group)
		releaseDUA(m_vec);
		*/
	return S_OK;
}
STDMETHODIMP ItemSet::Copy(IBindCtx *bctx, IGroup *dest, IUnknown *aggr, REFIID riid, void **ret) {
	METHTRACE(ItemSet,Copy);
	RUN(CoCreateInstance(CLSID_ItemSet,aggr,CLSCTX_ALL,riid,ret));
	CComQIPtr<IItemSet> is = (IUnknown*)*ret;
	RUN(CComQIPtr<IGroupItem>(is)->Connect(dest));
	for(mkSet::iterator i = m_set.begin();i!=m_set.end();i++) {
		CComPtr<IMoniker> mkc;
		RUN(dest->Paste(bctx,GP_SrcIsCont,m_group,*i,IID_NULL,0,&mkc));
		RUN(is->Add(mkc));
	}
	return S_OK;
}
STDMETHODIMP ItemSet::Realize(IBindCtx *bctx) {
	METHTRACE(ItemSet,Realize);
	if(m_realized)
		return S_FALSE;
	CComPtr<IBindCtx> bctxRef;
	if(!bctx) {
		RUN(CreateBindCtx(0,&bctx));
		bctxRef.Attach(bctx);
	}
	m_realized = true;
	CComPtr<IMoniker> mkG;
	RUN(FoolishCreatePointerMoniker(m_group,&mkG));
	for(mkSet::iterator i = m_set.begin();i!=m_set.end();i++) {
		CComPtr<IUnknown> u;
		RUN((*i)->BindToObject(bctx,mkG,riidret(&u)));
		if(CComQIPtr<IGroupItem> gi = u)
			RUN(gi->Realize(bctx));
	}
	return S_OK;
}
STDMETHODIMP ItemSet::Remove() {
	METHTRACE(ItemSet,Remove);
	for(mkSet::iterator i = m_set.begin();i!=m_set.end();i++)
		RUN(m_group->ReleaseStgRef(*i));
	return S_OK;
}
// IPersistStream
STDMETHODIMP ItemSet::GetClassID(CLSID *pClassID) {
	METHTRACE(ItemSet,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP ItemSet::IsDirty(void) {
	METHTRACE(ItemSet,IsDirty);
	return S_FALSE;
}
STDMETHODIMP ItemSet::Load(IStream *pStm) {
	METHTRACE(ItemSet,Load);
	unsigned n;
	RUN(pStm->Read(&n,sizeof(unsigned),0));
	while(n--) {
		CComPtr<IMoniker> mk;
		RUN(FixedOleLoadFromStream(pStm,riidret(&mk)));
		RUN(add(mk,false));
	}
	return S_OK;
}
STDMETHODIMP ItemSet::Save(IStream *pStm,BOOL fClearDirty) {
	METHTRACE(ItemSet,Save);
	unsigned n = m_set.size();
	RUN(pStm->Write(&n,sizeof(unsigned),0));
	for(mkSet::iterator i = m_set.begin();i!=m_set.end();i++)
		RUN(FixedOleSaveToStream(*i,pStm));
	return S_OK;
}
STDMETHODIMP ItemSet::GetSizeMax(ULARGE_INTEGER *pcbSize) {
	METHTRACE(ItemSet,GetSizeMax);
	pcbSize->QuadPart = sizeof(unsigned);
	for(mkSet::iterator i = m_set.begin();i!=m_set.end();i++) {
		ULARGE_INTEGER ul;
		RUN((*i)->GetSizeMax(&ul));
		pcbSize->QuadPart += ul.QuadPart;
	}
	return S_OK;
}
