/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "OnePointerPersistor.h"

OnePointerPersistor::OnePointerPersistor() : m_realized(false) {
	CONSTRUCTORTRACE(OnePointerPersistor);
}
OnePointerPersistor::~OnePointerPersistor() {
	DESTRUCTORTRACE(OnePointerPersistor);
}
// DIPersistor
STDMETHODIMP OnePointerPersistor::SetSubject(VARIANT_BOOL realized,IUnknown *subject) {
	METHTRACE(OnePointerPersistor,SetSubject);
	m_subject = subject;
	m_realized = realized!=0; // supports "realize this" mode
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::get_Subject(IUnknown **ret) {
	METHTRACE(OnePointerPersistor,GetSubject);
	m_subject.CopyTo(ret);
	return S_OK;
}
// IHoldOnePointer
STDMETHODIMP OnePointerPersistor::put_Moniker(IMoniker *mk) {
	METHTRACE(OnePointerPersistor,put_Moniker);
	if(!m_group)
		ERR(E_PENDING);
	CComPtr<IMoniker> old = m_moniker;
	m_moniker = mk;
	RUN(m_group->AddStgRef(mk));
	if(old)
		RUN(m_group->ReleaseStgRef(old));
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::get_Moniker(IMoniker **ret) {
	METHTRACE(OnePointerPersistor,get_Moniker);
	m_moniker.CopyTo(ret);
	return S_OK;
}
// DIHoldOnePointer
STDMETHODIMP OnePointerPersistor::put_Moniker(IUnknown *mk) {
    METHTRACE(OnePointerPersistor,put_Moniker);
	CComPtr<IMoniker> mk2;
	if(mk)
		RUNRET(mk->QueryInterface(riidret(&mk2)));
	return put_Moniker(mk2);
}
STDMETHODIMP OnePointerPersistor::get_Moniker(IUnknown **ret) {
    METHTRACE(OnePointerPersistor,get_Moniker);
	return get_Moniker((IMoniker**)ret);
}
//IGroupItem
STDMETHODIMP OnePointerPersistor::Connect(IGroup *group) {
	METHTRACE(OnePointerPersistor,Connect);
	m_group = group;
	if(!m_group)
		m_subject.Release();
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::Copy(IBindCtx *bctx, IGroup *dest, IUnknown *aggr, REFIID riid, void **ret) {
	METHTRACE(OnePointerPersistor,Copy);
	RUN(CoCreateInstance(CLSID_OnePointerPersistor,aggr,CLSCTX_ALL,riid,ret));
	CComQIPtr<IHoldOnePointer> hop = (IUnknown*)*ret;
	CComPtr<IMoniker> mk2;
	RUN(dest->Paste(bctx,GP_SrcIsCont,m_group,m_moniker,IID_NULL,0,&mk2));
	RUN(CComQIPtr<IGroupItem>(hop)->Connect(dest));
	RUN(hop->put_Moniker(mk2));
	RUN(CComQIPtr<DIPersistor>(hop)->SetSubject(m_realized,m_subject));
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::Realize(IBindCtx *bctx) {
	METHTRACE(OnePointerPersistor,Realize);
	if(m_realized)
		return S_FALSE;
	CComPtr<IBindCtx> bctxRef;
	if(!bctx) {
		RUN(CreateBindCtx(0,&bctx));
		bctxRef.Attach(bctx);
	}
	m_realized = true;
	CComPtr<IMoniker> mkG;
	RUN(FoolishCreatePointerMoniker(m_group,&mkG));
	CComPtr<IUnknown> pointer;
	if(m_moniker)
		RUN(m_moniker->BindToObject(bctx,mkG,riidret(&pointer)));
	CComQIPtr<INeedOnePointer> nop = m_subject;
	if(!nop)
		ERR(E_UNEXPECTED);
	return nop->TakePointer(pointer,m_moniker);
}
STDMETHODIMP OnePointerPersistor::Remove() {
	METHTRACE(OnePointerPersistor,Remove);
	return m_group->ReleaseStgRef(m_moniker);
}
// IPersistStream
STDMETHODIMP OnePointerPersistor::GetClassID(CLSID *pClassID) {
	METHTRACE(OnePointerPersistor,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::IsDirty(void) {
	METHTRACE(OnePointerPersistor,IsDirty);
	return S_FALSE;
}
STDMETHODIMP OnePointerPersistor::Load(IStream *pStm) {
	METHTRACE(OnePointerPersistor,Load);
	RUN(FixedOleLoadFromStream(pStm,riidret(&m_moniker)));
	RUN(FixedOleLoadFromStream(pStm,riidret(&m_subject)));
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::Save(IStream *pStm,BOOL fClearDirty) {
	METHTRACE(OnePointerPersistor,Save);
	RUN(FixedOleSaveToStream(m_moniker,pStm));
	RUN(FixedOleSaveToStream(m_subject,pStm));
	return S_OK;
}
STDMETHODIMP OnePointerPersistor::GetSizeMax(ULARGE_INTEGER *pcbSize) {
	METHTRACE(OnePointerPersistor,GetSizeMax);
	ULARGE_INTEGER ul;
	pcbSize->QuadPart = 0;
	RUN(m_moniker->GetSizeMax(&ul));
	pcbSize->QuadPart += ul.QuadPart;
	if(CComQIPtr<IPersistStream> ps = m_subject) {
		RUN(ps->GetSizeMax(&ul));
		pcbSize->QuadPart += ul.QuadPart;
	}
	else if(CComQIPtr<IPersistStreamInit> psi = m_subject) {
		RUN(psi->GetSizeMax(&ul));
		pcbSize->QuadPart += ul.QuadPart;
	}
	else if(CComQIPtr<IPersist> p = m_subject)
		pcbSize->QuadPart += sizeof(CLSID);
	return S_OK;
}
