/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


/////////////////////////////////////////////////////////////////////////////
// Group-IPersistStorage
//
#include "stdafx.h"
#include "GroupClass.h"

const LPOLESTR g_itemsName = L"GroupItems",
	g_itemsTempName = L"GroupItems.Temp",
	g_indexName = L"GroupIndex",
	g_namesName = L"GroupNames";
HRESULT Group::takeStorage(IStorage *pStg) {
    METHTRACE(Group,takeStorage);
	m_storage = pStg;
	RUN(open(m_storage,g_itemsName,&m_itemStream));
	RUN(open(m_storage,g_indexName,&m_indexStream));
	// mark the storage with the implementation of group storage that I
	// am, because the CLSID on the storage may be that of my client:
	RUN(m_indexStream->Write(&GetObjectCLSID(),sizeof(CLSID),0));
	return S_OK;
}
HRESULT Group::storageEvent(enum sEvent event) {
	METHTRACE(Group,storageEvent);
	CComPtr<IBindCtx> bctx;
	if(event==se_saveCompletedNew)
		RUN(CreateBindCtx(0,&bctx));
	for(itemMap::iterator i = m_data.begin(); i!=m_data.end();i++)
		if(i->second.method==persistStorage && i->second.pStorage) {
			IPersistStorage *p = i->second.pStorage;
			CComPtr<IStorage> s;
			switch(event) {
			case se_handsOff:
				if(i->second.isMainStorage)
					RUN(p->HandsOffStorage());
				i->second.storage.Release();
				break;
			case se_saveCompletedNew:
				if(i->second.isMainStorage)
					RUN(GetObjectStorage(i->first,bctx,riidret(&s)));
			case se_saveCompleted: // if entering here, s==0
				if(i->second.isMainStorage)
					RUN(p->SaveCompleted(s));
				break;
			default:
				ERR(E_UNEXPECTED);
			}
		}
	return S_OK;
}
STDMETHODIMP Group::GetClassID(CLSID *pClassID) {
	METHTRACE(Group,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP Group::IsDirty(void) {
	METHTRACE(Group,IsDirty);
	HRESULT hr = S_FALSE;
	for(itemMap::iterator i = m_data.begin();i!=m_data.end();i++) {
		switch(i->second.method) {
		case persist:
			hr = S_OK;
			break;
		case persistStorage:
			hr = i->second.pStorage->IsDirty();
			break;
		case persistStream:
			hr = i->second.pStream->IsDirty();
			break;
		case persistStreamInit:
			hr = i->second.pStreamInit->IsDirty();
			break;
		default:
			ERR(E_UNEXPECTED);
		}
		if(hr!=S_FALSE)
			break;
	}
	return hr;
}
STDMETHODIMP Group::InitNew(IStorage *pStg) {
	METHTRACE(Group,InitNew);
	if(m_storage)
		ERR(E_UNEXPECTED);
	return takeStorage(pStg);
}
STDMETHODIMP Group::Load(IStorage *pStg) {
	METHTRACE(Group,Load);
	if(m_storage)
		ERR(E_UNEXPECTED);
	m_storage = pStg;
	RUN(pStg->OpenStream(g_itemsName,0,STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,&m_itemStream));
	RUN(pStg->OpenStream(g_indexName,0,STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,&m_indexStream));
	// ensure that we are the same implementation of groups as the storage
	// (perhaps this is counterproductive?  what about compatible file formats?
	// well, if you, Future Self or anyone else, gotta problem wid it, change it!)
	CLSID clsid;
	RUN(m_indexStream->Read(&clsid,sizeof(CLSID),0));
	if(!InlineIsEqualGUID(clsid,CLSID_Group))
		ERR(E_FAIL);
	RUN(m_indexStream->Read(&m_anonID,sizeof(DWORD),0));
	DWORD n;
	RUN(m_indexStream->Read(&n,sizeof(DWORD),0));
	while(n--) {
		CComBSTR bstr;
		RUN(bstr.ReadFromStream(m_indexStream));
		_ASSERTE(bstr.m_str); // something has gone wrong with persistence
		Item &item = m_data[bstr]; // (creates item)
		bstr.Detach(); // so CComBSTR doesn't destroy
		RUN(ReadClassStm(m_indexStream,&item.clsid));
		RUN(m_indexStream->Read(&item.method,sizeof(enum persistMethod),0));
		item.p = 0;
		RUN(m_indexStream->Read(&item.pos,sizeof(ULONGLONG),0));
		RUN(m_indexStream->Read(&item.size,sizeof(ULONGLONG),0));
		RUN(m_indexStream->Read(&item.count,sizeof(ULONG),0));
		RUN(m_indexStream->Read(&item.element,sizeof(bool),0));
	}
	return S_OK;
}
STDMETHODIMP Group::Save(IStorage *pStgSave,BOOL fSameAsLoad) {
	METHTRACE(Group,Save);
	HRESULT hr;
	LARGE_INTEGER zero;
	zero.QuadPart = 0;
	CComPtr<IStream> indexStream,itemStream,nameStream;
	if(fSameAsLoad) {
		RUN((indexStream = m_indexStream)->Seek(zero,STREAM_SEEK_SET,0));
		// have to keep the old item stream around to copy from
		RUN(pStgSave->CreateStream(g_itemsTempName,STGM_CREATE|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,&itemStream));
	}
	else {
		RUN(pStgSave->CreateStream(g_itemsName,STGM_CREATE|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,&itemStream));
		RUN(pStgSave->CreateStream(g_indexName,STGM_CREATE|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,&indexStream));
	}
	RUN(indexStream->Write(&GetObjectCLSID(),sizeof(CLSID),0));
	RUN(indexStream->Write(&m_anonID,sizeof(DWORD),0));

	// size of group
	DWORD n = m_data.size();
	RUN(indexStream->Write(&n,sizeof(DWORD),0));

	// save all
	CComBSTR bstr;
	for(itemMap::iterator i = m_data.begin();i!=m_data.end();i++) {
		itemMap::value_type &p = *i;
		Item &item = p.second;
		ULONGLONG pos,size;
		RUN(itemStream->Seek(zero,STREAM_SEEK_CUR,&castULarge(pos)));
		if(item.p) { // loaded
			switch(item.method) {
			case persist:
				RUN(WriteClassStm(itemStream,item.clsid));
				break;
			case persistStream:
			case persistStreamInit: { // (typecast)
				CComPtr<IStream> savestm;
				if(!item.element)
					savestm = itemStream;
				else {
					if(fSameAsLoad) {
						if(!item.stream)
							RUN(open(m_storage,i->first,&item.stream));
						savestm = item.stream;
					}
					else
						RUN(open(pStgSave,i->first,&savestm));
				}
				RUN(OleSaveToStream(item.pStream,savestm));
				break;
			}
			case persistStorage: {
				if(fSameAsLoad) {
					if(!item.storage)
						RUN(open(m_storage,i->first,&item.storage));
					RUN(OleSave(item.pStorage,item.storage,item.isMainStorage));
				}
				else {
					CComPtr<IStorage> s;
					RUN(open(pStgSave,i->first,&s));
					RUN(OleSave(item.pStorage,s,false));
				}
				break;
			}
			default:
				ERR(E_UNEXPECTED);
			}
		}
		else  // passive
			switch(item.method) {
			case persist:
			case persistStream:
			case persistStreamInit:
				_ASSERTE(item.pos!=-1 && item.size!=-1);
				RUN(m_itemStream->Seek(castLarge(item.pos),STREAM_SEEK_SET,0));
				RUN(m_itemStream->CopyTo(itemStream,castULarge(item.size),0,0));
				break;
			case persistStorage:
				if(!fSameAsLoad) {
					if(!item.storage)
						RUN(m_storage->OpenStorage(i->first,0,STGM_TRANSACTED|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,&item.storage));
					CComPtr<IStorage> dest;
					RUN(pStgSave->CreateStorage(i->first,STGM_TRANSACTED|STGM_CREATE|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,0,&dest));
					RUN(item.storage->CopyTo(0,0,0,dest));
				}
			}
		ULONGLONG end;
		RUN(itemStream->Seek(zero,STREAM_SEEK_CUR,&castULarge(end)));
		size = end-pos;
		// write index data
		bstr.Attach(i->first);
		RUN(bstr.WriteToStream(indexStream));
		bstr.Detach();
		RUN(WriteClassStm(indexStream,item.clsid));
		RUN(indexStream->Write(&item.method,sizeof(enum persistMethod),0));
		RUN(indexStream->Write(&pos,sizeof(ULONGLONG),0));
		RUN(indexStream->Write(&size,sizeof(ULONGLONG),0));
		RUN(indexStream->Write(&item.count,sizeof(ULONG),0));
		RUN(indexStream->Write(&item.element,sizeof(bool),0));
		if(fSameAsLoad) {
			item.pos = pos;
			item.size = size;
		}
	}
	if(fSameAsLoad) {
		// replace old item stream with new
		m_itemStream.Release();
		itemStream.Release();
		hr = pStgSave->DestroyElement(g_itemsName); // this might not exist (because no data has been written?)
		RUN(pStgSave->RenameElement(g_itemsTempName,g_itemsName));
		RUN(m_storage->OpenStream(g_itemsName,0,STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,&m_itemStream));
	}
	m_noScribble = true;
	return S_OK;
}
STDMETHODIMP Group::SaveCompleted(IStorage *pStgNew) {
	METHTRACE(Group,SaveCompleted);
	// Group never "scribbles" but its children do, so take the new
	// storage if any and then propogate the event
	/*
	// hack: "replace with same" case was causing problems in DataObject (why?)
	if(pStgNew==m_storage)
		return S_FALSE;
	*/
	// this is how it's *supposed* to work
	if(pStgNew==m_storage)
		pStgNew = 0;
	m_noScribble = false;
	if(pStgNew) {
		if(m_storage)
			RUN(HandsOffStorage());
		RUN(takeStorage(pStgNew));
	}
	return storageEvent(pStgNew?se_saveCompletedNew:se_saveCompleted);
}
STDMETHODIMP Group::HandsOffStorage(void) {
	METHTRACE(Group,HandsOffStorage);
	if(!m_storage)
		return S_OK;
	// recurse first
	RUN(storageEvent(se_handsOff));
	// then forget all stg pointers
	for(itemMap::iterator i = m_data.begin();i!=m_data.end();i++)
		i->second.storage.Release();
	m_indexStream.Release();
	m_itemStream.Release();
	m_storage.Release();
	return S_OK;
}
