/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "ItemSetRenderer.h"
#include "getSites.h"
#include "Metafile.h"
#include "RenderPolicy.h"

ItemSetRenderer::ItemSetRenderer() : m_descriptor(0),m_gingerbreadMan(0),
		m_haveMousePt(false),m_overrideBounds(false) {
	CONSTRUCTORTRACE(ItemSetRenderer);
}
ItemSetRenderer::~ItemSetRenderer() {
	DESTRUCTORTRACE(ItemSetRenderer);
	if(m_descriptor)
		freeDesc();
	m_advice.Release();
}
void ItemSetRenderer::changed() {
	if(m_advice)
		m_advice->SendOnDataChange(this,0,0);
	freeDesc();
}
CComPtr<IItemSet> ItemSetRenderer::expandDataSet() {
	// expand the data set to get labels and other annotations
	CComPtr<IItemSet> fullData;
	RUN(CREATE(fullData,CLSID_ItemSet));
	RUN(CComQIPtr<IGroupItem>(fullData)->Connect(m_srcGroup));
	RUN(fullData->Assign(m_isetData));
	VS *vs = getSites(m_srcGroup,m_isetData);
	if(!vs)
		return 0;
	for(VS::iterator is = vs->begin(); is!=vs->end(); ++is)
		RUN(fireSiteEvent(*is,fireOneParam<IMCCExpandSelection,IItemSet*>(&IMCCExpandSelection::Expand,fullData)));
	delete vs;
	return fullData;
}
HRESULT freeExpandedData(IItemSet *data) {
	SRUN(data->Clear());
	return S_OK;
}

// boundary needed for most operations
HRESULT ItemSetRenderer::calcBoundary(IItemSet *data,RECTL *ret) {
	CComPtr<IMCCTransform> t;
	RUN(mc()->get_Transform(&t));
	RECTL minmax = {0,0,0,0};
	bool mminit = false;
	VS *sites = getSites(m_srcGroup,data);
	if(!sites)
		ERR(E_PENDING);
	for(VS::iterator i = sites->begin();i!=sites->end();++i) {
		long intent;
		RUN((*i)->get_Intent(&intent));
		if(intent<MCCS_SHOWN)
			continue;
		RECTL rcl;
		if(FAILED((*i)->get_VirtRect(&rcl))) {
			RECT rc;
			RUN((*i)->get_Rect(&rc));
			RUN(t->W2CR(&rc,&rcl));
		}
		if(!mminit) {
			minmax = rcl;
			mminit = true;
		}
		else
			minmax = minmax|rcl;
	}
	delete sites;
	*ret = minmax;
	return S_OK;
}

struct zval {
	int layerZ,siteZ;
	zval(int layerZ,int siteZ) : layerZ(layerZ),siteZ(siteZ) {}
	bool operator<(zval zv) const {
		return layerZ>zv.layerZ || siteZ>zv.siteZ; // greater Z is behind; we want furthest first
	}
};
typedef map<zval,IMCCSite*> orderEm;

HRESULT ItemSetRenderer::getMetafile(RECTL bounds,IItemSet *data,DWORD tymed,STGMEDIUM *stgm) {
	METHTRACE(ItemSetRenderer,getMetafile);
	SIZEL extentHI = siz(bounds),
		extentRender = RenderPolicy::inSize(extentHI);
	RECTL rcHI = {0,0,extentHI.cx,extentHI.cy},
		rcRender = {0,0,extentRender.cx,extentRender.cy};
	HDC hdc = startMetafile(tymed,extentHI,extentRender,reinterpret_cast<RECT*>(&rcHI));
	if(tymed & TYMED_ENHMF)  // flip over rect for HIMETRIC
		swap(rcRender.top,rcRender.bottom);
    CComPtr<IMontageContainer> MC = mc();
	if(wid(bounds) && hei(bounds)) {
		VS *sites = getSites(m_srcGroup,data);
		if(!sites)
			ERR(E_PENDING);
		// sites in VS are basically in random order (ordered by moniker).  so z-order them.
		orderEm zor;
		for(VS::iterator i = sites->begin();i!=sites->end();++i) {
			long lz,sz;
			CComPtr<IMCCLayer> l;
			RUN((*i)->get_Layer(&l));
			RUN(l->get_Z(&lz));
			RUN((*i)->get_Z(&sz));
			zor[zval(lz,sz)] = *i;
		}
		CComPtr<IMCCTransform> t;
		RUN(MC->get_Transform(&t));
		RUN(t->Viewing(&rcRender,&bounds));
		RUN(fireEvent(MC,fireThreeParams<IMontageContainerEvents,HDC,LPCRECTL,LPCRECTL>(&IMontageContainerEvents::Viewing,hdc,&rcRender,&bounds)));
		for(orderEm::iterator si = zor.begin(); si!=zor.end(); ++si) {
			RECTL rcl;
			if(SUCCEEDED(si->second->get_VirtRect(&rcl))) {
				RECTL rc2;
				RUN(t->C2VR(&rcl,reinterpret_cast<RECT*>(&rc2)));
				CComPtr<IViewObject> vo;
				RUN(si->second->GetControl(riidret(&vo)));
				RUN(vo->Draw(DVASPECT_CONTENT,-1,0,0,0,hdc,&rc2,&rcRender,0,0));
			}
		}
		RUN(t->Viewing(0,0));
		RUN(fireEvent(MC,fireThreeParams<IMontageContainerEvents,HDC,LPCRECTL,LPCRECTL>(&IMontageContainerEvents::Viewing,0,0,0)));
		delete sites;
	}
	RUN(storeMetafile(hdc,tymed,extentHI,stgm));
	return S_OK;
}
HRESULT ItemSetRenderer::calcDescriptor(RECTL bounds) {
	METHTRACE(ItemSetRenderer,calcDescriptor);
	if(m_descriptor)
		freeDesc();
	size_t s = sizeof(OBJECTDESCRIPTOR)+16;
	m_descriptor = GlobalAlloc(GHND,s);
	OBJECTDESCRIPTOR *pOD = (OBJECTDESCRIPTOR*)GlobalLock(m_descriptor);

	// calculate OBJECTDESCRIPTOR
	pOD->cbSize = s;
	pOD->clsid = m_clsid;
	pOD->dwDrawAspect = DVASPECT_CONTENT;

	pOD->sizel = siz(bounds);

	// descriptor point is upside-down HIMETRIC
	if(!m_haveMousePt)
		pOD->pointl.x = pOD->pointl.y = 0;
	else {
		pOD->pointl.x = m_mousePt.x - bounds.left;
		pOD->pointl.y = bounds.top - m_mousePt.y;
	}

	// pOD->dwStatus (& leave the last two fields of OBJECTDESCRIPTOR empty)
	RUN(OleRegGetMiscStatus(m_clsid,DVASPECT_CONTENT,&pOD->dwStatus));
	// names
	pOD->dwFullUserTypeName = sizeof(OBJECTDESCRIPTOR);
	pOD->dwSrcOfCopy = sizeof(OBJECTDESCRIPTOR)+8;
	WCHAR *dest = (WCHAR*)(&pOD->dwSrcOfCopy+1),
		*mon = L"mon";
	wcscpy(dest,mon);
	wcscpy(dest+4,mon);
	GlobalUnlock(m_descriptor);
	return S_OK;
}

HRESULT ItemSetRenderer::getStorage(RECTL bounds,IStorage *stg) {
	METHTRACE(ItemSetRenderer,getStorage);
	SIZEL extent = siz(bounds);

	// standard binding stuff
	CComPtr<IBindCtx> bctx;
	RUN(CreateBindCtx(0,&bctx));

    CComPtr<IMontageContainer> MC = mc();
	// get transform
	CComPtr<IMCCTransform> t;
	RUN(MC->get_Transform(&t));

	// mark as Montage doc.
	RUN(WriteClassStg(stg,m_clsid));

	// create & init same group implementation as source
	CLSID clsidMainGroup;
	RUN(CComQIPtr<IPersist>(m_srcGroup)->GetClassID(&clsidMainGroup));
	CComPtr<IGroup> group;
	RUN(CREATE(group,clsidMainGroup));
	RUN(CComQIPtr<IPersistStorage>(group)->InitNew(stg));

	// drop an empty selection in there because otherwise the current will get dragged in.
	ItemMoniker mkSelection(L"Selection");
	CComPtr<DIMCCSiteSet> noselect;
	RUN(CREATE(noselect,CLSID_MCCSiteSet));
	RUN(group->Paste(bctx,GP_SrcIsObj,noselect,mkSelection,IID_NULL,0,0));
	// initialize gets called below ***

	// paste m_isetData as the Data set...
	// because items (such as MontagePersistor) may point at the old Data set, we have to
	// paste a blank set first, then copy m_isetData, then move the monikers from the copy to dataSet
	ItemMoniker mkDataSet(L"Data");
	CComPtr<DIMCCSiteSet> dataSet;
	RUN(CREATE(dataSet,CLSID_MCCSiteSet));
	RUN(group->Paste(bctx,GP_SrcIsObj,dataSet,mkDataSet,IID_NULL,0,0)); // initialized below ***
	CComPtr<IItemSet> s2;
	RUN(CComQIPtr<IGroupItem>(m_isetData)->Copy(bctx,group,0,riidret(&s2)));
	RUN(CComQIPtr<IItemSet>(dataSet)->Assign(s2));
	RUN(CComQIPtr<IGroupItem>(s2)->Remove());

	// montage's persistor has presumably already been pasted but we force here
	// - allows render({}) to mean, "copy the configuration"
	// - gives us access to the InitialExtent property
	// note this means an extra stgref on Montage, which shouldn't cause any trouble.
	CComPtr<IMoniker> mkMP,mkMP2;
	RUN(MC->get_RelMoniker(&mkMP));
	CComPtr<IMontagePersistor> mp;
	RUN(group->Paste(bctx,GP_SrcIsCont,m_srcGroup,mkMP,riidret(&mp),&mkMP2));
	RUN(mp->put_InitialExtent(extent));

	// finish initializing selection and data set ***
	CComPtr<IMoniker> mkS,mkM2;
	RUN(CREATE(mkS,CLSID_GetSubjectMoniker));
	RUN(mkMP2->ComposeWith(mkS,false,&mkM2));
	RUN(noselect->Initialize(mkM2));
	RUN(dataSet->Initialize(mkM2));

	// now go through copied site persistors and relocate them
	RECT windOffset;
	RUN(t->C2WR(&bounds,&windOffset));
	CComPtr<IMoniker> mkG2;
	RUN(FoolishCreatePointerMoniker(group,&mkG2));
	CComPtr<IEnumMoniker> em;
	RUN(s2->Enum(&em));
	for(mk_iterator i(em);i!=mk_iterator();++i) {
		CComPtr<IMCCSitePersistor> sp;
		if(SUCCEEDED((*i)->BindToObject(bctx,mkG2,riidret(&sp)))) {
			MCCSITEPERSISTORDATA spd;
			RUN(sp->get_Data(&spd));
			switch(spd.pos.postype) {
			case MCC_POS_CANVASRECT:
				spd.pos.rcl -= size(bounds.left,bounds.bottom);
				break;
			case MCC_POS_WINDOWRECT:
				spd.pos.rc -= size(windOffset.left,windOffset.top);
				break;
			default:
				ERR(E_UNEXPECTED);
			}
			RUN(sp->put_Data(&spd));
		}
	}
	// disconnect all so there aren't any backrefs on the group
	RUN(group->Unload(0));
	// not necessary for std. group implementation,because Unload calls Save
	//RUN(CComQIPtr<IPersistStorage>(group)->Save(stg,true));

	// check these to make sure the group gets destroyed.
	// if not, the data object won't work (because the group holds onto its storage)
	group.Release();
	mkG2.Release();
	return S_OK;
}
// IMCCItemSetRenderer
STDMETHODIMP ItemSetRenderer::Initialize(IGroup *srcGroup, IItemSet *isetData,REFCLSID rcls,IMontageContainer *mc, IMCCDataObject *mdo) {
	METHTRACE(ItemSetRenderer,Initialize);
	m_srcGroup = srcGroup;
	m_isetData = isetData;
	RUN(AtlAdvise(m_isetData,(IItemSetEvents*)this,IID_IItemSetEvents,&m_gingerbreadMan));
	// the worst HACK ever!  there seems to be no other way to eliminate the circular reference between me
	// and random item set owned by whoever except to wait for Montage shutdown event
	DWORD mallomar;
	RUN(AtlAdvise(mc,(IMontageContainerEvents*)this,IID_IMontageContainerEvents,&mallomar));
	m_clsid = rcls;
    RUN(mc->get_RefObject(&m_mref));
	if(m_buffer = mdo) {
		RUN(mdo->SetDelayedData(g_clipformats.cc,this));
		RUN(mdo->SetDelayedData(g_clipformats.embedsource,this));
		RUN(mdo->SetDelayedData(g_clipformats.objectdescriptor,this));
		RUN(mdo->SetDelayedData(CF_ENHMETAFILE,this));
		RUN(mdo->SetDelayedData(CF_METAFILEPICT,this));
		// put them in order of descriptiveness
		RUN(mdo->Prioritize(g_clipformats.cc,200)); // first
		RUN(mdo->Prioritize(g_clipformats.embedsource,100));	// second
		RUN(mdo->Prioritize(CF_ENHMETAFILE,50)); // good
		RUN(mdo->Prioritize(CF_METAFILEPICT,49)); // okay
		RUN(mdo->Prioritize(g_clipformats.objectdescriptor,25));	// "next-to-last"
	}
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::Stop() {
    METHTRACE(ItemSetRenderer,Stop);
	if(m_buffer) {
		RUN(m_buffer->SetDelayedData(g_clipformats.cc,0));
		RUN(m_buffer->SetDelayedData(g_clipformats.embedsource,0));
		RUN(m_buffer->SetDelayedData(g_clipformats.objectdescriptor,0));
		RUN(m_buffer->SetDelayedData(CF_ENHMETAFILE,0));
		RUN(m_buffer->SetDelayedData(CF_METAFILEPICT,0));
		m_buffer.Release();
	}
	if(m_isetData) {
		LOOK(AtlUnadvise(m_isetData,IID_IItemSetEvents,m_gingerbreadMan));
		m_isetData.Release();
	}
	m_advice.Release();
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::put_MousePoint(POINT val) {
	METHTRACE(ItemSetRenderer,put_MousePoint);

	// get transform
	CComPtr<IMCCTransform> t;
	RUN(mc()->get_Transform(&t));

	RUN(t->W2CP(val,&m_mousePt));
	m_haveMousePt = true;
	freeDesc();
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::SetBoundary(RECTL *val) {
    METHTRACE(ItemSetRenderer,put_UpperLeft);
	if(m_overrideBounds = val!=0)
		m_override = *val;
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::get_Bounds(RECTL *ret) {
	CComPtr<IItemSet> fullData = expandDataSet();
	if(!fullData)
		ERR(E_PENDING);
	HRESULT hr = calcBoundary(fullData,ret);
	RUN(freeExpandedData(fullData));
	return hr;
}
// DIMCCItemSetRenderer
STDMETHODIMP ItemSetRenderer::Initialize(DIGroup *srcGroup,DIItemSet *isetData,BSTR rcls,DIMontageContainer *mc,
	IUnknown *buffer) {
    METHTRACE(ItemSetRenderer,Initialize);
	CComPtr<IGroup> g2;
	if(srcGroup)
		RUNRET(srcGroup->QueryInterface(riidret(&g2)));
	CComPtr<IItemSet> isd2;
	if(isetData)
		RUNRET(isetData->QueryInterface(riidret(&isd2)));
	CLSID clsid;
	RUNRET(CLSIDFromString(rcls,&clsid));
	CComPtr<IMontageContainer> mc2;
	if(mc)
		RUNRET(mc->QueryInterface(riidret(&mc2)));
	CComPtr<IMCCDataObject> do2;
	if(buffer)
		RUNRET(buffer->QueryInterface(riidret(&do2)));
	return Initialize(g2,isd2,clsid,mc2,do2);
}
STDMETHODIMP ItemSetRenderer::SetMousePoint(long x,long y) {
    METHTRACE(ItemSetRenderer,SetMousePoint);
	POINT pt = {x,y};
	return put_MousePoint(pt);
}
// IItemSetEvents
STDMETHODIMP ItemSetRenderer::Added(IMoniker *mk) {
	changed();
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::Removed(IMoniker *mk) {
	changed();
	return S_OK;
}
// IMontageContainerEvents
STDMETHODIMP ItemSetRenderer::Shutdown() { // bad bad hack
    METHTRACE(ItemSetRenderer,Shutdown);
	return Stop();
}
STDMETHODIMP ItemSetRenderer::Resize(SIZE s) {
    METHTRACE(ItemSetRenderer,Resize);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::NewExtent(SIZEL) {
    METHTRACE(ItemSetRenderer,NewExtent);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::Viewing(HDC hdc, LPCRECTL lpcrOut, LPCRECTL lpcrCanv) {
    METHTRACE(ItemSetRenderer,Viewing);
	return S_FALSE;
}
STDMETHODIMP ItemSetRenderer::GroupOp(BOOL) {
    METHTRACE(ItemSetRenderer,GroupOp);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::NewTransform(IMCCTransform*) {
    METHTRACE(ItemSetRenderer,NewTransform);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::SiteDeleted(IMCCSite*) {
    METHTRACE(ItemSetRenderer,SiteDeleted);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::LayerDeleted(IMCCLayer*) {
    METHTRACE(ItemSetRenderer,LayerDeleted);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::Shifting(long oldState,long newState) {
    METHTRACE(ItemSetRenderer,Shifting);
	return S_FALSE;
}
// IDataObject
HRESULT ItemSetRenderer::getOD(HGLOBAL glob) {
	METHTRACE(ItemSetRenderer,getOD);
	OBJECTDESCRIPTOR *pOD = (OBJECTDESCRIPTOR*)GlobalLock(m_descriptor);
	if(GlobalSize(glob)<pOD->cbSize) {
		GlobalUnlock(m_descriptor);
		ERR(STG_E_MEDIUMFULL);
	}
	void *dest = GlobalLock(glob);
	memcpy(dest,pOD,pOD->cbSize);
	GlobalUnlock(glob);
	GlobalUnlock(m_descriptor);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::GetData(FORMATETC *pformatetcIn,STGMEDIUM *pmedium) {
	METHTRACE(ItemSetRenderer,GetData);
	RUNRET(QueryGetData(pformatetcIn));
	// bounds must be calculated for all formats
	CComPtr<IItemSet> fullData = expandDataSet();
	if(!fullData)
		ERR(E_PENDING);
	RECTL bounds;
	HRESULT hr;
	if(m_overrideBounds)
		bounds = m_override;
	else if(FAILED(hr = calcBoundary(fullData,&bounds))) {
		RUN(freeExpandedData(fullData));
		return hr;
	}
	hr = E_FAIL;
	if(pformatetcIn->cfFormat==g_clipformats.objectdescriptor) {
		RUN(calcDescriptor(bounds));
		pmedium->tymed = TYMED_HGLOBAL;
		OBJECTDESCRIPTOR *pOD = (OBJECTDESCRIPTOR*)GlobalLock(m_descriptor);
		pmedium->hGlobal = GlobalAlloc(GHND|GMEM_SHARE,pOD->cbSize);
		GlobalUnlock(m_descriptor);
		RUN(getOD(pmedium->hGlobal));
		hr = S_OK;
	}
	else if(pformatetcIn->cfFormat==CF_METAFILEPICT || pformatetcIn->cfFormat==CF_ENHMETAFILE) {
		RUN(getMetafile(bounds,fullData,pformatetcIn->tymed,pmedium));
		hr = S_OK;
	}
	else if(pformatetcIn->tymed&TYMED_ISTORAGE) {
		pmedium->tymed = TYMED_ISTORAGE;
		CComPtr<ILockBytes> bytes;
		RUN(CreateILockBytesOnHGlobal(0,true,&bytes));
		RUN(StgCreateDocfileOnILockBytes(bytes,STGM_READWRITE|STGM_SHARE_EXCLUSIVE|STGM_CREATE,0,&pmedium->pstg));
		RUN(getStorage(bounds,pmedium->pstg));
		hr = S_OK;
	}
	RUN(freeExpandedData(fullData));
	return LOOK(hr);
}
STDMETHODIMP ItemSetRenderer::GetDataHere(FORMATETC *pformatetc,STGMEDIUM *pmedium) {
	METHTRACE(ItemSetRenderer,GetDataHere);
	RUN(QueryGetData(pformatetc));
	CComPtr<IItemSet> fullData = expandDataSet();
	if(!fullData)
		ERR(E_PENDING);
	RECTL bounds;
	if(m_overrideBounds)
		bounds = m_override;
	else RUNRET(calcBoundary(fullData,&bounds));
	HRESULT hr = E_FAIL;
	if(pformatetc->cfFormat==g_clipformats.objectdescriptor) {
		RUN(calcDescriptor(bounds));
		RUN(getOD(pmedium->hGlobal));
		hr = S_OK;
	}
	else if(pformatetc->tymed&TYMED_ISTORAGE) {
		RUN(getStorage(bounds,pmedium->pstg));
		hr = S_OK;
	}
	LOOK(hr);
	RUN(freeExpandedData(fullData));
	return LOOK(hr);
}
STDMETHODIMP ItemSetRenderer::QueryGetData(FORMATETC *pformatetc) {
	METHTRACE(ItemSetRenderer,QueryGetData);
	if(pformatetc->cfFormat!=g_clipformats.cc &&
		pformatetc->cfFormat!=g_clipformats.embedsource &&
		pformatetc->cfFormat!=CF_METAFILEPICT &&
		pformatetc->cfFormat!=CF_ENHMETAFILE &&
		pformatetc->cfFormat!=g_clipformats.objectdescriptor)
		ERR(DV_E_FORMATETC);
	if(pformatetc->dwAspect!=DVASPECT_CONTENT)
		ERR(DV_E_DVASPECT);
	if(pformatetc->lindex!=-1)
		ERR(DV_E_LINDEX);
	if(pformatetc->cfFormat==g_clipformats.objectdescriptor) {
		if(!(pformatetc->tymed&TYMED_HGLOBAL))
			ERR(DV_E_TYMED);
	}
	else if(pformatetc->cfFormat==CF_METAFILEPICT) {
		if(!(pformatetc->tymed&(TYMED_MFPICT|TYMED_HGLOBAL)))
			ERR(DV_E_TYMED);
	}
	else if(pformatetc->cfFormat==CF_ENHMETAFILE) {
		if(!(pformatetc->tymed&TYMED_ENHMF))
			ERR(DV_E_TYMED);
	}
	else if(!(pformatetc->tymed&TYMED_ISTORAGE))
		ERR(DV_E_TYMED);
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::GetCanonicalFormatEtc(FORMATETC *pformatetcIn,FORMATETC *pformatetcOut) {
	METHTRACE(ItemSetRenderer,GetCanonicalFormatEtc);
	RUN(QueryGetData(pformatetcIn));
	*pformatetcOut = *pformatetcIn;
	pformatetcOut->ptd = 0;
	return DATA_S_SAMEFORMATETC;
}
STDMETHODIMP ItemSetRenderer::SetData(FORMATETC *pformatetc,STGMEDIUM *pmedium,BOOL fRelease) {
	METHTRACE(ItemSetRenderer,SetData);
	ERR(E_NOTIMPL);
}
STDMETHODIMP ItemSetRenderer::EnumFormatEtc(DWORD dwDirection,IEnumFORMATETC **ppenumFormatEtc) {
	METHTRACE(ItemSetRenderer,EnumFormatEtc);
	if(dwDirection!=DATADIR_GET)
		ERR(E_NOTIMPL);
	FORMATETC *fet = new FORMATETC[5];
	fet->ptd = 0;
	fet->dwAspect = DVASPECT_CONTENT;
	fet->lindex = -1;
	fet[4] = fet[3] = fet[2] = fet[1] = fet[0];
	fet[0].cfFormat = g_clipformats.cc;
	fet[1].cfFormat = g_clipformats.embedsource;
	fet[0].tymed = fet[1].tymed = TYMED_ISTORAGE;
	fet[2].cfFormat = CF_ENHMETAFILE;
	fet[2].tymed = TYMED_ENHMF;
	fet[3].cfFormat = CF_METAFILEPICT;
	fet[3].tymed = TYMED_MFPICT;
	fet[4].cfFormat = g_clipformats.objectdescriptor;
	fet[4].tymed = TYMED_HGLOBAL;
	fetcenum *fe = new fetcenum;
	fe->Init(fet,fet+5,0,AtlFlagTakeOwnership);
	(*ppenumFormatEtc = fe)->AddRef();
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::DAdvise(FORMATETC *pformatetc,DWORD advf,IAdviseSink *pAdvSink,
     DWORD *pdwConnection) {
	METHTRACE(ItemSetRenderer,DAdvise);
	if(!m_advice)
		RUN(CreateDataAdviseHolder(&m_advice));
	return m_advice->Advise(this,pformatetc,advf,pAdvSink,pdwConnection);
}
STDMETHODIMP ItemSetRenderer::DUnadvise(DWORD dwConnection) {
	METHTRACE(ItemSetRenderer,DUnadvise);
	if(!m_advice)
		ERR(E_FAIL);
	return m_advice->Unadvise(dwConnection);
}
STDMETHODIMP ItemSetRenderer::EnumDAdvise(IEnumSTATDATA **ppenumAdvise) {
	METHTRACE(ItemSetRenderer,EnumDAdvise);
	if(!m_advice) {
		*ppenumAdvise = 0;
		return S_OK;
	}
	return m_advice->EnumAdvise(ppenumAdvise);
}
// INeedSomePointers
STDMETHODIMP ItemSetRenderer::TakePointers(IEnumUnknown *os,IEnumMoniker *ms) {
    METHTRACE(ItemSetRenderer,TakePointers);
	if(m_mref)
		return S_FALSE;
	// data set, container
	CComPtr<IUnknown> punk;
	RUN(os->Next(1,&punk,0));
	CComQIPtr<IItemSet> isetData = punk;
	if(!isetData)
		ERR(E_UNEXPECTED);
	punk.Release();
	RUN(os->Next(1,&punk,0));
	CComQIPtr<IMontageContainer> mc = punk;
	if(!mc)
		ERR(E_UNEXPECTED);
	punk.Release();
	RUN(os->Next(1,&punk,0));
	CComQIPtr<IMCCDataObject> mdo = punk; // may be 0
	CComPtr<IGroup> g;
	RUN(mc->GetMainGroup(riidret(&g)));
	return Initialize(g,isetData,m_clsid,mc,mdo);
}
// IPersistStream
STDMETHODIMP ItemSetRenderer::GetClassID(CLSID *pClassID) {
    METHTRACE(ItemSetRenderer,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::IsDirty(void) {
    METHTRACE(ItemSetRenderer,IsDirty);
	return S_FALSE;
}
STDMETHODIMP ItemSetRenderer::Load(IStream *pStm) {
    METHTRACE(ItemSetRenderer,Load);
	RUN(ReadClassStm(pStm,&m_clsid));
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::Save(IStream *pStm,BOOL fClearDirty) {
    METHTRACE(ItemSetRenderer,Save);
	RUN(WriteClassStm(pStm,m_clsid));
	return S_OK;
}
STDMETHODIMP ItemSetRenderer::GetSizeMax(ULARGE_INTEGER *pcbSize) {
    METHTRACE(ItemSetRenderer,GetSizeMax);
	ERR(E_NOTIMPL);
}
