/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "enums.h"
#include "Container.h"
#include "Collection.h"

Layer::Layer(Container *pCont) :
	m_cont(pCont),m_ct(0),m_amTransformed(false), m_sugarWafer(0) {
	CONSTRUCTORTRACE(Layer);
}
Layer::~Layer() {
	DESTRUCTORTRACE(Layer);
	for(sitelist::iterator i = m_sites.begin();i!=m_sites.end();i++) {
		(*i)->m_pLayer = 0;
		(*i)->Release();
	}
}
HRESULT Layer::construct() {
	METHTRACE(Layer,construct);
	return CREATEA(m_points,CLSID_MCCAllPointsContainer,(IMCCLayer*)this);
}
STDMETHODIMP_(ULONG) Layer::AddRef() {
	METHTRACE(Layer,AddRef);
	return ++m_ct;
}
STDMETHODIMP_(ULONG) Layer::Release() {
	METHTRACE(Layer,Release);
	ULONG ret = --m_ct;
	if(!ret)
		delete this;
	return ret;
}
STDMETHODIMP Layer::QueryInterface(REFIID riid,void **ppRet) {
	METHTRACE(Layer,QueryInterface);
	if(InlineIsEqualGUID(riid,IID_IMCCLayer)||InlineIsEqualUnknown(riid)) {
		*ppRet = (IMCCLayer *)this;
		AddRef();
		return S_OK;
	}
	else if(InlineIsEqualGUID(riid,IID_IPropertyNotifySink)) {
		*ppRet = (IPropertyNotifySink*)this;
		AddRef();
		return S_OK;
	}
	else if(riid==IID_IConnectionPointContainer || riid==IID_IMCCAllPointsContainer)
		return m_points->QueryInterface(riid,ppRet);
	else if(riid==IID_DIMCCLayer || riid==IID_IDispatch) {
		*ppRet = (DIMCCLayer*)this;
		AddRef();
		return S_OK;
	}
	*ppRet = 0;
	ERR(E_NOINTERFACE);
}
STDMETHODIMP Layer::GetContainer(REFIID riid, void **ret) {
	METHTRACE(Layer,GetContainer);
	return ((IMontageContainer*)m_cont)->QueryInterface(riid,ret);
}
STDMETHODIMP Layer::get_RelMoniker(IMoniker **ret) {
	METHTRACE(Layer,get_RelMoniker);
	m_mkPersistor.CopyTo(ret);
	return S_OK;
}
STDMETHODIMP Layer::put_RelMoniker(IMoniker *val) {
	METHTRACE(Layer,put_RelMoniker);
	m_mkPersistor = val;
	return S_OK;
}
STDMETHODIMP Layer::get_OrderMk(IMoniker **ret) {
	METHTRACE(Layer,get_OrderMk);
	m_mkOrder.CopyTo(ret);
	return S_OK;
}
STDMETHODIMP Layer::put_OrderMk(IMoniker *val) {
	METHTRACE(Layer,put_OrderMk);
	m_mkOrder = val;
	return S_OK;
}

STDMETHODIMP Layer::put_IsTransformed(VARIANT_BOOL val) {
	METHTRACE(Layer,put_IsTransformed);
	if(m_cont->m_transform) {
		if(val && !m_amTransformed) {
			for(sitelist::iterator is = m_sites.begin();is!=m_sites.end();is++)
				if((*is)->m_bRectValid && (*is)->m_state>MCCS_HIDDEN)
					RUN(m_cont->m_transform->AddRect(&(*is)->m_virtRect));
		}
		else if(!val && m_amTransformed) {
			for(sitelist::iterator is = m_sites.begin();is!=m_sites.end();is++)
				if((*is)->m_bRectValid && (*is)->m_state>MCCS_HIDDEN)
					RUN(m_cont->m_transform->RemoveRect(&(*is)->m_virtRect));
		}
	}
	m_amTransformed = val?true:false;
	return S_OK;
}
STDMETHODIMP Layer::get_IsTransformed(VARIANT_BOOL *ret) {
	METHTRACE(Layer,get_IsTransformed);
	*ret = m_amTransformed?-1:0;
	return S_OK;
}
STDMETHODIMP Layer::put_DefaultAmbiance(IDispatch *style) {
	METHTRACE(Layer,put_DefaultAmbiance);
	if(m_defaultStyle && m_sugarWafer) {
		RUN(AtlUnadvise(m_defaultStyle,IID_IPropertyNotifySink,m_sugarWafer));
		m_sugarWafer = 0;
	}
	m_defaultStyle = style;
	if(style)
		AtlAdvise(style,(IPropertyNotifySink*)this,IID_IPropertyNotifySink,&m_sugarWafer);
	RUN(OnChanged(DISPID_UNKNOWN));
	return S_OK;
}
STDMETHODIMP Layer::get_DefaultAmbiance(IDispatch **ret) {
	METHTRACE(Layer,get_DefaultAmbiance);
	m_defaultStyle.CopyTo(ret);
	return S_OK;
}
STDMETHODIMP Layer::get_Z(long *ret) {
	METHTRACE(Layer,get_Z);
	long depth(0);
	for(layerlist::iterator i = m_cont->m_layers.begin(); i != m_cont->m_layers.end(); i++)
		if(*i == this)
			break;
		else depth++;
	if(i==m_cont->m_layers.end())
		ERR(E_FAIL);
	else {
		*ret = depth;
		return S_OK;
	}
}
STDMETHODIMP Layer::PutBefore(IMCCLayer *other) {
	METHTRACE(Layer,PutBefore);
	return m_cont->beforeLayer(this,(Layer*)other);
}
STDMETHODIMP Layer::PutBehind(IMCCLayer *other) {
	METHTRACE(Layer,PutBehind);
	return m_cont->afterLayer(this,(Layer*)other);
}
STDMETHODIMP Layer::Delete() {
	METHTRACE(Layer,Delete);\
	if(!m_cont)
		return S_FALSE;
	// bug: there isn't yet real control over unloading things vs. deleting them.
	// I would think if you deleted a Layer the sites should be unloaded, not
	// erased.  But there's no way to do that right now (only ~CtlContainer/followCCstate
	// can delete without repercussions in the group).
	if(m_sites.size())
		ERR(E_PENDING);
	// try both
	m_cont->m_unlayers.remove((Layer*)this);
	m_cont->m_layers.remove((Layer*)this);
	if(m_defaultStyle)
		RUN(put_DefaultAmbiance(0));
	m_cont = 0;
	DWORD c = Release();
	return S_OK;
}


STDMETHODIMP Layer::FindControl(POINT pt, IMCCSite **ppFound) {
	METHTRACE(Layer,FindControl);
	sitelist::iterator itest = find_if(m_sites.begin(),m_sites.end(),hitpoint(pt,MCCS_SHOWN));
	if(itest==m_sites.end()) {
		if(ppFound)
			*ppFound = NULL;
		return S_FALSE;
	}
	else {
		if(ppFound)
			(*ppFound = *itest)->AddRef();
		return S_OK;
	}
}
struct hitrectset {
	vector<IMCCSite *> m_vec;
	hitrect hr;
	hitrectset(hitrect hr) : hr(hr) {}
	void operator ()(site *s) {
		if(hr(s))
			m_vec.push_back(s);
	}
};
STDMETHODIMP Layer::FindControls(LPCRECT lpcr, IEnumMCCSites **ret) {
	METHTRACE(Layer,FindControls);
	hitrectset result = for_each(m_sites.begin(),m_sites.end(),hitrectset(hitrect(*lpcr,MCCS_SHOWN)));
	siteenum *se = new siteenum;
	se->Init(&*result.m_vec.begin(),&*result.m_vec.end(),0,AtlFlagCopy);
	(*ret = se)->AddRef();
	return S_OK;
}

STDMETHODIMP Layer::EnumSites(IEnumMCCSites **ret) {
	METHTRACE(Layer,EnumSites);
	IMCCSite **sites = new IMCCSite *[m_sites.size()],
		**si = sites;
	for(sitelist::iterator i = m_sites.begin();i!=m_sites.end();i++)
		(*si++ = *i)->AddRef();
	siteenum *se = new siteenum;
	se->Init(sites,sites+m_sites.size(),0,AtlFlagTakeOwnership);
	(*ret = se)->AddRef();
	return S_OK;
}
// DIMCCLayer
STDMETHODIMP Layer::get_Container(DIMontageContainer **ret) {
    METHTRACE(Layer,get_Container);
	return GetContainer(riidret(ret));
}
STDMETHODIMP Layer::get_RelMoniker(IUnknown **ret) {
    METHTRACE(Layer,get_RelMoniker);
	return get_RelMoniker((IMoniker**)ret);
}
STDMETHODIMP Layer::put_RelMoniker(IUnknown *val) {
    METHTRACE(Layer,put_RelMoniker);
	CComPtr<IMoniker> m2;
	if(val)
		RUNRET(val->QueryInterface(riidret(&m2)));
	return put_RelMoniker(m2);
}
STDMETHODIMP Layer::get_OrderMk(IUnknown **ret) {
    METHTRACE(Layer,get_OrderMk);
	return get_OrderMk((IMoniker**)ret);
}
STDMETHODIMP Layer::put_OrderMk(IUnknown *val) {
    METHTRACE(Layer,put_OrderMk);
	CComPtr<IMoniker> m2;
	if(val)
		RUNRET(val->QueryInterface(riidret(&m2)));
	return put_OrderMk(m2);
}
STDMETHODIMP Layer::PutBefore(DIMCCLayer *other) {
    METHTRACE(Layer,PutBefore);
	CComPtr<IMCCLayer> l2;
	if(other)
		RUNRET(other->QueryInterface(riidret(&l2)));
	return PutBefore(l2);
}
STDMETHODIMP Layer::PutBehind(DIMCCLayer *other) {
    METHTRACE(Layer,PutBehind);
	CComPtr<IMCCLayer> l2;
	if(other)
		RUNRET(other->QueryInterface(riidret(&l2)));
	return PutBehind(l2);
}
STDMETHODIMP Layer::FindControl(long x, long y, DIMCCSite **ret) {
    METHTRACE(Layer,FindControl);
	POINT pt = {x,y};
	CComPtr<IMCCSite> s;
	RUN(FindControl(pt,&s));
	return s.QueryInterface(ret);
}
STDMETHODIMP Layer::FindControls(long left,long top, long right, long bottom, IDispatch **ret) {
    METHTRACE(Layer,FindControls);
	hitrectset result = for_each(m_sites.begin(),m_sites.end(),hitrectset(hitrect(rect(left,top,right,bottom),MCCS_SHOWN)));
	(*ret = new collection(make_colldata(result.m_vec)))->AddRef();
	return S_OK;
}
// Site has two IDispatch impls: choose the DIMCCSite one.
struct conv_site2dimccsite {
	IDispatch *operator()(Site *s) {
		return (IDispatchImpl<DIMCCSite>*)s;
	}
};
STDMETHODIMP Layer::Sites(IDispatch **ret) {
    METHTRACE(Layer,Sites);
	(*ret = new collection(make_colldata_conv(m_sites,conv_site2dimccsite())))->AddRef();
	return S_OK;
}

// IPropertyNotifySink
STDMETHODIMP Layer::OnChanged(DISPID id) {
	METHTRACE(Layer,OnChanged);
	for(sitelist::iterator i = m_sites.begin();i!=m_sites.end();i++) {
		site *s = *i;
		CComQIPtr<IOleControl> oc = (*i)->m_pControl;
		if(oc)
			RUN(oc->OnAmbientPropertyChange(id));
	}
	return S_OK;
}
STDMETHODIMP Layer::OnRequestEdit(DISPID id) {
	METHTRACE(Layer,OnRequestEdit);
	return S_OK;
}
