/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "RectEngine.h"
#include "borders.h"

RectDelegate::RectDelegate() : m_constrains(false), m_eng(0), m_type(normal) {
	CONSTRUCTORTRACE(RectDelegate);
}
RectDelegate::~RectDelegate() {
	DESTRUCTORTRACE(RectDelegate);
	// inner and outer will only get deleted when engine's going down; nothing to do there
	if(m_type==normal) {
		rd2 *self = (rd2*)this;
		if(m_constrains)
			RUN(m_eng->change(self,0));
		m_eng->m_data.remove(self);
		m_eng->GetUnknown()->Release();
	}
}
HRESULT RectDelegate::set(LPCRECT val) {
	METHTRACE(RectDelegate,set);
    RECT r2 = *val;
	if(val) {
		m_rect = *val;
		m_constrains = true;
	}
	else m_constrains = false;
	for(IUnknown **iu = m_vec.begin();iu!=m_vec.end();iu++)
		if(CComQIPtr<IMoveRect> mr = *iu)
			RUN(mr->Move(val));
	return S_OK;
}
// IMoveRect
STDMETHODIMP RectDelegate::Move(LPCRECT lpcr) {
	METHTRACE(RectDelegate,Move);
    RECT r2 = *lpcr;
	switch(m_type) {
	case normal:
		return m_eng->change((rd2*)this,lpcr);
	case special_inner:
		return m_eng->changeInner(lpcr);
	case special_outer:
		return m_eng->changeOuter(lpcr);
	default:
		_ASSERTE(1==0);
	}
	ERR(E_FAIL);
}
RectEngine::RectEngine() {
	CONSTRUCTORTRACE(RectEngine);
	// m_inner and m_outer have opposite refcounting from the other delegates
	(m_inner = new rd2(this))->AddRef();
	m_inner->m_type = RectDelegate::special_inner;
	(m_outer = new rd2(this))->AddRef();
	m_outer->m_type = RectDelegate::special_outer;
}
RectEngine::~RectEngine() {
	DESTRUCTORTRACE(RectEngine);
	// m_inner and m_outer have opposite refcounting from the other delegates
	m_inner->Release();
	m_outer->Release();
	// all other delegates must already be deleted for engine refcount to drop to zero.
}
// IRectEngine
STDMETHODIMP RectEngine::new_Delegate(IMoveRect **ret) {
	METHTRACE(RectEngine,new_Delegate);
	rd2 *rv = new rd2(this);
	m_data.push_back(rv);
	GetUnknown()->AddRef(); // layout object refs engine
	(*ret = rv)->AddRef();
	return S_OK;
}
STDMETHODIMP RectEngine::get_Outer(IMoveRect **ret) {
	METHTRACE(RectEngine,get_Outer);
	(*ret = m_outer)->AddRef();
	return S_OK;
}
STDMETHODIMP RectEngine::get_Inner(IMoveRect **ret) {
	METHTRACE(RectEngine,get_Inner);
	(*ret = m_inner)->AddRef();
	return S_OK;
}
STDMETHODIMP RectEngine::put_OuterFlexible(VARIANT_BOOL val) {
	ERR(E_NOTIMPL);
}
STDMETHODIMP RectEngine::get_OuterFlexible(VARIANT_BOOL *ret) {
	ERR(E_NOTIMPL);
}
// IMontageInclusion
STDMETHODIMP RectEngine::ContainsRect(LPCRECT lpcr) {
	METHTRACE(RectEngine,ContainsRect);
	// the rect doesn't touch our bounds if it's within the inner or
	// doesn't intersect with the outer
	if(m_inner->m_constrains)
		if(within(*lpcr,m_inner->m_rect))
			return S_FALSE;
	if(m_outer->m_constrains)
		if(intersectType(*lpcr,m_outer->m_rect)==int_none)
			return S_FALSE;
	return S_OK;
}
// IPersist
STDMETHODIMP RectEngine::GetClassID(CLSID *ret) {
	METHTRACE(RectEngine,GetClassID);
	*ret = GetObjectCLSID();
	return S_OK;
}

int guessborder(const RECT &r, const RECT &inner) {
	// see which side would need to be stretched the least:
	int dl = r.right - inner.left,
		dt = r.bottom - inner.top,
		dr = inner.right - r.left,
		db = inner.bottom - r.top;
	int smallest = dl, choice = Bleft;
	if(dt<smallest) {
		smallest = dt;
		choice = Btop;
	}
	if(dr<smallest) {
		smallest = dr;
		choice = Bright;
	}
	if(db<smallest) {
		smallest = db;
		choice = Bbottom;
	}
	if(smallest<0) // already in border
		return 0;
	return choice;
}
// here's the layout!
// there's a strict hierarchy: the inner must get out of the way of
// the data must get out of the way of the outer.

// get the inner rectangle out of the way in the easiest way
void adjustInner(const RECT &r,RECT &inner) {
	switch(guessborder(r,inner)) {
	case Bleft:
		inner.left = r.right+1;
		break;
	case Btop:
		inner.top = r.bottom+1;
		break;
	case Bright:
		inner.right = r.left-1;
		break;
	case Bbottom:
		inner.bottom = r.top-1;
		break;
	}
}
// gets a normal rectangle out of the way of the outer rect
bool adjustNormal(const RECT &outer,RECT &r) {
	int isect = intersectType(r,outer);
	if(isect==int_inside) // within: okay
		return false;
	int dx = 0,dy = 0;
	switch(isect&int_x) {
	case int_ll: // off left side
		dx = outer.left - r.left;
		break;
	case int_rr: // off right side
		dx = outer.right - r.right;
		break;
	case int_rl: // okay, inside
		break;
	case int_lr:
		r.left = outer.left; // rect's just too big.
		r.right = outer.right;
	}
	switch(isect&int_y) {
	case int_tt: // off top side
		dy = outer.top - r.top;
		break;
	case int_bb:
		dy = outer.bottom - r.bottom;
		break;
	case int_tb:
		break;
	case int_bt:
		r.top = outer.top;
		r.bottom = outer.bottom;
	}
	r +=  size(dx,dy);
	return true;
}
HRESULT RectEngine::change(rd2 *o,LPCRECT r) {
	METHTRACE(RectEngine,change);
	if(r) {
		RECT rc = *r;
		if(m_outer->m_constrains)
			adjustNormal(m_outer->m_rect,rc);
		if(m_inner->m_constrains) {
			int isect = intersectType(rc,m_inner->m_rect);
			if(isect!=int_none) {
				RECT move = m_inner->m_rect;
				adjustInner(rc,move);
				RUN(m_inner->set(&move));
			}
		}
		RUN(o->set(&rc));
	}
	else
		RUN(o->set(0));
	return spring();
}
HRESULT RectEngine::changeInner(LPCRECT r) {
	METHTRACE(RectEngine,changeInner);
	if(!r)
		ERR(E_FAIL); // ?
	// correct for what's actually there
	RECT rc = *r;
	for(rdlist::iterator i = m_data.begin();i!=m_data.end();i++)
		if((*i)->m_constrains)
			adjustInner((*i)->m_rect,rc);
	return m_inner->set(&rc);
}
HRESULT RectEngine::changeOuter(LPCRECT r) {
	METHTRACE(RectEngine,changeOuter);
	RUN(m_outer->set(r));
	if(r)
		for(rdlist::iterator i = m_data.begin();i!=m_data.end();i++)
			if((*i)->m_constrains) {
				RECT rc = (*i)->m_rect;
				if(adjustNormal(*r,rc))
					RUN(change(*i,&rc));
			}
	return spring();
}
// this is tricky.  I think it will work to just try to move the inner to the outer.
HRESULT RectEngine::spring() {
	METHTRACE(RectEngine,spring);
	return changeInner(&m_outer->m_rect);
}