/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "ToolPanel.h"

STDMETHODIMP MontageSink::Resize(SIZE size) {
    METHTRACE(MontageSink,Resize);
	RECT r = {0,0,size.cx,size.cy};
	if(m_panel->m_backs[0])
		m_panel->m_backs[0]->put_Rect(&r);
	if(m_panel->m_engine) {
		// typical just-averted infinite loop: this event may be caused by the engine resizing the
		// boundary -> OnPosRectChange -> here... but if so, then this is the same rect, so the engine
		// will ignore this change!
		CComPtr<IMoveRect> boundary;
		RUN(m_panel->m_engine->get_Outer(&boundary));
		RUN(boundary->Move(&r));
	}
	return 0;
}
STDMETHODIMP MontageSink::Shifting(long oldState,long newState) {
    METHTRACE(MontageSink,Shifting);
    // when ToolPanel is handling the tool border, its client will call Start
    // with the IOleInPlaceUIWindow of the border.  However, when it's handling
    // a normal embedded object full of tools, it must call Start itself
    // when the Montage part goes active.
	if(!m_panel->m_dest && oldState<MCCS_ACTIVE && newState>=MCCS_ACTIVE)
		RUN(m_panel->Start(0));
	return S_OK;
}
STDMETHODIMP RectSink::Move(LPCRECT lpcr) {
    METHTRACE(RectSink,Move);
	return m_panel->rectCallin(id,lpcr);
}
ToolPanel::ToolPanel() {
	CONSTRUCTORTRACE(ToolPanel);
	for(int i=0;i<4;i++)
		m_backs[i] = 0;
}
ToolPanel::~ToolPanel() {
	DESTRUCTORTRACE(ToolPanel);
	for(int i=0;i<4;i++)
		if(m_backs[i])
			m_backs[i]->Release();
}
HRESULT ToolPanel::FinalConstruct() {
	METHTRACE(ToolPanel,FinalConstruct);
	RUN(CREATEA(m_montage,CLSID_MontageControlContainer,(IMCCToolPanel*)this));
	m_sink = new CComObject<MontageSink>();
	m_sink->m_panel = this;
	RUN(AtlAdvise(m_montage,m_sink,IID_IMontageContainerEvents,&m_sink->m_nilla));
	return S_OK;
}
HRESULT ToolPanel::FinalRelease() {
	METHTRACE(ToolPanel,FinalRelease);
	m_montage.Release();
	m_sink->m_panel = 0;
	m_sink.Release();
	return S_OK;
}

enum sides {frame_left,frame_top,frame_right,frame_bottom};
// IMCCToolPanel
STDMETHODIMP ToolPanel::Start(IOleInPlaceUIWindow *dest) {
	METHTRACE(ToolPanel,Start);
	CComQIPtr<IMontageContainer> montage = m_montage;

	m_dest = dest;
	RECT rcBorder;
	if(dest) { // do tricks to host montage in the uiwindow
		// get border rect & check if we can put up tools
		if(dest->GetBorder(&rcBorder)==INPLACE_E_NOTOOLSPACE)
			return S_FALSE;
	}
    else { // act as an activex control
		rcBorder.left = rcBorder.top = 0;
		RUN(montage->get_Size((SIZE*)&rcBorder.right));
    }


	if(!m_engine)
		RUN(getEngine());

	// set up engine as a MontageInclusion sink
	CComPtr<IConnectionPoint> cpMI;
	RUN(CComQIPtr<IMCCAllPointsContainer>(m_montage)->EnsurePoint(IID_IMontageInclusion,&cpMI));
	RUN(cpMI->Advise(m_engine,&m_oatmealInclusion));

	CComPtr<IMoveRect> mrInner;
	if(m_dest) {
		// listen to inner rect
		RUN(m_engine->get_Inner(&mrInner));
		m_borderSink = new CComObject<RectSink>();
		m_borderSink->m_panel = this;
		m_borderSink->id = rect_sink_border;
		RUN(AtlAdvise(mrInner,m_borderSink,IID_IMoveRect,&m_mintBrusselsInner));

		// activate aggregated Montage into frame window
		HWND hWnd;
		RUN(dest->GetWindow(&hWnd));
		if(!hWnd)
			ERR(E_UNEXPECTED);
		RUN(montage->ActivateAs(hWnd));
	}
	else {
		CComPtr<IMoveRect> mrOuter;
		RUN(m_engine->get_Outer(&mrOuter));
		m_boundarySink = new CComObject<RectSink>();
		m_boundarySink->m_panel = this;
		m_boundarySink->id = rect_sink_boundary;
		RUN(AtlAdvise(mrOuter,m_boundarySink,IID_IMoveRect,&m_circularBakedSweet));
	}
	// create toolbacks (zero-size until the inner rect callback)
	if(!m_backLayer) {
		RUN(montage->new_Layer(&m_backLayer));
		RUN(m_backLayer->PutBefore(0));
	}

	// HACK: why should the panel know about the data reader just to tell toolbacks?
	CComPtr<IMoniker> mkGroup;
	ItemMoniker mkReader(L"Panel Data Reader");
	CComPtr<IGroup> g;
	RUN(montage->GetMainGroup(riidret(&g)));
	RUN(FoolishCreatePointerMoniker(g,&mkGroup));
	CComPtr<IBindCtx> bctx;
	RUN(CreateBindCtx(0,&bctx));
	CComPtr<IMCCDataObjectReader> reader;
	RUN(mkReader->BindToObject(bctx,mkGroup,riidret(&reader)));

	CComPtr<IClassFactory> backFactory;
	RUN(CoGetClassObject(CLSID_MCCToolBack,CLSCTX_ALL,0,riidret(&backFactory)));
	int N = m_dest?4:1;
	for(int i = 0;i<N;i++) {
		if(!m_backs[i]) {
			CComPtr<IMCCToolBack> tb; // empty interface (add something to it?)
			RUN(backFactory->CreateInstance(0,riidret(&tb)));
			RUN(CComQIPtr<INeedOnePointer>(tb)->TakePointer(reader,0));
			RUN(montage->new_Site(&m_backs[i]));
			RUN(m_backs[i]->put_Control(tb));
			RUN(m_backs[i]->put_Layer(m_backLayer));
		}
		RECT rcPos = rcBorder;
		if(dest)
			switch(i) {
			case frame_left:
				rcPos.right = rcPos.left;
				break;
			case frame_top:
				rcPos.bottom = rcPos.top;
				break;
			case frame_right:
				rcPos.left = rcPos.right;
				break;
			case frame_bottom:
				rcPos.top = rcPos.bottom;
				break;
			}
		RUN(m_backs[i]->put_Rect(&rcPos));
		RUN(m_backs[i]->put_Intent(MCCS_ACTIVE)); // actually will wait until toolbar MCC gets ActivateAs
	}

	// initialize outer rect
	RUN(Resize(&rcBorder));

	// start with an inner border with no space; the engine will correct me
	//RECT rc2 = {ctr(rcBorder).x,ctr(rcBorder).y,ctr(rcBorder).x+1,ctr(rcBorder).y+1};
	return mrInner?mrInner->Move(&rcBorder):S_OK;
}
STDMETHODIMP ToolPanel::Stop() {
	METHTRACE(ToolPanel,Stop);
	if(!m_engine)
		return S_FALSE; // as did Start(), as no tool space was given us
	CComPtr<IMoveRect> mrInner;
	RUN(m_engine->get_Inner(&mrInner));
	RUN(AtlUnadvise(mrInner,IID_IMoveRect,m_mintBrusselsInner));
	RUN(AtlUnadvise(m_montage,IID_IMontageInclusion,m_oatmealInclusion));
	RUN(CComQIPtr<IMontageContainer>(m_montage)->ActivateAs(0));
	m_dest.Release();
	return S_OK;
}
STDMETHODIMP ToolPanel::Resize(LPCRECT lpcr) {
	METHTRACE(ToolPanel,Resize);
	if(lpcr->top>lpcr->bottom)
		// powerpoint is sending me {0,0,0,-16}-like rects when minimized (??)
		return S_OK;
	// shouldn't get Resize w/o Start - is something activated wrong?
	_ASSERTE(m_engine);
	CComPtr<IMoveRect> mrOuter;
	RUN(m_engine->get_Outer(&mrOuter));
	return mrOuter->Move(lpcr);
}
// IMoveRect (from inner rect)
HRESULT ToolPanel::rectCallin(int id,LPCRECT lpcr) {
	METHTRACE(ToolPanel,Move);
	if(id==rect_sink_border) {
		RECT rcOuter;
		RUN(m_dest->GetBorder(&rcOuter));

		// tell frame wassup
		BORDERWIDTHS bw = {lpcr->left - rcOuter.left,lpcr->top - rcOuter.top, rcOuter.right - lpcr->right, rcOuter.bottom - lpcr->bottom};
		RUN(m_dest->SetBorderSpace(&bw));

		// size backs to match
		RECT rcPos = rcOuter;
		rcPos.right = lpcr->left;
		RUN(m_backs[frame_left]->put_Rect(&rcPos));
		rcPos.right = rcOuter.right;
		rcPos.bottom = lpcr->top;
		RUN(m_backs[frame_top]->put_Rect(&rcPos));
		rcPos.bottom = rcOuter.bottom;
		rcPos.left = lpcr->right;
		RUN(m_backs[frame_right]->put_Rect(&rcPos));
		rcPos.left = rcOuter.left;
		rcPos.top = lpcr->bottom;
		RUN(m_backs[frame_bottom]->put_Rect(&rcPos));
	}
	else if(id==rect_sink_boundary) {
		// hmmph all this to get rect?
		CComPtr<IOleClientSite> ocs;
		RUN(CComQIPtr<IOleObject>(m_montage)->GetClientSite(&ocs));
        _ASSERTE(ocs);
        if(ocs) {
		    CComPtr<IOleInPlaceFrame> fr; CComPtr<IOleInPlaceUIWindow> ui;
		    RECT pos,clip; OLEINPLACEFRAMEINFO info;
		    info.cb = sizeof OLEINPLACEFRAMEINFO;
		    CComQIPtr<IOleInPlaceSite> ips(ocs);
		    RUN(ips->GetWindowContext(&fr,&ui,&pos,&clip,&info));
		    RECT r2 = {pos.left+lpcr->left,pos.top+lpcr->top,pos.left+lpcr->right,pos.top+lpcr->bottom};
		    if(r2!=pos)
			    RUN(ips->OnPosRectChange(&r2));
        }
	}
	return S_OK;
}
HRESULT ToolPanel::getEngine() {
	METHTRACE(ToolPanel,getEngine);
	// this is a hack.  everything about containers holding the names of
	// contained objects is a little weird (consider the montage persistor)
	CComPtr<IOleItemContainer> oic;
	RUN(CComQIPtr<IMontageContainer>(m_montage)->GetMainGroup(riidret(&oic)));
	CComPtr<IBindCtx> bctx;
	RUN(CreateBindCtx(0,&bctx));
	RUN(oic->GetObject(L"Toolbar Engine",BINDSPEED_INDEFINITE,bctx,riidret(&m_engine)));
	return S_OK;
}

/*
// IPersistStorage
STDMETHODIMP ToolPanel::GetClassID(CLSID *pClassID) {
	METHTRACE(ToolPanel,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP ToolPanel::IsDirty(void) {
	METHTRACE(ToolPanel,IsDirty);
	return CComQIPtr<IPersistStorage>(m_montage)->IsDirty();
}
STDMETHODIMP ToolPanel::InitNew(IStorage *pStg) {
	METHTRACE(ToolPanel,InitNew);
	HRESULT hr;
	RUN(CComQIPtr<IMontageContainer>(m_montage)->put_Bootstrapper(bp));
	RUN(CComQIPtr<IPersistStorage>(m_montage)->InitNew(pStg));
	return S_OK;
}
STDMETHODIMP ToolPanel::Load(IStorage *pStg) {
	METHTRACE(ToolPanel,Load);
	HRESULT hr;
	RUN(CComQIPtr<IPersistStorage>(m_montage)->Load(pStg));
	RUN(getEngine());
	return S_OK;
}
STDMETHODIMP ToolPanel::Save(IStorage *pStgSave,BOOL fSameAsLoad) {
	METHTRACE(ToolPanel,Save);
	return CComQIPtr<IPersistStorage>(m_montage)->Save(pStgSave,fSameAsLoad);
}
STDMETHODIMP ToolPanel::SaveCompleted(IStorage *pStgNew) {
	METHTRACE(ToolPanel,SaveCompleted);
	return CComQIPtr<IPersistStorage>(m_montage)->SaveCompleted(pStgNew);
}
STDMETHODIMP ToolPanel::HandsOffStorage(void) {
	METHTRACE(ToolPanel,HandsOffStorage);
	return CComQIPtr<IPersistStorage>(m_montage)->HandsOffStorage();
}
*/