/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "..\shared\rect.h"

#include <vector>
#include <set>
#include <hash_map>
using namespace std;
using namespace stdext;

struct BPState;
struct BPGame;
struct BPObject;
typedef vector<BPObject> BPBoard;
struct BPObject {
	typedef bool (*callOut_f)(BPBoard::iterator o, BPState *s, BPGame *g); // returns whether it dirtied board

	int priority; // when to play
	bool valid; // whether to play
	callOut_f callOut; // function to call out on each play
	RECT pos; // position of the box

	bool operator >(const BPObject &other) const {
		return priority>other.priority;
	}
};
enum BPPriorities {BPP_Bounds = 500, BPP_Border = 400, BPP_Normal = 300};
struct BPState : BPBoard {
	BPState(BPState *last) : BPBoard(last->begin(),last->end()), cost(last->cost) {}
	BPState(BPBoard *start) : BPBoard(start->begin(),start->end()), cost(0) {
		// clients rely on their ordering of the board, so let's assume they've
		// got it sorted already.
		//make_heap(begin(),end(),greater<BPObject>());
		//sort_heap(begin(),end(),greater<BPObject>());
	}
	int cost;
};
struct lessCost {
	inline bool operator ()(BPState *p1, BPState *p2) {
		return p1->cost < p2->cost;
	}
};
// this set is emulating a priority queue, but the game also keeps pointers into this set
// from the already-tried set.
// top is here; pop and push also touch the already-tried set, so they're in BPGame
typedef multiset<BPState*,lessCost> BPStateSet;
struct BPQueue :  BPStateSet {
	BPState *top() {
		return *begin();
	}
};
// silly ordering
struct BPTraits {
	static const size_t bucket_size = 4,
		min_buckets = 8;
	size_t operator()(BPState *s) const {
		size_t ret = 0;
		for(BPState::iterator i = s->begin();i!=s->end();++i)
			if(i->valid)
				ret += i->pos.left + (i->pos.top<<8) + (i->pos.right<<16) + (i->pos.bottom<<24);
		return ret;
	}
	bool operator()(BPState *s1, BPState *s2) const {
		if(s1->size()!=s2->size())
			return s1->size()<s2->size();
		int cmp = memcmp(&s1->front(),&s2->front(),sizeof(BPObject)*s1->size());
		return cmp<0;
	}
};
// this maps from board states to positions in the (pseudo) priority queue
typedef hash_map<BPState *,BPQueue::iterator,BPTraits> BPTriedSet;
struct BPGame : BPQueue {
	// a list of already-seen states.  "Owns" these states (when pushed, client forgets; when removed
	// with clear() state gets deleted).
	BPTriedSet tried;
	bool flexibleBoundary;
	~BPGame() {
		clear();
	}
	void clear() {
		for(BPTriedSet::iterator i = tried.begin();i!=tried.end();) {
			// weird order because iterator relies on value
			BPState *del = i->first;
			++i;
			delete del;
		}
		tried.clear();
		BPQueue::clear();
	}
	void pop() {
		// mark tried-set to show state is no longer in queue
		BPState *s = top();
		BPTriedSet::iterator i = tried.find(s);
		_ASSERTE(i!=tried.end());
		i->second = end();
		// pop from queue
		erase(begin());
	}
	// takes ownership of the passed-in state, either deleting if seen before, or holding a pointer
	void push(BPState *s) {
		BPTriedSet::iterator i = tried.find(s);
		if(i!=tried.end()) {
			if(s->cost<i->first->cost) {
				// keep old board but update cost
				// re-insert (move or add) board in queue
				if(i->second!=end())
					erase(i->second);
				i->first->cost = s->cost;
				i->second = insert(i->first);
			}
			delete s;
		}
		else
			tried.insert(make_pair<BPState * const,BPQueue::iterator>(s,insert(s))); // silly const!
	}
	typedef pair<bool,BPState*> stepResult;
	stepResult BPGame::step();
	BPBoard *run(BPBoard *start); // sorts board and steps til done.  Returns copy of result.
};

int moveRight(const RECT &sub,RECT &ob);
int moveDown(const RECT &sub,RECT &ob);
int moveLeft(const RECT &sub,RECT &ob);
int moveUp(const RECT &sub,RECT &ob);

// implementations of BPObject::callOut and constructors for each type of object
// normal boxes
bool box_play(BPBoard::iterator o,BPState *s, BPGame *g);
inline BPObject make_box(int priority,RECT pos, bool valid = true) {
	BPObject ret;
	ret.priority = priority;
	ret.valid = valid;
	ret.callOut = box_play;
	ret.pos = pos;
	return ret;
}
// the outer boundary
bool bounds_play(BPBoard::iterator o,BPState *s, BPGame *g);
inline BPObject make_bounds(RECT pos, bool valid = true) {
	BPObject ret;
	ret.priority = BPP_Bounds;
	ret.valid = valid;
	ret.callOut = bounds_play;
	ret.pos = pos;
	return ret;
}
// the inner boundary
bool border_play(BPBoard::iterator o,BPState *s,BPGame *g);
inline BPObject make_border(RECT pos, bool valid = true) {
	BPObject ret;
	ret.priority = BPP_Border;
	ret.valid = valid;
	ret.callOut = border_play;
	ret.pos = pos;
	return ret;
}