/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "Engine.h"

#include "allocDGShapes.h"
#include "shapeConv.h"

#include "EdgeLabel.h"
#include "RWAttrs.h"

#include <common/colors.h>
#include "olecolor2rgb.h"

/*
#include "persistShape.h"
*/

Edge::Edge() : m_engine(0),m_edge(0),m_destroying(false),m_incrInserted(false) {
	CONSTRUCTORTRACE(Edge);
	//m_cache.geom.headClipped = m_cache.geom.tailClipped = false;
}
Edge::~Edge() {
	DESTRUCTORTRACE(Edge);
    m_destroying = true;
	HRESULT hr;
	if(m_edge) {
		if(m_engine->current.find(m_edge)) // gracefully remove self in case rest of graph is still there
			hr = Delete();
		freeEdge();
	}
	m_engine->GetUnknown()->Release();
}
void Edge::freeEdge() { // so node can call in when it causes deletion of m_edge
	m_engine->forget(m_edge);
	m_engine->destroy(m_edge);
	m_edge = 0;
}
void Edge::sh_modify(StrAttrs attrs) {
	SRUN(mod(stringsIn(m_engine->m_transform,m_edge,attrs,false)));
}
void Edge::sh_rename(DString name) {
	m_engine->rename(m_edge,name);
}

HRESULT Edge::mod(Update upd) {
    METHTRACE(Edge,mod);
    if(m_engine->current.find(m_edge)) {
		m_engine->Q.ModEdge(m_edge,upd);
		RUNRET(m_engine->go());
	}
	return S_OK;
}
void Edge::show() {
	METHTRACE(Edge,show);
	RUN(fireEvent(GetUnknown(),fireVoid<IDGEdgeShadow>(&IDGEdgeShadow::Insert)));
}
void Edge::hide(Engine *engine,Layout::Edge *edge) {
	METHTRACE(Edge,hide);
    if(!m_destroying)
	    RUN(fireEvent(GetUnknown(),fireVoid<IDGEdgeShadow>(&IDGEdgeShadow::Delete)));
}
/////////////////////////////////////////////////////////////////////////////
// IDGEdge
STDMETHODIMP Edge::get_Graph(IDGGraph **ppOut) {
	IMETHTRACE(Edge,IDGEdge,get_Graph);
	if(!ppOut)
		ERR(E_POINTER);
	(*ppOut = m_engine)->AddRef();
	return S_OK;
}
// the two put_....Node methods here are the only ones
// that take object parameters in the engines.  Montage uses
// them conspicuously.
STDMETHODIMP Edge::get_TailNode(IDGNode **ppNode) {
	METHTRACE(DGEdge,get_TailNode);
	if(!ppNode)
		ERR(E_POINTER);
	Node *n = m_engine->m_nodeMap[m_edge->tail];
	(*ppNode = n)->AddRef();
	return S_OK;
}
STDMETHODIMP Edge::get_HeadNode(IDGNode **ppNode) {
	IMETHTRACE(Edge,IDGEdge,get_HeadNode);
	if(!ppNode)
		ERR(E_POINTER);
	Node *n = m_engine->m_nodeMap[m_edge->head];
	(*ppNode = n)->AddRef();
	return S_OK;
}
STDMETHODIMP Edge::get_TailPos(PPOINTD ptpos) {
	IMETHTRACE(Edge,IDGDrawnEdge,get_TailPos);
	if(!ptpos)
		ERR(E_POINTER);
	*ptpos = DG2Win(m_engine->reorient(geom().tailPort.pos,false));
	return S_OK;
}
STDMETHODIMP Edge::put_TailPos(POINTD tpos) {
	IMETHTRACE(Edge,IDGDrawnEdge,put_TailPos);
	geom().tailPort.pos = m_engine->reorient(Win2DG(tpos),true);
	mod(DG_UPD_TAIL);
	return S_OK;
}
STDMETHODIMP Edge::get_HeadPos(PPOINTD phpos) {
	IMETHTRACE(Edge,IDGDrawnEdge,get_HeadPos);
	if(!phpos)
		ERR(E_POINTER);
	*phpos = DG2Win(m_engine->reorient(geom().headPort.pos,false));
	return S_OK;
}
STDMETHODIMP Edge::put_HeadPos(POINTD hpos) {
	IMETHTRACE(Edge,IDGDrawnEdge,put_HeadPos);
	geom().headPort.pos = m_engine->reorient(Win2DG(hpos),true);
	mod(DG_UPD_HEAD);
	return S_OK;
}
STDMETHODIMP Edge::get_Length(double *pLength) {
	IMETHTRACE(Edge,IDGDrawnEdge,get_Length);
	if(!pLength)
		ERR(E_POINTER);
	*pLength = geom().minLength;
	return S_OK;
}
STDMETHODIMP Edge::put_Length(double length) {
	IMETHTRACE(Edge,IDGDrawnEdge,put_Length);
	geom().minLength = length;
	mod(DG_UPD_LENGTH);
	return S_OK;
}
STDMETHODIMP Edge::get_Constraint(VARIANT_BOOL *pConstraint) {
	IMETHTRACE(Edge,IDGDrawnEdge,get_Constraint);
	if(!pConstraint)
		ERR(E_POINTER);
	*pConstraint = geom().constraint?-1:0;
	return S_OK;
}
STDMETHODIMP Edge::put_Constraint(VARIANT_BOOL constraint) {
	IMETHTRACE(Edge,IDGDrawnEdge,put_Constraint);
	geom().constraint = constraint!=VARIANT_FALSE;
	mod(DG_UPD_CONSTRAINT);
	return S_OK;
}
STDMETHODIMP Edge::SetPoints(DGSHAPEDESC *val) {
	readDGSHAPEDESC(m_engine,val,geom().pos,true);
	mod(DG_UPD_MOVE);
	return S_OK;
}
STDMETHODIMP Edge::GetPoints(DGSHAPEDESC **ret) {
	IMETHTRACE(Edge,IDGDrawnEdge,get_Points);
	Line &l = geom().pos;
	*ret = writeDGSHAPEDESC(m_engine,l,true);
	if(!*ret)
		ERR(E_PENDING);
	return S_OK;
}
STDMETHODIMP Edge::new_Label(IDGEdgeLabel **ret) {
	EdgeLabelWrap *el = new CComObject<EdgeLabelWrap>;
	el->e = this;
	el->x = labels().size();
	labels().push_back(EdgeLabel());
	labelz.push_back(el);
	el->AddRef();
	(*ret = el)->AddRef();
	return S_OK;
}
STDMETHODIMP Edge::Insert() {
	IMETHTRACE(Edge,IDGDrawnEdge,Insert);
	if(m_engine->current.find(m_edge))
		return S_FALSE;
	if(!(m_engine->current.find(m_edge->tail)||m_engine->Q.insN.find(m_edge->tail))||
		!(m_engine->current.find(m_edge->head)||m_engine->Q.insN.find(m_edge->head)))
		ERR(E_PENDING); // nodes not inserted yet
	m_engine->Q.InsEdge(m_edge);
	_ASSERTE(m_engine->Q.insE.find(m_edge));
	RUNRET(m_engine->go());
	_ASSERTE(m_engine->locks);
	return S_OK;
}
STDMETHODIMP Edge::Delete() {
	IMETHTRACE(Edge,IDGDrawnEdge,Delete);
	if(!m_engine->current.find(m_edge))
		return S_FALSE;
	m_engine->Q.DelEdge(m_edge);
	RUNRET(m_engine->go());
	return S_OK;
}
STDMETHODIMP Edge::Optimize(VARIANT_BOOL full) {
	if(full)
		geom().pos.clear();
	mod(DG_UPD_MOVE);
	return S_OK;
}
// DIDGDrawnEdge
STDMETHODIMP Edge::put_BorderColor(OLE_COLOR clr) {
    METHTRACE(Edge,put_BorderColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(m_edge).put("color",color.toString());
    return S_OK;
}
STDMETHODIMP Edge::get_BorderColor(OLE_COLOR* pclr) {
    METHTRACE(Edge,get_BorderColor);
    StrAttrs::iterator ai = gd<StrAttrs>(m_edge).find("color");
    if(ai==gd<StrAttrs>(m_edge).end()||!ai->second.length())
        *pclr = RGB(0,0,0); // default black lines
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Edge::put_ForeColor(OLE_COLOR clr) {
    METHTRACE(Edge,put_ForeColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(m_edge).put("fontcolor",color.toString());
    return S_OK;
}
STDMETHODIMP Edge::get_ForeColor(OLE_COLOR* pclr) {
    METHTRACE(Edge,get_ForeColor);
    StrAttrs::iterator ai = gd<StrAttrs>(m_edge).find("fontcolor");
    if(ai==gd<StrAttrs>(m_edge).end())
        *pclr = RGB(0,0,0); // default black text
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Edge::put_FillColor(OLE_COLOR clr) {
    METHTRACE(Edge,put_FillColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(m_edge).put("fillcolor",color.toString());
    return S_OK;
}
STDMETHODIMP Edge::get_FillColor(OLE_COLOR* pclr) {
    METHTRACE(Edge,get_FillColor);
    StrAttrs::iterator ai = gd<StrAttrs>(m_edge).find("fillcolor");
    if(ai==gd<StrAttrs>(m_edge).end()||!ai->second.length())
        return get_BorderColor(pclr); // default to line color
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Edge::get_Engine(DIDGLayout **ppEngine) {
    METHTRACE(Edge,get_Engine);
	CComPtr<IDGGraph> e;
	RUN(get_Graph(&e));
	return e.QueryInterface(ppEngine);
}
STDMETHODIMP Edge::get_TailNode(DIDGDrawnNode **pNode) {
    METHTRACE(Edge,get_TailNode);
	CComPtr<IDGNode> n;
	RUN(get_TailNode(&n));
	return n.QueryInterface(pNode);
}
STDMETHODIMP Edge::get_HeadNode(DIDGDrawnNode **pNode) {
    METHTRACE(Edge,get_HeadNode);
	CComPtr<IDGNode> n;
	RUN(get_HeadNode(&n));
	return n.QueryInterface(pNode);
}
STDMETHODIMP Edge::get_TailX(double *ret) {
    METHTRACE(Edge,get_TailX);
	POINTD ptd;
	RUN(get_TailPos(&ptd));
	*ret = ptd.x;
	return S_OK;
}
STDMETHODIMP Edge::get_TailY(double *ret) {
    METHTRACE(Edge,get_TailY);
	POINTD ptd;
	RUN(get_TailPos(&ptd));
	*ret = ptd.x;
	return S_OK;
}
STDMETHODIMP Edge::SetTailPos(double x, double y) {
    METHTRACE(Edge,SetTailPos);
	POINTD ptd = {x,y};
	return put_TailPos(ptd);
}
STDMETHODIMP Edge::get_HeadX(double *ret) {
    METHTRACE(Edge,get_HeadX);
	POINTD ptd;
	RUN(get_TailPos(&ptd));
	*ret = ptd.x;
	return S_OK;
}
STDMETHODIMP Edge::get_HeadY(double *ret) {
    METHTRACE(Edge,get_HeadY);
	POINTD ptd;
	RUN(get_TailPos(&ptd));
	*ret = ptd.x;
	return S_OK;
}
STDMETHODIMP Edge::SetHeadPos(double x,double y) {
    METHTRACE(Edge,SetHeadPos);
	POINTD ptd = {x,y};
	return put_HeadPos(ptd);
}
STDMETHODIMP Edge::get_Points(DIDGShape **ret) {
    METHTRACE(Edge,get_Points);
	DGSHAPEDESC *sd;
	RUN(GetPoints(&sd));
	CComPtr<IDGShape> s;
	RUN(CREATE(s,CLSID_DGShape));
	RUN(s->put_Shape(sd));
	freeDGShape(sd);
	RUN(s.QueryInterface(ret));
	return S_OK;
}

// IPersistStream
STDMETHODIMP Edge::GetClassID(CLSID *pClassID) {
	METHTRACE(Edge,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP Edge::IsDirty(void) {
	METHTRACE(Edge,IsDirty);
	return S_FALSE;
}
// note position is not saved, this currently because dg-win32's EdgeConnector doesn't know
// how to offset properly.
STDMETHODIMP Edge::Load(IStream *stream) {
	METHTRACE(Edge,Load);
	StrAttrs stuff;
	RUN(readAttrs(stream,stuff));
	mod(stringsIn(m_engine->m_transform,m_edge,stuff,true));
	return S_OK;
}
STDMETHODIMP Edge::Save(IStream *stream,BOOL fClearDirty) {
	METHTRACE(Edge,Save);
	RUN(writeAttrs(stream,gd<StrAttrs>(m_edge)));
	return S_OK;
}
STDMETHODIMP Edge::GetSizeMax(ULARGE_INTEGER *pcbSize) {
	METHTRACE(Edge,GetSizeMax);
	ERR(E_NOTIMPL);
}
/*
//////////////////////////////////////////
// CEnumDGEdges
extern "C" {
	#include "incr\engine.h"
}
CEnumDGEdges::CEnumDGEdges(EEdgeType typ, Agnode_t *n) {
	METHTRACE(CEnumDGEdges,CEnumDGEdges);
	int cnt = 0;
	Agedge_t *ie;
	switch(typ) {
	case EDGE_IN:
		for(ie = agfstin(n); ie; ie = agnxtin(ie)) cnt++;
		break;
	case EDGE_OUT:
		for(ie = agfstout(n); ie; ie = agnxtout(ie)) cnt++;
		break;
	case EDGE_ALL:
		for(ie = agfstedge(n); ie; ie = agnxtedge(ie,n)) cnt++;
	}
	IDGEdge **arr = new IDGEdge *[cnt],
		**ia = arr;
	switch(typ) {
	case EDGE_IN:
		for(ie = agfstin(n); ie; ie = agnxtin(ie))
			*ia++ = (Edge *)DG_edge(ie);
		break;
	case EDGE_OUT:
		for(ie = agfstout(n); ie; ie = agnxtout(ie))
			*ia++ = (Edge *)DG_edge(ie);
		break;
	case EDGE_ALL:
		for(ie = agfstedge(n); ie; ie = agnxtedge(ie,n))
			*ia++ = (Edge *)DG_edge(ie);
	}
	Init(arr,ia,NULL,AtlFlagTakeOwnership);
}
*/