/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


// Engine.cpp : Implementation of Engine

#include "stdafx.h"
#include "Engine.h"
#include "engine_ids.h" // dispatch IDs
#include "Collection.h"
#include "allocDGShapes.h"
#include "shapeConv.h"
#include "RWAttrs.h"
#include <common/colors.h>
#include "olecolor2rgb.h"
#include "readHGlobalDotfile.h"
#include "StringGraph.h"

#include <sstream>
#include <fstream>
#include <iomanip>
using namespace std;

LPCOLESTR g_szCLSID = L"CLSID",
	g_szEngName = L"EngName";
LPCOLESTR g_szEngineDesc = L"Dynagraph Layout Engines";

// The incrface parser will call into here
struct IncrWants : IncrCallbacks {
    IncrWants() {
        g_incrCallback = this;
    }
	typedef CComObject<AbstractGraph> AbsG;
    IncrLangEvents *incr_cb_create_handler(Name name,const StrAttrs &attrs) {
		// return only graphs that have been created by the user.
		// "open graph" with a new name is an error
        DinoMachine::Node *dn = g_dinoMachine.ndict[name];
		return dn?gd<DinoMachNode>(dn).handler:0;
	}
	void incr_cb_destroy_handler(IncrLangEvents *h) {
		// ignore: graph lifetimes are only affected by reference count
	}
	// These should turn into IncrClient events
    void incr_cb_fulfil_graph(DString name,StrGraph *sg) {
	}
    void incr_cb_fulfil_node(DString graphname,DString nodename,const StrAttrs &attrs) {
	}
    void incr_cb_fulfil_edge(DString graphname,DString edgename,const StrAttrs &attrs) {
	}
    void incr_cb_message(const char *msg) {
	}
};
IncrWants g_comdg_incr_callback;
Engine::Engine() :
	m_orient(DGO_DOWN),
	secret_code(17),
	DynaView(DString(),new Transform(Coord(1,1),Coord(1,1)),false)
{
	CONSTRUCTORTRACE(Engine);
	// we go ahead and register before incr_open_graph happens
    incr_set_handler(gd<Name>(&layout),this);
}

Engine::~Engine() {
	DESTRUCTORTRACE(Engine);
}
Coord Engine::reorient(Coord val,bool in,enum DG_ORIENTATION dir) {
	Coord ret;
	switch(dir) {
	case DGO_UP:
		ret.x = -val.x;
		ret.y = -val.y;
		break;
	case DGO_DOWN:
		ret.x = val.x;
		ret.y = val.y;
		break;
	case DGO_LEFT:
		if(in)
			goto right;
	left:
		ret.x = val.y;
		ret.y = -val.x;
		break;
	case DGO_RIGHT:
		if(in)
			goto left;
	right:
		ret.x = -val.y;
		ret.y = val.x;
		break;
	}
	return ret;
}
HRESULT Engine::mod(Update upd) {
	Q.GraphUpdateFlags() |= upd.flags;
	if(dgserver)
		go();
	return S_OK;
}
HRESULT Engine::go() {
    try {
        if(!maybe_go())
    		return S_FALSE; // later...
    }
    catch(DGException xep) {
        handleDGException(xep);
        ERR(E_FAIL);
    }
	catch(...) {
		BREAK();
	}
	return S_OK;
}
// View events
void Engine::IncrHappened() {
    METHTRACE(Engine,IncrHappened);
	// these events can cause changes, so clear the change queue first
	ChangeQueue Q2 = Q;
	Q.Okay(false);
	++locks;
	Layout::Node *n;
	Layout::Edge *e;
	// do callbacks
    for(Layout::node_iter ni = Q2.insN.nodes().begin(); ni!=Q2.insN.nodes().end(); ++ni) {
		if(Node *N = m_nodeMap[n = layout.find(*ni)])
			N->show();
        gd<StrAttrChanges>(*ni).clear();
    }
    for(Layout::graphedge_iter ei = Q2.insE.edges().begin(); ei!=Q2.insE.edges().end(); ++ei) {
		if(Edge *E = m_edgeMap[e = layout.find(*ei)])
			E->show();
        gd<StrAttrChanges>(*ei).clear();
    }
    for(ni = Q2.modN.nodes().begin(); ni!=Q2.modN.nodes().end(); ++ni) {
        StrAttrChanges &changes = gd<StrAttrChanges>(*ni);
        if(Node *N = m_nodeMap[n = layout.find(*ni)]) {
            Update update = igd<Update>(*ni);
            int modf = 0;
	        if(update.flags&DG_UPD_MOVE)
                modf |= DGMF_POS;
	        else if(update.flags&DG_UPD_LABEL)
                modf |= DGMF_LABELS;
	        if(update.flags&(DG_UPD_POLYDEF|DG_UPD_REGION))
                modf |= DGMF_SHAPE;
            if(changes.find("color")!=changes.end()||changes.find("fontcolor")!=changes.end()||changes.find("fillcolor")!=changes.end())
                modf |= DGMF_COLORS;
            if(!changes.empty())
                modf |= DGMF_STRINGS;
            if(modf)
                RUN(fireEvent((IDGNode*)N,fireOneParam<IDGNodeShadow,int>(IDGNodeShadow::Modify,modf)));
        }
        changes.clear();
    }
    for(ei = Q2.modE.edges().begin(); ei!=Q2.modE.edges().end(); ++ei) {
        if(Edge *E = m_edgeMap[e = layout.find(*ei)]) {
            Update update = igd<Update>(*ei);
            int modf = 0;
	        if(update.flags&DG_UPD_MOVE)
                modf |= DGMF_POS;
	        else if(update.flags&DG_UPD_LABEL)
                modf |= DGMF_LABELS;
            if(gd<StrAttrChanges>(*ei).find("color")!=gd<StrAttrChanges>(*ei).end()||
                    gd<StrAttrChanges>(*ei).find("fillcolor")!=gd<StrAttrChanges>(*ei).end())
                modf |= DGMF_COLORS;
            if(!gd<StrAttrChanges>(*ei).empty())
                modf |= DGMF_STRINGS;
            if(modf)
                RUN(fireEvent((IDGEdge*)E,fireOneParam<IDGEdgeShadow,int>(IDGEdgeShadow::Modify,modf)));
        }
        gd<StrAttrChanges>(*ei).clear();
    }
    for(ei = Q2.delE.edges().begin(); ei!=Q2.delE.edges().end();) {
		e = layout.find(*ei++);
        gd<StrAttrChanges>(e).clear(); // hide() may result in deletion, but may not...
		if(Edge *E = m_edgeMap[e])
			E->hide(this,e);
    }
    for(ni = Q2.delN.nodes().begin(); ni!=Q2.delN.nodes().end();) {
		n = layout.find(*ni++);
        gd<StrAttrChanges>(n).clear();
		if(Node *N = m_nodeMap[n])
			N->hide();
    }
    // catch any whose string attrs changed but didn't actually have any layout action
    for(ni = Q2.current->nodes().begin(); ni!=Q2.current->nodes().end(); ++ni) {
        StrAttrChanges &changes = gd<StrAttrChanges>(*ni);
        if(Node *N = m_nodeMap[n = layout.find(*ni)]) {
            if(!changes.empty())
                RUN(fireEvent((IDGNode*)N,fireOneParam<IDGNodeShadow,int>(IDGNodeShadow::Modify,DGMF_STRINGS)));
        }
        changes.clear();
    }
    for(ei = Q2.current->edges().begin(); ei!=Q2.current->edges().end(); ++ei) {
        if(Edge *E = m_edgeMap[e = layout.find(*ei)]) {
            if(!gd<StrAttrChanges>(*ei).empty())
                RUN(fireEvent((IDGEdge*)E,fireOneParam<IDGEdgeShadow,int>(IDGEdgeShadow::Modify,DGMF_STRINGS)));
        }
        gd<StrAttrChanges>(*ei).clear();
    }
    if(Q2.GraphUpdateFlags()&DG_UPD_DRAWN)
		RUN(fireEvent(GetUnknown(),fireVoid<IDGLayoutShadow>(&IDGLayoutShadow::OnLinesChanged)));
	if(Q2.GraphUpdateFlags()&DG_UPD_BOUNDS) {
		RECTD r = DG2Win(gd<GraphGeom>(&current).bounds);
		RUN(fireEvent(GetUnknown(),fireOneParam<IDGLayoutShadow,PRECTD>(&IDGLayoutShadow::OnBoundsChanged,&r)));
	}
	if(Q2.GraphUpdateFlags()&DG_UPD_CHANGERECT) {
		Bounds &cr = gd<GraphGeom>(&current).changerect;
		RECTD r = DG2Win(cr);
		RUN(fireEvent(GetUnknown(),fireOneParam<IDGLayoutShadow,PRECTD>(&IDGLayoutShadow::OnChangeRect,&r)));
	}
    if(!gd<StrAttrChanges>(&current).empty())
		RUN(fireEvent(GetUnknown(),fireVoid<IDGLayoutShadow>(&IDGLayoutShadow::OnStringsChanged)));
    gd<StrAttrChanges>(&current).clear();
	--locks;
}
void Engine::IncrNewNode(Layout::Node *n) {
    METHTRACE(Engine,IncrNewNode);
	CComPtr<IDGDrawnNode> dgn;
	RUN(wrapNode(n,&dgn));
	RUN(fireEvent(GetUnknown(),fireOneParam<IDGLayoutShadow,IDGDrawnNode*>(&IDGLayoutShadow::OnNewNode,dgn)));
}
void Engine::IncrNewEdge(Layout::Edge *e) {
    METHTRACE(Engine,IncrNewEdge);
	CComPtr<IDGDrawnEdge> dge;
	RUN(wrapEdge(e,&dge));
	RUN(fireEvent(GetUnknown(),fireOneParam<IDGLayoutShadow,IDGDrawnEdge*>(&IDGLayoutShadow::OnNewEdge,dge)));
}

// IDGLayout
STDMETHODIMP Engine::Initialize() {
	METHTRACE(Engine,Initialize);
    if(dgserver)
        m_replacementFlag = true;
    else
        m_allowOneReopen = true;
    createServer();
	RUN(go());
    return S_OK;
}
STDMETHODIMP Engine::get_Resolution(POINTD *ret) {
    METHTRACE(Engine,get_Resolution);
    *ret = pointd(gd<GraphGeom>(&current).resolution);
    return S_OK;
}
STDMETHODIMP Engine::put_Resolution(POINTD val) {
    METHTRACE(Engine,put_Resolution);
    if(assign(gd<GraphGeom>(&current).resolution,coord(val))) {
		ostringstream o;
        o << gd<GraphGeom>(&current).resolution.x << ',' << gd<GraphGeom>(&current).resolution.y;
        gd<StrAttrs2>(&current).put("resolution",o.str());
	}
    return S_OK;
}
STDMETHODIMP Engine::get_Ticks(float *ret) {
    METHTRACE(Engine,get_Ticks);
    *ret = gd<GraphGeom>(&current).ticks;
    return S_OK;
}
STDMETHODIMP Engine::put_Ticks(float val) {
    METHTRACE(Engine,put_Ticks);
    if(assign(gd<GraphGeom>(&current).ticks,val)) {
		ostringstream o;
        o << gd<GraphGeom>(&current).ticks;
        gd<StrAttrs2>(&current).put("ticks",o.str());
    }
    return S_OK;
}
STDMETHODIMP Engine::get_Separation(POINTD *ret) {
    METHTRACE(Engine,get_Separation);
    *ret = pointd(gd<GraphGeom>(&current).separation);
    return S_OK;
}
STDMETHODIMP Engine::put_Separation(POINTD val) {
    METHTRACE(Engine,put_Separation);
    if(assign(gd<GraphGeom>(&current).separation,coord(val))) {
		ostringstream o;
        o << gd<GraphGeom>(&current).separation;
        gd<StrAttrs2>(&current).put("separation",o.str());
	}
    return S_OK;
}
STDMETHODIMP Engine::get_LabelGap(POINTD *ret) {
    *ret = pointd(gd<GraphGeom>(&current).labelGap);
	return S_OK;
}
STDMETHODIMP Engine::put_LabelGap(POINTD val) {
    if(assign(gd<GraphGeom>(&current).labelGap,coord(val))) {
		ostringstream o;
        o << gd<GraphGeom>(&current).labelGap;
        gd<StrAttrs2>(&current).put("labelGap",o.str());
	}
	return S_OK;
}
STDMETHODIMP Engine::get_DefaultNodeSize(POINTD *ret) {
    *ret = pointd(gd<GraphGeom>(&current).defaultSize);
	return S_OK;
}
STDMETHODIMP Engine::put_DefaultNodeSize(POINTD val) {
    if(assign(gd<GraphGeom>(&current).defaultSize,coord(val))) {
		ostringstream o;
        o << gd<GraphGeom>(&current).defaultSize;
        gd<StrAttrs2>(&current).put("defaultSize",o.str());
	}
	return S_OK;
}
STDMETHODIMP Engine::get_Orientation(enum DG_ORIENTATION *ret) {
    METHTRACE(Engine,get_Orientation);
	*ret = m_orient;
	return S_OK;
}
STDMETHODIMP Engine::put_Orientation(enum DG_ORIENTATION val) {
    METHTRACE(Engine,put_Orientation);
	if(val==m_orient)
		return S_FALSE;
	RUN(GraphLock());
	// positions stay the same, but the shapes rotate
	for(Layout::node_iter ni = layout.nodes().begin(); ni!=layout.nodes().end(); ++ni) {
		Region &poly = gd<NodeGeom>(*ni).region;
		for(Line::iterator pi = poly.shape.begin(); pi!=poly.shape.end(); ++pi) {
			*pi = reorient(*pi,false); // back to outside world
			*pi = reorient(*pi,true,val); // to new world
		}
        poly.updateBounds();
		m_nodeMap[*ni]->mod(DG_UPD_REGION|DG_UPD_MOVE|DG_UPD_LABEL|DG_UPD_POLYDEF); // move flag just for us
	}
	m_orient = val;
	gd<Translation>(&layout).orientation = (Orientation)val;
	RUN(GraphUnlock());
	return S_OK;
}
STDMETHODIMP Engine::get_SplineLevel(enum DG_SPLINELEVEL *ret) {
    METHTRACE(Engine,get_SplineLevel);
	switch(gd<GraphGeom>(&layout).splineLevel) {
	case DG_SPLINELEVEL_VNODE:
		*ret = DGSL_VNODE;
		break;
	case DG_SPLINELEVEL_BOUNDS:
		*ret = DGSL_BOUNDS;
		break;
	case DG_SPLINELEVEL_SHORTEST:
		*ret = DGSL_SHORTEST;
		break;
	case DG_SPLINELEVEL_SPLINE:
		*ret = DGSL_SPLINE;
		break;
	default:
		ERR(E_FAIL);
	}
	return S_OK;
}
STDMETHODIMP Engine::put_SplineLevel(enum DG_SPLINELEVEL val) {
    METHTRACE(Engine,put_SplineLevel);
	SpliningLevel &level = gd<GraphGeom>(&layout).splineLevel;
	switch(val) {
	case DGSL_VNODE:
		level = DG_SPLINELEVEL_VNODE;
		break;
	case DGSL_BOUNDS:
		level = DG_SPLINELEVEL_BOUNDS;
		break;
	case DGSL_SHORTEST:
		level = DG_SPLINELEVEL_SHORTEST;
		break;
	case DGSL_SPLINE:
		level = DG_SPLINELEVEL_SPLINE;
		break;
	default:
		ERR(E_FAIL);
	}
	mod(DG_UPD_EDGESTYLE);
	return S_OK;
}
STDMETHODIMP Engine::get_Nodes(IEnumDGNodes **ppRet) {
    METHTRACE(Engine,get_Nodes);
	/*
	METHTRACE(Engine,get_Nodes);
	(*ppRet = new CEnumDGNodes(agfstnode(ilmodel(this))))->AddRef();
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}
STDMETHODIMP Engine::get_Bounds(RECTD *ret) {
	Bounds &b = gd<GraphGeom>(&current).bounds;
	if(!b.valid)
		return E_FAIL;
	*ret = DG2Win(b);
	return S_OK;
}
STDMETHODIMP Engine::get_ChangeRect(RECTD *ret) {
	Bounds &b = gd<GraphGeom>(&current).changerect;
	if(!b.valid)
		ERR(E_FAIL);
	*ret = DG2Win(b);
	return S_OK;
}
STDMETHODIMP Engine::get_Lines(DGSHAPESDESC **ret) {
    METHTRACE(Engine,get_Lines);
	Drawn &lines = gd<Drawn>(&current);
	*ret = allocDGShapes(lines.size());
	int i=0;
	for(Lines::iterator li = lines.begin(); li!=lines.end(); ++i,++li)
		(*ret)->sps[i] = writeDGSHAPEDESC(this,*li,true);
	return S_OK;
}
STDMETHODIMP Engine::GraphLock() {
	METHTRACE(Engine,Hold);
	locks++;
	return S_OK;
}
STDMETHODIMP Engine::GraphUnlock() {
	METHTRACE(Engine,Unleash);
    if(locks) {
		locks--;
		RUNRET(go());
	} // else be lenient about overzealous unleashing
	return S_OK;
}
STDMETHODIMP Engine::get_Locks(int *ret) {
    METHTRACE(Engine,get_Locks);
    *ret = locks;
    return S_OK;
}

HRESULT Engine::wrapNode(Layout::Node *n, IDGDrawnNode **ret) {
    METHTRACE(Engine,wrapNode);
	CComObject<Node> *pNode = new CComObject<Node>();
	((IDGLayout*)(pNode->m_engine = this))->AddRef(); // clients ref engine.
	pNode->m_node = n;
	m_nodeMap[n] = pNode;
	(*ret = pNode)->AddRef();
	return S_OK;
}
STDMETHODIMP Engine::new_Node(BSTR name, IDGNode **ppRet) {
	METHTRACE(Engine,new_Node);
	if(!ppRet)
		ERR(E_POINTER);
	USES_CONVERSION;
	pair<Layout::Node*,bool> nb = getNode(W2A(name),true);
	if(!nb.second)
		ERR(CTL_E_FILEALREADYOPEN);
	RUN(wrapNode(nb.first,(IDGDrawnNode**)ppRet));
	return S_OK;
}
HRESULT Engine::wrapEdge(Layout::Edge *e,IDGDrawnEdge **ret) {
    METHTRACE(Engine,wrapEdge);
	CComObject<Edge> *pEdge = new CComObject<Edge>();
	ULONG d = ((IDGLayout*)(pEdge->m_engine = this))->AddRef(); // clients ref engine.
	pEdge->m_edge = e;
	m_edgeMap[e] = pEdge;
	(*ret = pEdge)->AddRef();
	return S_OK;
}
STDMETHODIMP Engine::new_Edge(BSTR nombre, IDGNode *tail, IDGNode *head, IDGEdge ** ppRet) {
	METHTRACE(Engine,new_Edge);
	if(!ppRet)
		ERR(E_POINTER);
	// since these are the only dynamic casts I decided to forgo RTTI
	// pN = dynamic_cast<Node *>(pNode);
	Node *t = static_cast<Node *>(tail),
		*h = static_cast<Node *>(head);
	if(t->secret_code!=17 || h->secret_code!=17)
		ERR(E_INVALIDARG);
	USES_CONVERSION;
	pair<Layout::Edge*,bool> eb = getEdge(W2A(nombre),t->m_node,h->m_node,true);
	if(!eb.second) // already exists
		ERR(CTL_E_FILEALREADYEXISTS);
	RUN(wrapEdge(eb.first,(IDGDrawnEdge**)ppRet));
	return S_OK;
}
STDMETHODIMP Engine::get_Node(BSTR name, IDGNode **ret) {
    METHTRACE(Engine,get_Node);
	USES_CONVERSION;
	Layout::Node *n = getNode(W2A(name)).first;
	if(!n) {
		*ret = 0;
		ERR(E_FAIL);
	}
	(*ret = m_nodeMap[n])->AddRef();
	return S_OK;
}
STDMETHODIMP Engine::get_Edge(BSTR name, IDGEdge **ret) {
    METHTRACE(Engine,get_Edge);
	USES_CONVERSION;
	Layout::Edge *e = getEdge(W2A(name));
	if(!e) {
		*ret = 0;
		ERR(E_FAIL);
	}
	(*ret = m_edgeMap[e])->AddRef();
	return S_OK;
}
/*
STDMETHODIMP Engine::get_EdgeByNodes(BSTR tail, BSTR head, IDGEdge **ret) {
    METHTRACE(Engine,get_EdgeByNodes);
	USES_CONVERSION;
	Layout::Edge *e = getEdge(0,W2A(tail),W2A(head));
	if(!e) {
		*ret = 0;
		ERR(E_FAIL);
	}
	(*ret = m_edgeMap[e])->AddRef();
	return S_OK;
}
*/
STDMETHODIMP Engine::OptimizeAll(VARIANT_BOOL full) {
    METHTRACE(Engine,OptimizeAll);
	RUN(GraphLock());
	for(Layout::node_iter ni = current.nodes().begin(); ni!=current.nodes().end(); ++ni) {
		for(Layout::nodeedge_iter ei = (*ni)->alledges().begin(); ei!=(*ni)->alledges().end(); ++ei) {
			Layout::Edge *le = layout.find(*ei);
			Edge *e = m_edgeMap[le];
			RUN(e->Optimize(full));
		}
		Layout::Node *ln = layout.find(*ni);
		Node *n = m_nodeMap[ln];
		RUN(n->Optimize(full));
	}
	RUN(GraphUnlock());
	return S_OK;
}
STDMETHODIMP Engine::LoadDotfile(HGLOBAL data,VARIANT_BOOL merge,VARIANT_BOOL del) {
    METHTRACE(Engine,LoadDotfile);
    StrGraph *sg = readHGlobalDotfile(data);
	if(!sg)
		ERR(E_FAIL);
	incr_ev_lock();
	incr_ev_load_strgraph(sg,merge!=VARIANT_FALSE,del!=VARIANT_FALSE);
	incr_ev_unlock();
	return S_OK;
}
STDMETHODIMP Engine::SaveDotfile(HGLOBAL *ret) {
	ostringstream f;
	emitGraph(f,&layout);
    f << ends;
	string s = f.str();
	*ret = GlobalAlloc(GHND,s.length());
	void *p = GlobalLock(*ret);
	s.copy((char*)p,s.length());
	GlobalUnlock(*ret);
	return S_OK;
}
// DIDGLayout
/*
// doin' this the hard way?
struct nodeIterator {
	Agnode_t *curr;
	nodeIterator(Agnode_t *curr) : curr(curr) {}
	DIDGDrawnNode *operator *() {
		return (Node*)il_node(curr);
	}
	void operator ++() {
		curr = agnxtnode(curr);
	}
	bool operator !=(nodeIterator &other) {
		return curr!=other.curr;
	}
};
*/
STDMETHODIMP Engine::put_BackColor(OLE_COLOR clr) {
    METHTRACE(Engine,put_BackColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(&layout).put("bgcolor",color.toString());
    return S_OK;
}
STDMETHODIMP Engine::get_BackColor(OLE_COLOR* pclr) {
    METHTRACE(Engine,get_BackColor);
    StrAttrs::iterator ai = gd<StrAttrs>(&layout).find("bgcolor");
    if(ai==gd<StrAttrs>(&layout).end())
        *pclr = RGB(255,255,255); // default white
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Engine::SetResolution(double x,double y) {
    METHTRACE(Engine,SetResolution);
    POINTD res = {x,y};
    return put_Resolution(res);
}
STDMETHODIMP Engine::get_ResolutionX(double *ret) {
    METHTRACE(Engine,get_ResolutionX);
    *ret = gd<GraphGeom>(&current).resolution.x;
    return S_OK;
}
STDMETHODIMP Engine::get_ResolutionY(double *ret) {
    METHTRACE(Engine,get_ResolutionY);
    *ret = gd<GraphGeom>(&current).resolution.y;
    return S_OK;
}
STDMETHODIMP Engine::SetSeparation(double x,double y) {
    METHTRACE(Engine,SetSeparation);
    POINTD sep = {x,y};
    return put_Separation(sep);
}
STDMETHODIMP Engine::get_SeparationX(double *ret) {
    METHTRACE(Engine,get_SeparationX);
    *ret = gd<GraphGeom>(&current).separation.x;
    return S_OK;
}
STDMETHODIMP Engine::get_SeparationY(double *ret) {
    METHTRACE(Engine,get_SeparationY);
    *ret = gd<GraphGeom>(&current).separation.y;
    return S_OK;
}
STDMETHODIMP Engine::get_Nodes(IDispatch ** ppRet) {
    METHTRACE(Engine,get_Nodes);
	/*
	(*ppRet = new collection(make_colldata(nodeIterator(agfstnode(ilmodel(this))),nodeIterator(0))))->AddRef();
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}
STDMETHODIMP Engine::new_Node(BSTR name, DIDGDrawnNode ** ppRet) {
    METHTRACE(Engine,new_Node);
	CComPtr<IDGNode> n;
	RUN(new_Node(name,&n));
	return n.QueryInterface(ppRet);
}
STDMETHODIMP Engine::new_Edge(BSTR name, DIDGDrawnNode *tail, DIDGDrawnNode *head, DIDGDrawnEdge ** ppRet) {
    METHTRACE(Engine,new_Edge);
	CComPtr<IDGEdge> e;
	RUN(new_Edge(name,CComQIPtr<IDGNode>(tail),CComQIPtr<IDGNode>(head),&e));
	return e.QueryInterface(ppRet);
}

// IPersistStream
STDMETHODIMP Engine::IsDirty(void) {
	METHTRACE(Engine,IsDirty);
	return S_FALSE;
}
#define SIMPER(x) &x,sizeof x,0
STDMETHODIMP Engine::Load(IStream *stream) {
	METHTRACE(Engine,Load);
	Name name;
	RUN(readDString(stream,name));
	rename(name);
	StrAttrs stuff;
	RUN(readAttrs(stream,stuff));
	RUN(stream->Read(SIMPER(m_orient)));
	RUN(stream->Read(&locks,sizeof(int),0));
	mod(stringsIn(m_transform,false,&layout,stuff,true));
	RUN(Initialize());
	return S_OK;
}
STDMETHODIMP Engine::Save(IStream *stream,BOOL fClearDirty) {
	METHTRACE(Engine,Save);
	if(!dgserver)
		ERR(E_PENDING);
	RUN(writeDString(stream,gd<Name>(&layout)));
	RUN(writeAttrs(stream,gd<StrAttrs>(&layout)));
	RUN(stream->Write(SIMPER(m_orient)));
    int locksNotPersistent = 0; // don't think they should be
	RUN(stream->Write(&locksNotPersistent,sizeof(int),0));
	return S_OK;
}
STDMETHODIMP Engine::GetSizeMax(ULARGE_INTEGER *pcbSize) {
	METHTRACE(Engine,GetSizeMax);
	ERR(E_NOTIMPL);
}
