/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "includeDynagraph.h"
#include "pointd2coord.h"
#include "Conversion.h"
#include "RegCategory.h"
#include "Node.h"
#include "Edge.h"



struct Engine :
	CComObjectRoot,
	CComCoClass<Engine,&CLSID_DynagraphEngine>,
	IDGLayout,
	IDispatchImpl<DIDGLayout>,
	IConnectionPointImpl<Engine, &IID_IDGLayoutShadow>,
	IConnectionPointContainerImpl<Engine>,
	IPersistStream,
	ChStringHandler<Engine>,
	DynaView
{
	enum DG_ORIENTATION m_orient; // which way is down?
	// find COM objects from Layout objects
	typedef map<Layout::Node*,Node*> vn2com;
	typedef map<Layout::Edge*,Edge*> ve2com;
	vn2com m_nodeMap;
	ve2com m_edgeMap;

	const int secret_code;

	Engine();
	~Engine();

	// for ChStringHandler
	Layout *sh_object() { return &layout; }
	void sh_modify(StrAttrs attrs) {
		SRUN(mod(stringsIn(m_transform,m_useDotDefaults,&layout,attrs,false)));
	}
	void sh_rename(DString name) {
		rename(name);
	}
	HRESULT mod(Update upd);

	// call-in from nodes&edges when they've made a change
	HRESULT go();

	// change internal coords to/from external
	Coord reorient(Coord val,bool in,enum DG_ORIENTATION dir);
	Coord reorient(Coord val,bool in) {
		return reorient(val,in,m_orient);
	}
	Rect reorient(Rect val,bool in,enum DG_ORIENTATION dir) {
		Coord p1 = reorient(val.NW(),in,dir),
			p2 = reorient(val.SE(),in,dir);
		return Rect(min(p1.x,p2.x),max(p1.y,p2.y),max(p1.x,p2.x),min(p1.y,p2.y));
	}
	Rect reorient(Rect val,bool in) {
		return reorient(val,in,m_orient);
	}

	HRESULT wrapNode(Layout::Node *n, IDGDrawnNode **ret);
	HRESULT wrapEdge(Layout::Edge *e,IDGDrawnEdge **ret);

// View events
	void IncrHappened();
	void IncrNewNode(Layout::Node *n);
	void IncrNewEdge(Layout::Edge *e);

	REGISTER_GENERIC(L"LayoutEngine");

BEGIN_COM_MAP(Engine)
	COM_INTERFACE_ENTRY(IDGGraph)
	COM_INTERFACE_ENTRY(IDGLayout)
	COM_INTERFACE_ENTRY(DIDGLayout)
	COM_INTERFACE_ENTRY(IDGStringAttributes)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(IConnectionPointContainer)
	COM_INTERFACE_ENTRY(IPersistStream)
	COM_INTERFACE_ENTRY2(IPersist,IPersistStream)
END_COM_MAP()
BEGIN_CONNECTION_POINT_MAP(Engine)
   CONNECTION_POINT_ENTRY(IID_IDGLayoutShadow)
END_CONNECTION_POINT_MAP()
// IDGGraph
    STDMETHODIMP get_Nodes(IEnumDGNodes **ppRet);
    STDMETHODIMP new_Node(BSTR name, IDGNode **ppRet);
	STDMETHODIMP new_Edge(BSTR name, IDGNode *tail, IDGNode *head, IDGEdge ** ppRet);
	STDMETHODIMP get_Node(BSTR name, IDGNode **ret);
	STDMETHODIMP get_Edge(BSTR name, IDGEdge **ret);
	STDMETHODIMP LoadDotfile(HGLOBAL data,VARIANT_BOOL merge,VARIANT_BOOL del);
	STDMETHODIMP SaveDotfile(HGLOBAL *ret);
    STDMETHODIMP GraphLock();
    STDMETHODIMP GraphUnlock();
// IDGLayout
	STDMETHODIMP Initialize();
    STDMETHODIMP get_Resolution(POINTD *ret);
    STDMETHODIMP put_Resolution(POINTD val);
    STDMETHODIMP get_Ticks(float *ret);
    STDMETHODIMP put_Ticks(float val);
    STDMETHODIMP get_Separation(POINTD *ret);
    STDMETHODIMP put_Separation(POINTD val);
    STDMETHODIMP get_LabelGap(POINTD *ret);
    STDMETHODIMP put_LabelGap(POINTD val);
    STDMETHODIMP get_DefaultNodeSize(POINTD *ret);
    STDMETHODIMP put_DefaultNodeSize(POINTD val);
	STDMETHODIMP get_Orientation(enum DG_ORIENTATION *ret);
	STDMETHODIMP put_Orientation(enum DG_ORIENTATION val);
	STDMETHODIMP get_SplineLevel(enum DG_SPLINELEVEL *ret);
	STDMETHODIMP put_SplineLevel(enum DG_SPLINELEVEL val);
	STDMETHODIMP get_Bounds(RECTD *ret);
	STDMETHODIMP get_ChangeRect(RECTD *ret);
	STDMETHODIMP get_Lines(DGSHAPESDESC **ret);
    STDMETHODIMP get_Locks(int *ret);
	STDMETHODIMP OptimizeAll(VARIANT_BOOL full);
	STDMETHODIMP LoadDotfile(LPCOLESTR filename,VARIANT_BOOL merge,VARIANT_BOOL del);
	STDMETHODIMP SaveDotfile(LPCOLESTR filename);
// DIDGLayout
	STDMETHODIMP put_BackColor(OLE_COLOR clr);
	STDMETHODIMP get_BackColor(OLE_COLOR* pclr);
    STDMETHODIMP SetResolution(double x,double y);
    STDMETHODIMP get_ResolutionX(double *ret);
    STDMETHODIMP get_ResolutionY(double *ret);
    STDMETHODIMP SetSeparation(double x,double y);
    STDMETHODIMP get_SeparationX(double *ret);
    STDMETHODIMP get_SeparationY(double *ret);
	STDMETHODIMP get_Nodes(IDispatch ** ppRet);
	STDMETHODIMP new_Node(BSTR name, DIDGDrawnNode ** ppRet);
	STDMETHODIMP new_Edge(BSTR name, DIDGDrawnNode *tail, DIDGDrawnNode *head, DIDGDrawnEdge ** ppRet);
// IPersistStream
    STDMETHODIMP GetClassID(CLSID *pClassID) {
		METHTRACE(EngineClass,GetClassID);
		*pClassID = GetObjectCLSID();
		return S_OK;
	}
    STDMETHODIMP IsDirty(void);
    STDMETHODIMP Load(IStream *pStm);
    STDMETHODIMP Save(IStream *pStm,BOOL fClearDirty);
    STDMETHODIMP GetSizeMax(ULARGE_INTEGER *pcbSize);
};
