/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "Engine.h"

#include "Collection.h"
#include "allocDGShapes.h"
#include "shapeConv.h"

#include "NodeLabel.h"
#include "RWAttrs.h"

#include <common/colors.h>
#include <common/stringsOut.h>
#include "olecolor2rgb.h"


Node::Node() : secret_code(17), m_engine(0),m_node(0),m_destroying(false),m_incrInserted(false) {
	CONSTRUCTORTRACE(Node);
}
Node::~Node() {
	DESTRUCTORTRACE(Node);
    m_destroying = true;
	if(m_engine->current.find(m_node)) {
		// if the rest of the graph is still alive this will gracefully remove itself
		LOOK(Delete());
	}
	m_engine->forget(m_node);
	for(Layout::nodeedge_iter ei = m_node->alledges().begin(); ei!=m_node->alledges().end();) {
		Edge *edge = m_engine->m_edgeMap[*ei++];
		edge->freeEdge();
	}
	m_engine->destroy(m_node);
	m_engine->GetUnknown()->Release();
}
void Node::sh_modify(StrAttrs attrs) {
	SRUN(mod(stringsIn(m_engine->m_transform,m_node,attrs,false)));
}
void Node::sh_rename(DString name) {
	m_engine->rename(m_node,name);
}
HRESULT Node::mod(Update upd) {
    METHTRACE(Node,mod);
	if(m_engine->current.find(m_node)) {
		m_engine->Q.ModNode(m_node,upd);
		RUNRET(m_engine->go());
	}
	return S_OK;
}
HRESULT Node::figureInside() {
	POINTD largest = {0,0};
	for(NodeLabels::iterator li = labels().begin(); li!=labels().end(); ++li)
		if(li->align==DG_NODELABEL_CENTER)
			largest.x = max(largest.x,li->size.x),
			largest.y = max(largest.y,li->size.y);
	RUN(SetInside(largest.x,largest.y));
	return S_OK;
}
// engine call-ins
void Node::show() {
    METHTRACE(Node,show);
	RUN(fireEvent(GetUnknown(),fireVoid<IDGNodeShadow>(&IDGNodeShadow::Insert)));
}
void Node::hide() {
	METHTRACE(Node,hide);
    if(!m_destroying)
	    RUN(fireEvent(GetUnknown(),fireVoid<IDGNodeShadow>(&IDGNodeShadow::Delete)));
}
// IDGNode
STDMETHODIMP Node::get_Graph(IDGGraph **ppOut) {
	METHTRACE(Node,get_Graph);
	if(!ppOut)
		ERR(E_POINTER);
	(*ppOut = m_engine)->AddRef();
	return S_OK;
}
STDMETHODIMP Node::get_Edges(IEnumDGEdges **ppEdges) {
	METHTRACE(Node,get_Edges);
	/*
	Agnode_t *n = il_find_node((engview_t *)(ILview_t*)m_engine,this);
	(*ppEdges = new CEnumDGEdges(EDGE_ALL,n))->AddRef();
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}

STDMETHODIMP Node::get_OutEdges(IEnumDGEdges **ppEdges) {
	METHTRACE(Node,get_OutEdges);
	/*
	Agnode_t *n = il_find_node((engview_t *)(ILview_t*)m_engine,this);
	(*ppEdges = new CEnumDGEdges(EDGE_OUT,n))->AddRef();
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}

STDMETHODIMP Node::get_InEdges(IEnumDGEdges **ppEdges) {
	METHTRACE(Node,get_InEdges);
	/*
	Agnode_t *n = il_find_node((engview_t *)(ILview_t*)m_engine,this);
	(*ppEdges = new CEnumDGEdges(EDGE_IN,n))->AddRef();
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}

STDMETHODIMP Node::get_Pos(
     PPOINTD ppos) {
	METHTRACE(Node,get_Pos);
	if(!ppos)
		ERR(E_POINTER);
	if(!geom().pos.valid)
		ERR(E_PENDING);
	*ppos = DG2Win(m_engine->reorient(geom().pos,false));
	return S_OK;
}
STDMETHODIMP Node::put_Pos(
     POINTD pos) {
	METHTRACE(Node,put_Pos);
	Position p = m_engine->reorient(Win2DG(pos),true);
	geom().pos = p;
	mod(DG_UPD_MOVE);
	return S_OK;
}
STDMETHODIMP Node::put_Shape(DGSHAPESDESC *val) {
	METHTRACE(Node,put_Shape);
    geom().region.shape.Clear();
    // !val -> use shape definition
    if(!val) {
        gd<IfPolyDef>(m_node).whether = true;
		return S_OK;
    }
    if(!val->n) // no-shape
		return S_OK;
    gd<IfPolyDef>(m_node).whether = false;
    // use first shape, as ShapeGenerator does
	readDGSHAPEDESC(m_engine,val->sps[0],geom().region.shape,false);
	geom().region.updateBounds();
	mod(DG_UPD_REGION);
	return S_OK;
}
STDMETHODIMP Node::get_Shape(DGSHAPESDESC **ret) {
    METHTRACE(Node,get_Shape);
	Drawn &drawn = gd<Drawn>(m_node);
	*ret = allocDGShapes(drawn.size());
	int i=0;
	for(Lines::iterator li = drawn.begin(); li!=drawn.end(); ++i,++li)
		(*ret)->sps[i] = writeDGSHAPEDESC(m_engine,*li,false);
	return S_OK;
}
STDMETHODIMP Node::GetRect(RECTD *ret) {
    METHTRACE(Node,GetRect);
    *ret = rectd(m_engine->reorient(geom().BoundingBox(),false));
	return S_OK;
}
STDMETHODIMP Node::new_Label(IDGNodeLabel **ret) {
    METHTRACE(Node,new_Label);
	NodeLabelWrap *nl = new CComObject<NodeLabelWrap>;
	nl->n = this;
	nl->x = labels().size();
	labels().push_back(NodeLabel());
	labelz.push_back(nl);
	nl->AddRef();
	(*ret = nl)->AddRef();
	return S_OK;
}
STDMETHODIMP Node::Insert() {
	METHTRACE(Node,Insert);
	if(m_engine->current.find(m_node))
		return S_FALSE;
	m_engine->Q.InsNode(m_node);
	RUNRET(m_engine->go());
	return S_OK;
}
STDMETHODIMP Node::Delete() {
	METHTRACE(Node,Delete);
	if(!m_engine->current.find(m_node))
		return S_FALSE;
	m_engine->Q.DelNode(m_node);
	RUNRET(m_engine->go());
	return S_OK;
}
STDMETHODIMP Node::Optimize(VARIANT_BOOL full) {
    METHTRACE(Node,Optimize);
	if(full)
		geom().pos.invalidate();
	mod(DG_UPD_MOVE);
	return S_OK;
}
// DIDGDrawnNode
STDMETHODIMP Node::put_BorderColor(OLE_COLOR clr) {
    METHTRACE(Node,put_BorderColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(m_node).put("color",color.toString());
    return S_OK;
}
STDMETHODIMP Node::get_BorderColor(OLE_COLOR* pclr) {
    METHTRACE(Node,get_BorderColor);
    StrAttrs::iterator ai = gd<StrAttrs>(m_node).find("color");
    if(ai==gd<StrAttrs>(m_node).end()||!ai->second.length())
        *pclr = RGB(0,0,0); // default black
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Node::put_ForeColor(OLE_COLOR clr) {
    METHTRACE(Node,put_ForeColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(m_node).put("fontcolor",color.toString());
    return S_OK;
}
STDMETHODIMP Node::get_ForeColor(OLE_COLOR* pclr) {
    METHTRACE(Node,get_ForeColor);
    StrAttrs::iterator ai = gd<StrAttrs>(m_node).find("fontcolor");
    if(ai==gd<StrAttrs>(m_node).end()||!ai->second.length())
        return get_BorderColor(pclr);
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Node::put_FillColor(OLE_COLOR clr) {
    METHTRACE(Node,put_FillColor);
    Color color;
    RUN(olecolor2rgb(clr,&color));
    gd<StrAttrs2>(m_node).put("fillcolor",color.toString());
    return S_OK;
}
STDMETHODIMP Node::get_FillColor(OLE_COLOR* pclr) {
    METHTRACE(Node,get_FillColor);
    StrAttrs::iterator ai = gd<StrAttrs>(m_node).find("fillcolor");
    if(ai==gd<StrAttrs>(m_node).end()||!ai->second.length())
        return m_engine->get_BackColor(pclr); // default to graph bgcolor
    else {
        Color clr = Color(ai->second).toRGB();
        *pclr = RGB(clr.rgba.r,clr.rgba.g,clr.rgba.b);
    }
    return S_OK;
}
STDMETHODIMP Node::get_Engine(DIDGLayout **ppEngine) {
    METHTRACE(Node,get_Engine);
	CComPtr<IDGGraph> e;
	RUNRET(get_Graph(&e));
	return e.QueryInterface(ppEngine);
}
STDMETHODIMP Node::get_Edges(IDispatch **pEdges) {
    METHTRACE(Node,get_Edges);
	/*
	Agnode_t *n = il_find_node((engview_t *)(ILview_t*)m_engine,this);
	make_edgecoll(n,nextEdge(n),pEdges);
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}
STDMETHODIMP Node::get_OutEdges(IDispatch **pEdges) {
    METHTRACE(Node,get_OutEdges);
	/*
	Agnode_t *n = il_find_node((engview_t *)(ILview_t*)m_engine,this);
	make_edgecoll(n,nextOutEdge(),pEdges);
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}
STDMETHODIMP Node::get_InEdges(IDispatch **pEdges) {
    METHTRACE(Node,get_InEdges);
	/*
	Agnode_t *n = il_find_node((engview_t *)(ILview_t*)m_engine,this);
	make_edgecoll(n,nextInEdge(),pEdges);
	return S_OK;
	*/
	ERR(E_NOTIMPL);
}

STDMETHODIMP Node::get_X(double *ret) {
    METHTRACE(Node,get_X);
	if(!geom().pos.valid)
		ERR(E_FAIL);
	*ret = geom().pos.x;
	return S_OK;
}
STDMETHODIMP Node::get_Y(double *ret) {
    METHTRACE(Node,get_Y);
	if(!geom().pos.valid)
		ERR(E_FAIL);
	*ret = geom().pos.y;
	return S_OK;
}
STDMETHODIMP Node::SetPos(double x, double y) {
    METHTRACE(Node,SetPos);
	POINTD ptd = {x,y};
	return put_Pos(ptd);
}
STDMETHODIMP Node::put_Shape(DIDGShapes *shapes) {
    METHTRACE(Node,put_Shape);
	CComPtr<IDGShapes> s;
	RUN(shapes->QueryInterface(riidret(&s)));
	DGSHAPESDESC *sd;
	RUN(s->get_Shapes(&sd));
	RUN(put_Shape(sd));
	freeDGShapes(sd);
	return S_OK;
}
// DIDGShapeDefinition
STDMETHODIMP Node::get_IsEllipse(VARIANT_BOOL *ret) {
    METHTRACE(Node,get_IsEllipse);
	*ret = gd<PolyDef>(m_node).isEllipse?VARIANT_TRUE:VARIANT_FALSE;
	return S_OK;
}
STDMETHODIMP Node::put_IsEllipse(VARIANT_BOOL val) {
    METHTRACE(Node,put_IsEllipse);
	if(assign(gd<PolyDef>(m_node).isEllipse,val!=VARIANT_FALSE))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Regular(VARIANT_BOOL *ret) {
	*ret = gd<PolyDef>(m_node).regular?VARIANT_TRUE:VARIANT_FALSE;
	return S_OK;
}
STDMETHODIMP Node::put_Regular(VARIANT_BOOL val) {
	if(assign(gd<PolyDef>(m_node).regular,val!=VARIANT_FALSE))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Sides(int *ret) {
    METHTRACE(Node,get_Sides);
	*ret = gd<PolyDef>(m_node).sides;
	return S_OK;
}
STDMETHODIMP Node::put_Sides(int val) {
    METHTRACE(Node,put_Sides);
    if(val<3)
        val = 3;
	if(assign(gd<PolyDef>(m_node).sides,val))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Peripheries(int *ret) {
    METHTRACE(Node,get_Peripheries);
	*ret = gd<PolyDef>(m_node).peripheries;
	return S_OK;
}
STDMETHODIMP Node::put_Peripheries(int val) {
    METHTRACE(Node,put_Peripheries);
    if(val<0)
        val = 0;
	if(assign(gd<PolyDef>(m_node).peripheries,val))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Spacing(double *ret) {
    METHTRACE(Node,get_Spacing);
	*ret = gd<PolyDef>(m_node).perispacing;
	return S_OK;
}
STDMETHODIMP Node::put_Spacing(double val) {
    METHTRACE(Node,put_Spacing);
    if(val<0.0)
        val = 0.0;
	if(assign(gd<PolyDef>(m_node).perispacing,val))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Rotation(double *ret) {
    METHTRACE(Node,get_Rotation);
	*ret = gd<PolyDef>(m_node).rotation;
	return S_OK;
}
STDMETHODIMP Node::put_Rotation(double val) {
    METHTRACE(Node,put_Rotation);
    val = fmod(val,2*M_PI);
	if(assign(gd<PolyDef>(m_node).rotation,val))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Skew(double *ret) {
    METHTRACE(Node,get_Skew);
	*ret = gd<PolyDef>(m_node).skew;
	return S_OK;
}
STDMETHODIMP Node::put_Skew(double val) {
    METHTRACE(Node,put_Skew);
	if(assign(gd<PolyDef>(m_node).skew,val))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_Distortion(double *ret) {
    METHTRACE(Node,get_Distortion);
	*ret = gd<PolyDef>(m_node).distortion;
	return S_OK;
}
STDMETHODIMP Node::put_Distortion(double val) {
    METHTRACE(Node,put_Distortion);
	if(assign(gd<PolyDef>(m_node).distortion,val))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_InsideWidth(double *ret) {
    METHTRACE(Node,get_InsideWidth);
	*ret = gd<PolyDef>(m_node).interior_box.x;
	return S_OK;
}
STDMETHODIMP Node::get_InsideHeight(double *ret) {
    METHTRACE(Node,get_InsideHeight);
	*ret = gd<PolyDef>(m_node).interior_box.y;
	return S_OK;
}
STDMETHODIMP Node::SetInside(double width,double height) {
    METHTRACE(Node,SetInside);
    if(width<0.0)
        width = 0.0;
    if(height<0.0)
        height = 0.0;
	if(assign(gd<PolyDef>(m_node).interior_box,Coord(width,height)))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}
STDMETHODIMP Node::get_OutsideWidth(double *ret) {
    METHTRACE(Node,get_OutsideWidth);
	*ret = gd<PolyDef>(m_node).exterior_box.x;
	return S_OK;
}
STDMETHODIMP Node::get_OutsideHeight(double *ret) {
    METHTRACE(Node,get_OutsideHeight);
	*ret = gd<PolyDef>(m_node).exterior_box.y;
	return S_OK;
}
STDMETHODIMP Node::SetOutside(double width,double height) {
    METHTRACE(Node,SetOutside);
    if(width<0.0)
        width = 0.0;
    if(height<0.0)
        height = 0.0;
	if(assign(gd<PolyDef>(m_node).exterior_box,Coord(width,height)))
		mod(DG_UPD_POLYDEF);
	return S_OK;
}

// IPersistStream
STDMETHODIMP Node::GetClassID(CLSID *pClassID) {
	METHTRACE(Node,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP Node::IsDirty(void) {
	METHTRACE(Node,IsDirty);
	return S_FALSE;
}
STDMETHODIMP Node::Load(IStream *stream) {
	METHTRACE(Node,Load);
	StrAttrs stuff;
	RUN(readAttrs(stream,stuff));
	mod(stringsIn(m_engine->m_transform,m_node,stuff,true));
	return S_OK;
}
STDMETHODIMP Node::Save(IStream *stream,BOOL fClearDirty) {
	METHTRACE(Node,Save);
	RUN(writeAttrs(stream,gd<StrAttrs>(m_node)));
	return S_OK;
}
STDMETHODIMP Node::GetSizeMax(ULARGE_INTEGER *pcbSize) {
	METHTRACE(Node,GetSizeMax);
	pcbSize->QuadPart = 100;
	return S_OK;
}
/*
//////////////////////////////////////
// CEnumDGNodes
CEnumDGNodes::CEnumDGNodes(Agnode_t *n) {
	METHTRACE(CEnumDGNodes,CEnumDGNodes);
	int cnt = 0;
	Agnode_t *in;
	for(in = n; in; in = agnxtnode(in)) cnt++;
	IDGNode **arr = new IDGNode *[cnt],
		**ia;
	for(in = n,ia = arr; in; in = agnxtnode(in))
		*ia++ = (Node *)il_node(in);
	Init(arr,ia,NULL,AtlFlagTakeOwnership); // "it's yours now"
}
*/