/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "NodePersistor.h"

#include "cacheStreamableObject.h"
#include "countStream.h"

NodePersistor::NodePersistor() {
	CONSTRUCTORTRACE(NodePersistor);
}
NodePersistor::~NodePersistor() {
	DESTRUCTORTRACE(NodePersistor);
}
// IDGNodePersistor
STDMETHODIMP NodePersistor::put_Name(BSTR val) {
	m_name = val;
	return S_OK;
}
STDMETHODIMP NodePersistor::put_Cache(IStream *dgdata) {
	m_cache = dgdata;
	return S_OK;
}
STDMETHODIMP NodePersistor::put_MkEngine(IMoniker *val) {
	METHTRACE(NodePersistor,put_MkEngine);
	if(!m_group)
		ERR(E_PENDING);
	CComPtr<IMoniker> old = m_mkEngine;
	m_mkEngine = val;
	RUN(m_group->AddStgRef(m_mkEngine));
	if(old)
		RUN(m_group->ReleaseStgRef(old));
	return S_OK;
}
// IGroupItem
STDMETHODIMP NodePersistor::Connect(IGroup *group) {
	METHTRACE(NodePersistor,Connect);
	m_group = group;
	if(!m_group)
		m_subject.Release();
	return S_OK;
}
STDMETHODIMP NodePersistor::Copy(IBindCtx *bctx, IGroup *dest, IUnknown *aggr, REFIID riid, void **ret) {
	METHTRACE(NodePersistor,Copy);
	CComPtr<IDGNodePersistor> np2;
	RUN(CREATE(np2,CLSID_DGNodePersistor));
	CComPtr<IMoniker> mkE2;
	RUN(dest->Paste(bctx,GP_SrcIsCont,m_group,m_mkEngine,IID_NULL,0,&mkE2));
	RUN(CComQIPtr<IGroupItem>(np2)->Connect(dest));
	RUN(np2->put_Cache(m_cache));
	RUN(np2->put_MkEngine(mkE2));
	if(m_subject)
		RUN(CComQIPtr<DIPersistor>(np2)->SetSubject(m_subject!=0, m_subject)); // realized==have subject
	return np2->QueryInterface(riid,ret);
}
STDMETHODIMP NodePersistor::Realize(IBindCtx *bctx) {
	METHTRACE(NodePersistor,Realize);
	if(m_subject)
		return S_FALSE;
	CComPtr<IDGLayout> eng;
	RUN(m_mkEngine->BindToObject(bctx?bctx:BindCtx(),PointerMoniker(m_group),riidret(&eng)));
	CComPtr<IDGNode> n;
	RUN(eng->new_Node(m_name,&n));
	RUN(n.QueryInterface(&m_subject));
	if(m_cache) {
		RUN(CComQIPtr<IPersistStream>(m_subject)->Load(m_cache));
		RUN(m_cache->Seek(castLarge(0),STREAM_SEEK_SET,0));
		m_cache.Release();
	}
	return S_OK;
}
STDMETHODIMP NodePersistor::Remove() {
	METHTRACE(NodePersistor,Remove);
	RUN(m_group->ReleaseStgRef(m_mkEngine));
	return S_OK;
}
// IPersistStream
STDMETHODIMP NodePersistor::GetClassID(CLSID *pClassID) {
	METHTRACE(NodePersistor,GetClassID);
	*pClassID = GetObjectCLSID();
	return S_OK;
}
STDMETHODIMP NodePersistor::IsDirty(void) {
	METHTRACE(NodePersistor,IsDirty);
	return S_FALSE;
}
STDMETHODIMP NodePersistor::Load(IStream *pStm) {
	METHTRACE(NodePersistor,Load);
	RUN(m_name.ReadFromStream(pStm));
	RUN(OleLoadFromStream(pStm,riidret(&m_mkEngine)));
	return cacheStreamableObject(pStm,&m_cache);
}
STDMETHODIMP NodePersistor::Save(IStream *pStm,BOOL fClearDirty) {
	METHTRACE(NodePersistor,Save);
	RUN(m_name.WriteToStream(pStm));
	RUN(OleSaveToStream(m_mkEngine,pStm));
	if(m_subject)
		return saveWithSize(pStm,CComQIPtr<IPersistStream>(m_subject),fClearDirty);
	else {
		if(!m_cache)
			ERR(E_PENDING);
		return writeCache(pStm,m_cache);
	}
}
STDMETHODIMP NodePersistor::GetSizeMax(ULARGE_INTEGER *pcbSize) {
	METHTRACE(NodePersistor,GetSizeMax);
	CComObjectStack<countStream> count;
	RUN(Save(&count,false));
	*pcbSize = count.m_pos;
	return S_OK;
}
// IPersistor
STDMETHODIMP NodePersistor::SetSubject(VARIANT_BOOL realized,IUnknown *pUnkSite) {
	METHTRACE(NodePersistor,SetSubject);
	// for our purposes, having a subject means being realized (this persistor doesn't support
	// connecting a live object with Realize because there's no demand (yet?)).  So we can ignore
	// realized param.
	QISET(m_subject = pUnkSite);
	return S_OK;
}
STDMETHODIMP NodePersistor::get_Subject(IUnknown **ret) {
	METHTRACE(NodePersistor,GetSubject);
	m_subject.CopyTo((IDGDrawnNode**)ret);
	return S_OK;
}
