/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include <common/emitGraph.h>
#include <sstream>
#include "readHGlobalDotfile.h"

template<class NG,class IncrHandler,class ThreadModel>
struct StringGraph;

template<class NG,class IH,class TM>
struct StringNode :
    CComObjectRoot,
    StringHandler<StringNode>,
    IDGNode
{
    StringGraph<NG,IH,TM> *m_graph;
    typename NG::Node *m_node;

BEGIN_COM_MAP(StringNode)
    COM_INTERFACE_ENTRY(IDGNode)
    COM_INTERFACE_ENTRY(IDGStringAttributes)
END_COM_MAP()

	// for StringHandler
    typename NG::Node *sh_object() {
        return m_node;
    }
    void sh_modify(StrAttrs attrs) {
        m_graph->incr_ev_mod_node(gd<Name>(m_node),attrs);
    }
    void sh_rename(DString name) {
        m_graph->g->rename(m_node,name);
    }

 // IDGNode
	STDMETHODIMP get_Graph(IDGGraph **ret);
	STDMETHODIMP get_Edges(IEnumDGEdges **pEdges);
	STDMETHODIMP get_OutEdges(IEnumDGEdges **pEdges);
	STDMETHODIMP get_InEdges(IEnumDGEdges **pEdges);
};
template<class NG,class IH,class TM>
struct StringEdge :
    CComObjectRoot,
    StringHandler<StringEdge>,
    IDGEdge
{
    typename NG::Edge *m_edge;
    StringGraph<NG,IH,TM> *m_graph;

BEGIN_COM_MAP(StringEdge)
    COM_INTERFACE_ENTRY(IDGEdge)
    COM_INTERFACE_ENTRY(IDGStringAttributes)
END_COM_MAP()

	// for StringHandler
    typename NG::Edge *sh_object() {
        return m_edge;
    }
    void sh_modify(StrAttrs attrs) {
        m_graph->incr_ev_mod_edge(gd<Name>(m_edge),attrs);
    }
    void sh_rename(DString name) {
        m_graph->g->rename(m_edge,name);
    }

// IDGEdge
	STDMETHODIMP get_Graph(IDGGraph **ret);
	STDMETHODIMP get_TailNode(IDGNode **pNode);
	STDMETHODIMP get_HeadNode(IDGNode **pNode);
};
template<class NG, class IncrHandler = AbsGraphHandler<NG>, class ThreadModel = CComObjectThreadModel>
struct StringGraph :
    CComObjectRootEx<ThreadModel>,
    IncrHandler,
    StringHandler<StringGraph>,
    IDGGraph
{
    typedef map<typename NG::Node*,StringNode<NG,IncrHandler,ThreadModel>*> sn2com;
    typedef map<typename NG::Edge*,StringEdge<NG,IncrHandler,ThreadModel>*> se2com;
    sn2com m_nodeMap;
    se2com m_edgeMap;

    // override to change node & edge types to something derived from StringNode, StringEdge
    virtual StringNode<NG,IncrHandler,ThreadModel> *createNode() {
        return new CComObject<StringNode<NG,IncrHandler,ThreadModel> >;
    }
    virtual StringEdge<NG,IncrHandler,ThreadModel> *createEdge() {
        return new CComObject<StringEdge<NG,IncrHandler,ThreadModel> >;
    }
    StringNode<NG,IncrHandler,ThreadModel> *getCOM(typename NG::Node *n) {
        StringNode<NG,IncrHandler,ThreadModel> *&entry = m_nodeMap[n];
        if(!entry) {
            entry = createNode();
            entry->GetUnknown()->AddRef();
            entry->m_node = n;
            entry->m_graph = this;
        }
        return entry;
    }
    StringEdge<NG,IncrHandler,ThreadModel> *getCOM(typename NG::Edge *e) {
        StringEdge<NG,IncrHandler,ThreadModel> *&entry = m_edgeMap[e];
        if(!entry) {
            entry = createEdge();
            entry->GetUnknown()->AddRef();
            entry->m_edge = e;
            entry->m_graph = this;
        }
        return entry;
    }

    StringGraph(NG *g = 0) : IncrHandler(g) {
        incr_ev_open_graph(gd<Name>(this->g),gd<StrAttrs>(this->g));
    }

    ~StringGraph() {
        for(sn2com::iterator ni = m_nodeMap.begin(); ni!=m_nodeMap.end(); ++ni)
            ni->second->GetUnknown()->Release();
        for(se2com::iterator ei = m_edgeMap.begin(); ei!=m_edgeMap.end(); ++ei)
            ei->second->GetUnknown()->Release();
    }

BEGIN_COM_MAP(StringGraph)
    COM_INTERFACE_ENTRY(IDGGraph)
    COM_INTERFACE_ENTRY(IDGStringAttributes)
END_COM_MAP()

	// for StringHandler
	NG *sh_object() {
        return g;
    }
    void sh_modify(StrAttrs attrs) {
        incr_ev_mod_graph(attrs);
    }
    void sh_rename(DString name) {
        g_dinoMachine.rename(g_dinoMachine.ndict[gd<Name>(g)],name);
        gd<Name>(g) = name;
    }

// IDGGraph
	STDMETHODIMP get_Nodes(IEnumDGNodes ** ppRet);
	STDMETHODIMP new_Node(BSTR name, IDGNode ** ppRet);
	STDMETHODIMP new_Edge(BSTR name, IDGNode *tail, IDGNode *head, IDGEdge ** ppRet);
	STDMETHODIMP get_Node(BSTR name, IDGNode **ret);
	STDMETHODIMP get_Edge(BSTR name, IDGEdge **ret);
	STDMETHODIMP LoadDotfile(HGLOBAL data,VARIANT_BOOL merge,VARIANT_BOOL del);
	STDMETHODIMP SaveDotfile(HGLOBAL *ret);
    STDMETHODIMP GraphLock();
    STDMETHODIMP GraphUnlock();
};

// StringNode
// StringNode's IDGNode
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringNode<NG,IH,TM>::get_Graph(IDGGraph **ret) {
    (*ret = m_graph)->AddRef();
	return S_OK;
}
template<typename NG,typename IH,typename TM,typename Coll>
HRESULT getEdges(Coll &c,StringGraph<NG,IH,TM> *graph,IEnumDGEdges **ret) {
    int n = c.size();
    IDGEdge **ray = new IDGEdge*[n];
    for(Coll::iterator ei = c.begin(); ei!=c.end(); ++ei)
        (*ray = graph->getCOM(*ei))->AddRef();
    typedef CComObject<CComEnum<IEnumDGEdges,&IID_IEnumDGEdges,IDGEdge*,_CopyInterface<IDGEdge> > > Enum;
    Enum *enume = new Enum;
    enume->Init(ray,ray+n,0,AtlFlagTakeOwnership);
    (*ret = enume)->AddRef();
    return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringNode<NG,IH,TM>::get_Edges(IEnumDGEdges **pEdges) {
    RUN(getEdges(m_node->alledges(),m_graph,pEdges));
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringNode<NG,IH,TM>::get_OutEdges(IEnumDGEdges **pEdges) {
    RUN(getEdges(m_node->outs(),m_graph,pEdges));
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringNode<NG,IH,TM>::get_InEdges(IEnumDGEdges **pEdges) {
    RUN(getEdges(m_node->ins(),m_graph,pEdges));
	return S_OK;
}
// StringEdge's IDGEdge
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringEdge<NG,IH,TM>::get_Graph(IDGGraph **ret) {
    (*ret = m_graph)->AddRef();
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringEdge<NG,IH,TM>::get_TailNode(IDGNode **pNode) {
    (*pNode = m_graph->getCOM(m_edge->tail))->AddRef();
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringEdge<NG,IH,TM>::get_HeadNode(IDGNode **pNode) {
    (*pNode = m_graph->getCOM(m_edge->head))->AddRef();
	return S_OK;
}
// IDGGraph
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::get_Nodes(IEnumDGNodes ** ppRet) {
    int n = g->nodes().size();
    IDGNode **ray = new IDGNode*[n];
    for(typename NG::node_iter ni = g->nodes().begin(); ni!=g->nodes().end(); ++ni)
        (*ray = getCOM(*ni))->AddRef();
    typedef CComObject<CComEnum<IEnumDGNodes,&IID_IEnumDGNodes,IDGNode*,_CopyInterface<IDGNode> > > Enum;
    Enum *enumn = new Enum;
    enumn->Init(ray,ray+n,0,AtlFlagTakeOwnership);
    (*ppRet = enumn)->AddRef();
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::new_Node(BSTR name, IDGNode ** ppRet) {
    USES_CONVERSION;
    DString name2 = W2A(name);
    incr_ev_ins_node(name2,StrAttrs(),true);
    typename NG::Node *nn = g->ndict[name2];
    (*ppRet = getCOM(nn))->AddRef();
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::new_Edge(BSTR name, IDGNode *tail, IDGNode *head, IDGEdge ** ppRet) {
    StringNode<NG,IH,TM> *st = static_cast<StringNode<NG,IH,TM>*>(tail),
        *sh = static_cast<StringNode<NG,IH,TM>*>(head);
    USES_CONVERSION;
    DString name2 = W2A(name);
    incr_ev_ins_edge(name2,gd<Name>(st->m_node),gd<Name>(sh->m_node),StrAttrs());
    typename NG::Edge *ne = g->edict[name2];
    (*ppRet = getCOM(ne))->AddRef();
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::get_Node(BSTR name, IDGNode **ret) {
    USES_CONVERSION;
    typename NG::Node *nn = g->ndict[W2A(name)];
    if(nn) {
        (*ret = getCOM(nn))->AddRef();
	    return S_OK;
    }
    else {
        *ret = 0;
        ERR(E_FAIL);
    }
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::get_Edge(BSTR name, IDGEdge **ret) {
    USES_CONVERSION;
    typename NG::Edge *ne = g->edict[W2A(name)];
    if(ne) {
        (*ret = getCOM(ne))->AddRef();
	    return S_OK;
    }
    else {
        *ret = 0;
        ERR(E_FAIL);
    }
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::LoadDotfile(HGLOBAL data,VARIANT_BOOL merge,VARIANT_BOOL del) {
    StrGraph *sg = readHGlobalDotfile(data);
    if(!sg)
        ERR(E_FAIL);
    incr_ev_load_strgraph(sg,merge,del);
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::SaveDotfile(HGLOBAL *ret) {
    ostringstream f;
	emitGraph(f,g);
    f << ends;
	string s = f.str();
	*ret = GlobalAlloc(GHND,s.length());
	void *p = GlobalLock(*ret);
	s.copy((char*)p,s.length());
	GlobalUnlock(*ret);
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::GraphLock() {
    incr_ev_lock();
	return S_OK;
}
template<typename NG,typename IH,typename TM>
STDMETHODIMP StringGraph<NG,IH,TM>::GraphUnlock() {
    incr_ev_unlock();
	return S_OK;
}

struct AbstractGraph : StringGraph<StrGraph>,
    CComCoClass<AbstractGraph,&CLSID_DynagraphAbstractGraph>
{
    REGISTER_GENERIC(L"AbstractGraph")
};
