/**********************************************************
*      This software is part of the graphviz toolset      *
*                http://www.graphviz.org/                 *
*                                                         *
*            Copyright (c) 1994-2005 AT&T Corp.           *
*                and is licensed under the                *
*            Common Public License, Version 1.0           *
*                      by AT&T Corp.                      *
*                                                         *
*        Information and Software Systems Research        *
*              AT&T Research, Florham Park NJ             *
*                                                         *
*               This version available from               *
*                   http://dynagraph.org                  *
**********************************************************/


#include "stdafx.h"
#include "DinographUI.h"
#include "MBasics.h"
#include "ambiance.h"
#include "fillcombo.h"
#include <common/StringDict.h>
#include <common/parsestr.h>
#include "defaultFont.h"
#include <sstream>

#define MAIN_WINDOW 5
#define PROP_WINDOW 6

struct ComboSink :
	CComObjectRoot,
	IMCCComboBoxEvents
{
	DinographUI *m_dino;
    int m_id;
BEGIN_COM_MAP(ComboSink)
	COM_INTERFACE_ENTRY(IMCCComboBoxEvents)
END_COM_MAP()
	STDMETHODIMP Chose(BSTR val) {
        return m_dino->selected(m_id,val);
	}
	STDMETHODIMP TextChanged(BSTR val) {
        return S_OK;
    }
};
struct PropTypeSink : ComboSink {
	STDMETHODIMP Chose(BSTR val) {
        return m_dino->setType(val);
	}
};
struct PropScopeSink : ComboSink {
	STDMETHODIMP Chose(BSTR val) {
        return m_dino->setScope(val);
	}
};

struct ButtonSink :
	CComObjectRoot,
	IMCCButtonEvents
{
	DinographUI *m_dino;
    int m_id;
BEGIN_COM_MAP(ButtonSink)
	COM_INTERFACE_ENTRY(IMCCButtonEvents)
END_COM_MAP()
};

// for search, pattern
struct SaveButtonSink : ButtonSink {
    STDMETHODIMP Click() {
        return m_dino->save(m_id);
    }
};
// for "text" checkboxes
struct TextCheckSink : ButtonSink {
    CComPtr<IMCCButton> m_check;
    STDMETHODIMP Click() {
        int state;
        RUN(m_check->get_State(&state));
        RUN(m_dino->textCheck(m_id,bool(state&BST_CHECKED)));
        return S_OK;
    }
};
// for input
struct StickyButtonSink :
	CComObjectRoot,
	IMCCStickyButtonEvents
{
	DinographUI *m_dino;
BEGIN_COM_MAP(StickyButtonSink)
	COM_INTERFACE_ENTRY(IMCCButtonEvents)
	COM_INTERFACE_ENTRY(IMCCStickyButtonEvents)
END_COM_MAP()
};
struct StickySaveSink :
	StickyButtonSink
{
	STDMETHODIMP Click() {
        METHTRACE(SaveSink,Click);
		RUN(m_dino->save(DINOGRAPH_INPUT_START));
		return S_OK;
	}
	STDMETHODIMP Stick(VARIANT_BOOL val) {
        METHTRACE(SaveSink,Stick);
        m_dino->alwaysSave = val!=0;
		return S_OK;
	}
};
struct GrabSink : StickyButtonSink {
	STDMETHODIMP Click() {
        METHTRACE(GrabSink,Click);
		RUN(m_dino->grab());
		return S_OK;
	}
	STDMETHODIMP Stick(VARIANT_BOOL val) {
        METHTRACE(GrabSink,Stick);
        if(m_dino->alwaysGrab = val!=0)
            RUN(m_dino->inputshowctl->put_Stuck(VARIANT_FALSE));
		return S_OK;
	}
};
struct ShowSink : StickyButtonSink {
	STDMETHODIMP Click() {
        METHTRACE(ShowSink,Click);
		RUN(m_dino->show());
		return S_OK;
	}
	STDMETHODIMP Stick(VARIANT_BOOL val) {
        METHTRACE(ShowSink,Stick);
        if(m_dino->alwaysShow = val!=0)
            RUN(m_dino->inputgrabctl->put_Stuck(VARIANT_FALSE));
		return S_OK;
	}
};
struct SelectionSink :
	CComObjectRoot,
	IItemSetEvents
{
	DinographUI *m_dino;
    int m_id;
BEGIN_COM_MAP(SelectionSink)
	COM_INTERFACE_ENTRY(IItemSetEvents)
END_COM_MAP();

	STDMETHODIMP Added(IMoniker *mk) {
        METHTRACE(SelectionSink,Added);
		RUN(m_dino->selectChange(m_id));
		return S_OK;
	}
	STDMETHODIMP Removed(IMoniker *mk) {
        METHTRACE(SelectionSink,Removed);
		RUN(m_dino->selectChange(m_id));
		return S_OK;
	}
};

DinographUI::DinographUI() :
    searchusetext(false),patternusetext(false),inputusetext(false),
    alwaysSave(false),alwaysGrab(false),alwaysShow(false),
    propNodes(true),propEdges(true),propAbstract(true)
{
}

CComPtr<IMCCComboBox> DinographUI::getChooser(int type) {
    CComPtr<IMCCComboBox> cbox;
    switch(type) {
    case DINOGRAPH_INPUT_PATTERN:
        cbox = patternchoicectl;
        break;
    case DINOGRAPH_INPUT_SEARCH:
        cbox = searchchoicectl;
        break;
    case DINOGRAPH_INPUT_START:
        cbox = inputchoicectl;
        break;
    }
    return cbox;
}
// IDinographCentralEvents
STDMETHODIMP DinographUI::InputAdded(DINOGRAPH_INPUT_TYPES type,LPCOLESTR name) {
    METHTRACE(DinographUI,InputAdded);
    CComPtr<IMCCComboBox> cbox = getChooser(type);
    if(!cbox)
        ERR(E_INVALIDARG);
    RUN(cbox->Add(CComBSTR(name)));
    if(type==DINOGRAPH_INPUT_SEARCH)
        RUN(runchoicectl->Add(CComBSTR(name)));
	return S_OK;
}
STDMETHODIMP DinographUI::InputDeleted(DINOGRAPH_INPUT_TYPES type,LPCOLESTR name) {
    METHTRACE(DinographUI,InputDeleted);
    CComPtr<IMCCComboBox> cbox = getChooser(type);
    if(!cbox)
        ERR(E_INVALIDARG);
    RUN(cbox->Remove(CComBSTR(name)));
    if(type==DINOGRAPH_INPUT_SEARCH)
        RUN(runchoicectl->Remove(CComBSTR(name)));
	return S_OK;
}
struct Throwaway {
    DWORD no;
    DWORD *operator &() {
        no = 0;
        return &no;
    }
};
HRESULT DinographUI::start(IDinographCentral *central, IDGIncrClient *incrClient, IUnknown *mp) {
    MBasics mb;
    RUN(mb.init(mp));
    topMC = mb.c;

    Throwaway compost;

    m_central = central;
    m_incrClient = incrClient;
    RUN(AtlAdvise(m_central,GetUnknown(),IID_IDinographCentralEvents,&compost));
    // blot out the original container with panes of other stuff
    RUN(mb.c->new_Layer(&topL));
    RUN(topL->PutBehind(0));
    RUN(topL->put_IsTransformed(VARIANT_FALSE));
    CComPtr<IDispatchEx> amb;
    CComPtr<IMoniker> mkAmbiance;
    RUN(ambiance(L"Top Layer Ambiance",mb,&amb,&mkAmbiance));
    RUN(topL->put_DefaultAmbiance(amb));
	// set ShowHatching = FALSE;
	CComBSTR showHatch(L"ShowHatching");
	DISPID idSH; // gets set to DISPID_AMBIENT_SHOWHATCHING
	RUN(amb->GetDispID(showHatch,fdexNameEnsure|fdexNameCaseInsensitive,&idSH));
	CComVariant v(false);
	RUN(CComDispatchDriver::PutProperty(amb,idSH,&v));

    RUN(main.CoCreateInstance(CLSID_DynagraphControl));
    RUN(search.CoCreateInstance(CLSID_DynagraphControl));
    RUN(pattern.CoCreateInstance(CLSID_DynagraphControl));
    RUN(input.CoCreateInstance(CLSID_DynagraphControl));

    // give these persiststorage so that they have group facilities
    RUN(mb.add(L"Dinograph Browser",main,dont_keep,0,true));
    RUN(mb.add(L"Dinograph Search",search,dont_keep,0,true));
    RUN(mb.add(L"Dinograph Pattern",pattern,dont_keep,0,true));
    RUN(mb.add(L"Dinograph Input",input,dont_keep,0,true));

    RUN(main->put_HatchMode(MHM_AUTO));
    RUN(search->put_HatchMode(MHM_AUTO));
    RUN(pattern->put_HatchMode(MHM_AUTO));
    RUN(input->put_HatchMode(MHM_AUTO));

	RUN(main->put_AutoSize(VARIANT_FALSE));
	RUN(search->put_AutoSize(VARIANT_FALSE));
	RUN(pattern->put_AutoSize(VARIANT_FALSE));
	RUN(input->put_AutoSize(VARIANT_FALSE));

    RUN(mb.c->new_Site(&mainS));
    RUN(mb.c->new_Site(&searchS));
    RUN(mb.c->new_Site(&patternS));
    RUN(mb.c->new_Site(&inputS));

    RUN(mainS->put_AutoSetExtent(VARIANT_TRUE));
    RUN(searchS->put_AutoSetExtent(VARIANT_TRUE));
    RUN(patternS->put_AutoSetExtent(VARIANT_TRUE));
    RUN(inputS->put_AutoSetExtent(VARIANT_TRUE));

	RUN(mainS->put_ClipMode(MCM_SITE_RECT));
	RUN(searchS->put_ClipMode(MCM_SITE_RECT));
	RUN(patternS->put_ClipMode(MCM_SITE_RECT));
	RUN(inputS->put_ClipMode(MCM_SITE_RECT));

    RUN(mainS->put_Layer(topL));
    RUN(searchS->put_Layer(topL));
    RUN(patternS->put_Layer(topL));
    RUN(inputS->put_Layer(topL));

    RUN(mainS->put_Control(main));
    RUN(searchS->put_Control(search));
    RUN(patternS->put_Control(pattern));
    RUN(inputS->put_Control(input));

    CComPtr<IFontDisp> fd;
    RUN(defaultFont(L"Arial Black",110000,&fd));
    titleFont = fd;

    RUN(columns.CoCreateInstance(CLSID_MCCPane));
    RUN(columns->put_IsVert(false));
    RECT r = rect(0,0,1000,1000);
    RUN(columns->put_Rect(&r));
    RUN(columns->put_Montage(mb.c));

    const int titleheight = 29,
        combowidth = 150;

    // left column: title+main viewer
    RUN(leftcol.CoCreateInstance(CLSID_MCCPane));
    RUN(columns->AddPane(0,leftcol));
    RUN(columns->put_PaneXtra(0,1));
	RUN(maintitle.CoCreateInstance(CLSID_MCCPaneTitle));
    RUN(maintitle->put_BorderSize(2));
    RUN(maintitle->put_Text(CComBSTR("main")));
    RUN(maintitle->put_Content(mainS));
    RUN(maintitle->put_Font(titleFont));
    RUN(maintitle->put_TitleHeight(titleheight));
    RUN(leftcol->put_IsVert(VARIANT_TRUE));
	RUN(leftcol->put_Border(rect(2,2,2,2)));
	RUN(mb.c->new_Site(&maintitleS));
	RUN(maintitleS->put_Control(maintitle));
    RUN(maintitleS->put_Layer(topL));
	RUN(leftcol->put_Content(maintitleS));
	RUN(maintitleS->put_Intent(MCCS_ACTIVE));
    RUN(maintitlespot.CoCreateInstance(CLSID_MCCPane));
    RUN(mainspot.CoCreateInstance(CLSID_MCCPane));
    RUN(maintitlespot->put_Border(rect(1,1,1,1)));
    RUN(leftcol->AddPane(0,maintitlespot));
    RUN(leftcol->put_PaneSize(0,titleheight));
    RUN(leftcol->AddPane(1,mainspot));
    RUN(leftcol->put_PaneXtra(1,1));
    RUN(mainspot->put_Content(mainS));

    // right column: search, pattern, and input, and their title bars
    RUN(rightcol.CoCreateInstance(CLSID_MCCPane));
    RUN(columns->AddPane(1,rightcol));
    RUN(columns->put_PaneSize(1,470));
    RUN(rightcol->put_IsVert(VARIANT_TRUE));
    RUN(searchwind.CoCreateInstance(CLSID_MCCPane));
    RUN(searchwind->put_Border(rect(2,2,2,2)));
    RUN(searchwind->put_IsVert(VARIANT_TRUE));
    RUN(rightcol->AddPane(0,searchwind));
	RUN(searchtitle.CoCreateInstance(CLSID_MCCPaneTitle));
    RUN(searchtitle->put_BorderSize(2));
    RUN(searchtitle->put_Text(CComBSTR("search")));
    RUN(searchtitle->put_Content(searchS));
    RUN(searchtitle->put_Font(titleFont));
    RUN(searchtitle->put_TitleHeight(titleheight));
	RUN(mb.c->new_Site(&searchtitleS));
	RUN(searchtitleS->put_Control(searchtitle));
    RUN(searchtitleS->put_Layer(topL));
	RUN(searchwind->put_Content(searchtitleS));
	RUN(searchtitleS->put_Intent(MCCS_ACTIVE));
        RUN(searchtitlespot.CoCreateInstance(CLSID_MCCPane));
        RUN(searchspot.CoCreateInstance(CLSID_MCCPane));
        RUN(searchtitlespot->put_Border(rect(1,1,1,1)));
        RUN(searchwind->AddPane(0,searchtitlespot));
        RUN(searchwind->AddPane(1,searchspot));
        RUN(searchwind->put_PaneSize(0,titleheight));
        RUN(searchwind->put_PaneXtra(1,1));
        RUN(searchspot->put_Content(searchS));
    RUN(patternwind.CoCreateInstance(CLSID_MCCPane));
    RUN(patternwind->put_Border(rect(2,2,2,2)));
    RUN(patternwind->put_IsVert(VARIANT_TRUE));
    RUN(rightcol->AddPane(1,patternwind));
	RUN(patterntitle.CoCreateInstance(CLSID_MCCPaneTitle));
    RUN(patterntitle->put_BorderSize(2));
    RUN(patterntitle->put_Text(CComBSTR("pattern")));
    RUN(patterntitle->put_Content(patternS));
    RUN(patterntitle->put_Font(titleFont));
    RUN(patterntitle->put_TitleHeight(titleheight));
	RUN(mb.c->new_Site(&patterntitleS));
	RUN(patterntitleS->put_Control(patterntitle));
    RUN(patterntitleS->put_Layer(topL));
	RUN(patternwind->put_Content(patterntitleS));
	RUN(patterntitleS->put_Intent(MCCS_ACTIVE));
        RUN(patterntitlespot.CoCreateInstance(CLSID_MCCPane));
        RUN(patternspot.CoCreateInstance(CLSID_MCCPane));
        RUN(patterntitlespot->put_Border(rect(1,1,1,1)));
        RUN(patternwind->AddPane(0,patterntitlespot));
        RUN(patternwind->AddPane(1,patternspot));
        RUN(patternwind->put_PaneSize(0,titleheight));
        RUN(patternwind->put_PaneXtra(1,1));
        RUN(patternspot->put_Content(patternS));
    RUN(inputwind.CoCreateInstance(CLSID_MCCPane));
    RUN(inputwind->put_Border(rect(2,2,2,2)));
    RUN(inputwind->put_IsVert(VARIANT_TRUE));
    RUN(rightcol->AddPane(2,inputwind));
	RUN(inputtitle.CoCreateInstance(CLSID_MCCPaneTitle));
    RUN(inputtitle->put_BorderSize(2));
    RUN(inputtitle->put_Text(CComBSTR("input")));
    RUN(inputtitle->put_Content(inputS));
    RUN(inputtitle->put_Font(titleFont));
    RUN(inputtitle->put_TitleHeight(titleheight));
	RUN(mb.c->new_Site(&inputtitleS));
	RUN(inputtitleS->put_Control(inputtitle));
    RUN(inputtitleS->put_Layer(topL));
	RUN(inputwind->put_Content(inputtitleS));
	RUN(inputtitleS->put_Intent(MCCS_ACTIVE));
        RUN(inputtitlespot.CoCreateInstance(CLSID_MCCPane));
        RUN(inputspot.CoCreateInstance(CLSID_MCCPane));
        RUN(inputtitlespot->put_Border(rect(1,1,1,1)));
        RUN(inputwind->AddPane(0,inputtitlespot));
        RUN(inputwind->put_PaneSize(0,titleheight));
        RUN(inputwind->AddPane(1,inputspot));
        RUN(inputwind->put_PaneXtra(1,1));
        RUN(inputspot->put_Content(inputS));
    // the property pane doesn't get activated
    RUN(propwind.CoCreateInstance(CLSID_MCCPane));
    RUN(propwind->put_Border(rect(2,2,2,2)));
    RUN(propwind->put_IsVert(VARIANT_TRUE));
    RUN(rightcol->AddPane(3,propwind));
	RUN(proptitle.CoCreateInstance(CLSID_MCCPaneTitle));
    RUN(proptitle->put_BorderSize(2));
    RUN(proptitle->put_Text(CComBSTR("properties")));
    RUN(proptitle->put_Font(titleFont));
    RUN(proptitle->put_TitleHeight(titleheight));
	RUN(mb.c->new_Site(&proptitleS));
	RUN(proptitleS->put_Control(proptitle));
    RUN(proptitleS->put_Layer(topL));
	RUN(propwind->put_Content(proptitleS));
	RUN(proptitleS->put_Intent(MCCS_ACTIVE));
        RUN(proptitlespot.CoCreateInstance(CLSID_MCCPane));
        RUN(proptitlespot->put_Border(rect(1,1,1,1)));
        RUN(propspot.CoCreateInstance(CLSID_MCCPane));
        RUN(propspot->put_IsVert(VARIANT_TRUE));
        RUN(propwind->AddPane(0,proptitlespot));
        RUN(propwind->put_PaneSize(0,titleheight));
        RUN(propwind->AddPane(1,propspot));
        RUN(propwind->put_PaneXtra(1,1));
        RUN(proppane.CoCreateInstance(CLSID_MCCPropertiesPane));
        RUN(proppane->put_CellHeight(titleheight-2));
        RUN(proppane->put_Layer(topL));
        RUN(proppane->put_Pane(propspot));
        RUN(proppane->put_NothingText(CComBSTR("<don't set>")));
    RUN(rightcol->put_PaneXtra(0,1));
    RUN(rightcol->put_PaneXtra(1,1));
    RUN(rightcol->put_PaneXtra(2,1));
    RUN(rightcol->put_PaneXtra(3,1));

    // the title bars
    int mainWid=50,searchWid=70,patternWid=90,inputWid=55,propWid=100; // overridden by PaneXtra
    RUN(runcaptionspot.CoCreateInstance(CLSID_MCCPane));
    RUN(runchoicespot.CoCreateInstance(CLSID_MCCPane));
    RUN(maintitlespot->AddPane(0,runcaptionspot));
    RUN(maintitlespot->AddPane(1,runchoicespot));
    RUN(maintitlespot->put_PaneSize(0,mainWid));
    RUN(maintitlespot->put_PaneXtra(0,1));
    RUN(maintitlespot->put_PaneSize(1,combowidth));
    RUN(mb.c->new_Site(&runchoicesite));
    RUN(runchoicesite->put_Layer(topL));
    RUN(runchoicectl.CoCreateInstance(CLSID_MCCComboBox));
    RUN(runchoicesite->put_Control(runchoicectl));
    RUN(runchoicespot->put_Content(runchoicesite));
    RUN(runchoicesite->put_Intent(MCCS_ACTIVE));

    RUN(searchcaptionspot.CoCreateInstance(CLSID_MCCPane));
    RUN(searchchoicespot.CoCreateInstance(CLSID_MCCPane));
    RUN(searchsavespot.CoCreateInstance(CLSID_MCCPane));
    RUN(searchtitlespot->AddPane(0,searchcaptionspot));
    RUN(searchtitlespot->AddPane(1,searchchoicespot));
    RUN(searchtitlespot->AddPane(2,searchsavespot));
    RUN(searchtitlespot->put_PaneSize(0,searchWid));
    RUN(searchtitlespot->put_PaneXtra(0,1));
    RUN(searchtitlespot->put_PaneSize(1,combowidth));
    RUN(searchtitlespot->put_PaneSize(2,50));
    RUN(mb.c->new_Site(&searchchoicesite));
    RUN(mb.c->new_Site(&searchsavesite));
    RUN(searchchoicesite->put_Layer(topL));
    RUN(searchsavesite->put_Layer(topL));
    RUN(searchchoicectl.CoCreateInstance(CLSID_MCCComboBox));
    RUN(searchsavectl.CoCreateInstance(CLSID_MCCButton));
    RUN(searchsavectl->put_Text(CComBSTR("Save")));
    RUN(searchchoicesite->put_Control(searchchoicectl));
    RUN(searchsavesite->put_Control(searchsavectl));
    RUN(searchchoicespot->put_Content(searchchoicesite));
    RUN(searchsavespot->put_Content(searchsavesite));
    RUN(searchchoicesite->put_Intent(MCCS_ACTIVE));
    RUN(searchsavesite->put_Intent(MCCS_ACTIVE));

    RUN(searchtextcheckctl.CoCreateInstance(CLSID_MCCButton));
    RUN(searchtextcheckctl->put_WStyle(BS_AUTOCHECKBOX));
    RUN(searchtextcheckctl->put_Text(CComBSTR("text")));
    RUN(searchtextcheckspot.CoCreateInstance(CLSID_MCCPane));
    RUN(searchtitlespot->AddPane(3,searchtextcheckspot));
    RUN(searchtitlespot->put_PaneSize(3,50));
    RUN(mb.c->new_Site(&searchtextchecksite));
    RUN(searchtextchecksite->put_Layer(topL));
    RUN(searchtextchecksite->put_Control(searchtextcheckctl));
    RUN(searchtextcheckspot->put_Content(searchtextchecksite));
    RUN(searchtextchecksite->put_Intent(MCCS_ACTIVE));

    RUN(searchtextctl.CoCreateInstance(CLSID_MCCEdit));
    RUN(mb.c->new_Site(&searchtextsite));
    RUN(searchtextsite->put_Layer(topL));
    RUN(searchtextsite->put_Control(searchtextctl));
    RUN(searchtextsite->put_Intent(MCCS_HIDDEN));

    RUN(patterncaptionspot.CoCreateInstance(CLSID_MCCPane));
    RUN(patternchoicespot.CoCreateInstance(CLSID_MCCPane));
    RUN(patternsavespot.CoCreateInstance(CLSID_MCCPane));
    RUN(patterntitlespot->AddPane(0,patterncaptionspot));
    RUN(patterntitlespot->AddPane(1,patternchoicespot));
    RUN(patterntitlespot->AddPane(2,patternsavespot));
    RUN(patterntitlespot->put_PaneSize(0,patternWid));
    RUN(patterntitlespot->put_PaneXtra(0,1));
    RUN(patterntitlespot->put_PaneSize(1,combowidth));
    RUN(patterntitlespot->put_PaneSize(2,50));
    RUN(mb.c->new_Site(&patternchoicesite));
    RUN(mb.c->new_Site(&patternsavesite));
    RUN(patternchoicesite->put_Layer(topL));
    RUN(patternsavesite->put_Layer(topL));
    RUN(patternchoicectl.CoCreateInstance(CLSID_MCCComboBox));
    RUN(patternsavectl.CoCreateInstance(CLSID_MCCButton));
    RUN(patternsavectl->put_Text(CComBSTR("Save")));
    RUN(patternchoicesite->put_Control(patternchoicectl));
    RUN(patternsavesite->put_Control(patternsavectl));
    RUN(patternchoicespot->put_Content(patternchoicesite));
    RUN(patternsavespot->put_Content(patternsavesite));
    RUN(patternchoicesite->put_Intent(MCCS_ACTIVE));
    RUN(patternsavesite->put_Intent(MCCS_ACTIVE));

    RUN(patterntextcheckctl.CoCreateInstance(CLSID_MCCButton));
    RUN(patterntextcheckctl->put_WStyle(BS_AUTOCHECKBOX));
    RUN(patterntextcheckctl->put_Text(CComBSTR("text")));
    RUN(patterntextcheckspot.CoCreateInstance(CLSID_MCCPane));
    RUN(patterntitlespot->AddPane(3,patterntextcheckspot));
    RUN(patterntitlespot->put_PaneSize(3,50));
    RUN(mb.c->new_Site(&patterntextchecksite));
    RUN(patterntextchecksite->put_Layer(topL));
    RUN(patterntextchecksite->put_Control(patterntextcheckctl));
    RUN(patterntextcheckspot->put_Content(patterntextchecksite));
    RUN(patterntextchecksite->put_Intent(MCCS_ACTIVE));

    RUN(patterntextctl.CoCreateInstance(CLSID_MCCEdit));
    RUN(mb.c->new_Site(&patterntextsite));
    RUN(patterntextsite->put_Layer(topL));
    RUN(patterntextsite->put_Control(patterntextctl));
    RUN(patterntextsite->put_Intent(MCCS_HIDDEN));

    RUN(inputcaptionspot.CoCreateInstance(CLSID_MCCPane));
    RUN(inputchoicespot.CoCreateInstance(CLSID_MCCPane));
    RUN(inputsavespot.CoCreateInstance(CLSID_MCCPane));
    RUN(inputgrabspot.CoCreateInstance(CLSID_MCCPane));
    RUN(inputshowspot.CoCreateInstance(CLSID_MCCPane));
    RUN(inputtitlespot->AddPane(0,inputcaptionspot));
    RUN(inputtitlespot->AddPane(1,inputchoicespot));
    RUN(inputtitlespot->AddPane(2,inputsavespot));
    RUN(inputtitlespot->AddPane(3,inputgrabspot));
    RUN(inputtitlespot->AddPane(4,inputshowspot));
    RUN(inputtitlespot->put_PaneSize(0,inputWid));
    RUN(inputtitlespot->put_PaneXtra(0,1));
    RUN(inputtitlespot->put_PaneSize(1,combowidth));
    RUN(inputtitlespot->put_PaneSize(2,70));
    RUN(inputtitlespot->put_PaneSize(3,70));
    RUN(inputtitlespot->put_PaneSize(4,70));
    RUN(mb.c->new_Site(&inputchoicesite));
    RUN(mb.c->new_Site(&inputsavesite));
    RUN(mb.c->new_Site(&inputgrabsite));
    RUN(mb.c->new_Site(&inputshowsite));
    RUN(inputchoicesite->put_Layer(topL));
    RUN(inputsavesite->put_Layer(topL));
    RUN(inputgrabsite->put_Layer(topL));
    RUN(inputshowsite->put_Layer(topL));
    RUN(inputchoicectl.CoCreateInstance(CLSID_MCCComboBox));
    RUN(inputsavectl.CoCreateInstance(CLSID_MCCStickyButton));
    RUN(inputsavectl->put_Text(CComBSTR("Save")));
    RUN(inputgrabctl.CoCreateInstance(CLSID_MCCStickyButton));
    RUN(inputgrabctl->put_Text(CComBSTR("Grab")));
    RUN(inputshowctl.CoCreateInstance(CLSID_MCCStickyButton));
    RUN(inputshowctl->put_Text(CComBSTR("Show")));
    RUN(inputchoicesite->put_Control(inputchoicectl));
    RUN(inputsavesite->put_Control(inputsavectl));
    RUN(inputgrabsite->put_Control(inputgrabctl));
    RUN(inputshowsite->put_Control(inputshowctl));
    RUN(inputchoicespot->put_Content(inputchoicesite));
    RUN(inputsavespot->put_Content(inputsavesite));
    RUN(inputgrabspot->put_Content(inputgrabsite));
    RUN(inputshowspot->put_Content(inputshowsite));
    RUN(inputchoicesite->put_Intent(MCCS_ACTIVE));
    RUN(inputsavesite->put_Intent(MCCS_ACTIVE));
    RUN(inputgrabsite->put_Intent(MCCS_ACTIVE));
    RUN(inputshowsite->put_Intent(MCCS_ACTIVE));

    RUN(inputtextcheckctl.CoCreateInstance(CLSID_MCCButton));
    RUN(inputtextcheckctl->put_WStyle(BS_AUTOCHECKBOX));
    RUN(inputtextcheckctl->put_Text(CComBSTR("text")));
    RUN(inputtextcheckspot.CoCreateInstance(CLSID_MCCPane));
    RUN(inputtitlespot->AddPane(5,inputtextcheckspot));
    RUN(inputtitlespot->put_PaneSize(5,50));
    RUN(mb.c->new_Site(&inputtextchecksite));
    RUN(inputtextchecksite->put_Layer(topL));
    RUN(inputtextchecksite->put_Control(inputtextcheckctl));
    RUN(inputtextcheckspot->put_Content(inputtextchecksite));
    RUN(inputtextchecksite->put_Intent(MCCS_ACTIVE));

    RUN(inputtextctl.CoCreateInstance(CLSID_MCCEdit));
    RUN(mb.c->new_Site(&inputtextsite));
    RUN(inputtextsite->put_Layer(topL));
    RUN(inputtextsite->put_Control(inputtextctl));
    RUN(inputtextsite->put_Intent(MCCS_HIDDEN));

    RUN(propcaptionspot.CoCreateInstance(CLSID_MCCPane));
    RUN(propscopespot.CoCreateInstance(CLSID_MCCPane));
    RUN(proptypespot.CoCreateInstance(CLSID_MCCPane));
    RUN(propsavespot.CoCreateInstance(CLSID_MCCPane));
    RUN(proptitlespot->AddPane(0,propcaptionspot));
    RUN(proptitlespot->AddPane(1,propscopespot));
    RUN(proptitlespot->AddPane(2,proptypespot));
    RUN(proptitlespot->AddPane(3,propsavespot));
    RUN(proptitlespot->put_PaneSize(0,propWid));
    RUN(proptitlespot->put_PaneXtra(0,1));
    RUN(proptitlespot->put_PaneSize(1,95));
    RUN(proptitlespot->put_PaneSize(2,136));
    RUN(proptitlespot->put_PaneSize(3,50));
    RUN(mb.c->new_Site(&propscopesite));
    RUN(mb.c->new_Site(&proptypesite));
    RUN(mb.c->new_Site(&propsavesite));
    RUN(propscopesite->put_Layer(topL));
    RUN(proptypesite->put_Layer(topL));
    RUN(propsavesite->put_Layer(topL));
    RUN(propscopectl.CoCreateInstance(CLSID_MCCComboBox));
    RUN(proptypectl.CoCreateInstance(CLSID_MCCComboBox));
    RUN(propsavectl.CoCreateInstance(CLSID_MCCButton));
    RUN(propsavectl->put_Text(CComBSTR("Save")));
    RUN(propscopesite->put_Control(propscopectl));
    RUN(proptypesite->put_Control(proptypectl));
    RUN(propsavesite->put_Control(propsavectl));
    RUN(propscopespot->put_Content(propscopesite));
    RUN(proptypespot->put_Content(proptypesite));
    RUN(propsavespot->put_Content(propsavesite));
    RUN(propscopesite->put_Intent(MCCS_ACTIVE));
    RUN(proptypesite->put_Intent(MCCS_ACTIVE));
    RUN(propsavesite->put_Intent(MCCS_ACTIVE));

	RUN(proptypectl->put_WStyle(CBS_DROPDOWNLIST|CBS_SORT));
    RUN(proptypectl->Add(CComBSTR("nodes&edges")));
    RUN(proptypectl->Add(CComBSTR("nodes")));
    RUN(proptypectl->Add(CComBSTR("edges")));
    RUN(proptypectl->Choose(CComBSTR("nodes&edges")));

	RUN(propscopectl->put_WStyle(CBS_DROPDOWNLIST|CBS_SORT));
    RUN(propscopectl->Add(CComBSTR("abstract")));
    RUN(propscopectl->Add(CComBSTR("layout")));
    RUN(propscopectl->Choose(CComBSTR("abstract")));

    // this may not be adequate
    RUN(columns->DoZ(0));

    RUN(mainS->put_Intent(MCCS_UIACTIVE));
    RUN(searchS->put_Intent(MCCS_UIACTIVE));
    RUN(patternS->put_Intent(MCCS_UIACTIVE));
    RUN(inputS->put_Intent(MCCS_UIACTIVE));

    // sinks for all that
    CComObject<ComboSink> *chooseSearch = new CComObject<ComboSink>,
        *choosePattern = new CComObject<ComboSink>,
        *chooseInput = new CComObject<ComboSink>,
        *chooseRun = new CComObject<ComboSink>;
    chooseSearch->m_dino = choosePattern->m_dino = chooseInput->m_dino = chooseRun->m_dino = this;
    chooseSearch->m_id = DINOGRAPH_INPUT_SEARCH;
    choosePattern->m_id = DINOGRAPH_INPUT_PATTERN;
    chooseInput->m_id = DINOGRAPH_INPUT_START;
    chooseRun->m_id = MAIN_WINDOW;
    RUN(AtlAdvise(searchchoicectl,chooseSearch,IID_IMCCComboBoxEvents,&compost));
    RUN(AtlAdvise(patternchoicectl,choosePattern,IID_IMCCComboBoxEvents,&compost));
    RUN(AtlAdvise(inputchoicectl,chooseInput,IID_IMCCComboBoxEvents,&compost));
    RUN(AtlAdvise(runchoicectl,chooseRun,IID_IMCCComboBoxEvents,&compost));

    CComObject<SaveButtonSink> *saveSearch = new CComObject<SaveButtonSink>,
        *savePattern = new CComObject<SaveButtonSink>;
    saveSearch->m_dino = savePattern->m_dino = this;
    saveSearch->m_id = DINOGRAPH_INPUT_SEARCH;
    savePattern->m_id = DINOGRAPH_INPUT_PATTERN;
    RUN(AtlAdvise(searchsavectl,saveSearch,IID_IMCCButtonEvents,&compost));
    RUN(AtlAdvise(patternsavectl,savePattern,IID_IMCCButtonEvents,&compost));

    CComObject<StickySaveSink> *saveInput = new CComObject<StickySaveSink>;
    saveInput->m_dino = this;
    RUN(AtlAdvise(inputsavectl,saveInput,IID_IMCCStickyButtonEvents,&compost));

    CComObject<TextCheckSink> *checkTextSearch = new CComObject<TextCheckSink>;
    checkTextSearch->m_dino = this;
    checkTextSearch->m_id = DINOGRAPH_INPUT_SEARCH;
    checkTextSearch->m_check = searchtextcheckctl;
    RUN(AtlAdvise(searchtextcheckctl,checkTextSearch,IID_IMCCButtonEvents,&compost));

    CComObject<TextCheckSink> *checkTextPattern = new CComObject<TextCheckSink>;
    checkTextPattern->m_dino = this;
    checkTextPattern->m_id = DINOGRAPH_INPUT_PATTERN;
    checkTextPattern->m_check = patterntextcheckctl;
    RUN(AtlAdvise(patterntextcheckctl,checkTextPattern,IID_IMCCButtonEvents,&compost));

    CComObject<TextCheckSink> *checkTextInput = new CComObject<TextCheckSink>;
    checkTextInput->m_dino = this;
    checkTextInput->m_id = DINOGRAPH_INPUT_START;
    checkTextInput->m_check = inputtextcheckctl;
    RUN(AtlAdvise(inputtextcheckctl,checkTextInput,IID_IMCCButtonEvents,&compost));

    CComObject<GrabSink> *grabInput = new CComObject<GrabSink>;
    grabInput->m_dino = this;
    RUN(AtlAdvise(inputgrabctl,grabInput,IID_IMCCStickyButtonEvents,&compost));

    CComObject<ShowSink> *showInput = new CComObject<ShowSink>;
    showInput->m_dino = this;
    RUN(AtlAdvise(inputshowctl,showInput,IID_IMCCStickyButtonEvents,&compost));

	CComBSTR blank("");
	RUN(patternchoicectl->Add(blank));
	RUN(searchchoicectl->Add(blank));
	RUN(inputchoicectl->Add(blank));
	RUN(runchoicectl->Add(blank));
	RUN(patternchoicectl->put_WStyle(CBS_DROPDOWN|CBS_SORT));
	RUN(searchchoicectl->put_WStyle(CBS_DROPDOWN|CBS_SORT));
	RUN(inputchoicectl->put_WStyle(CBS_DROPDOWN|CBS_SORT));
	RUN(runchoicectl->put_WStyle(CBS_DROPDOWNLIST|CBS_SORT));

    CComObject<PropTypeSink> *propType = new CComObject<PropTypeSink>;
    propType->m_dino = this;
    RUN(AtlAdvise(proptypectl,propType,IID_IMCCComboBoxEvents,&compost));

    CComObject<PropScopeSink> *propScope = new CComObject<PropScopeSink>;
    propScope->m_dino = this;
    RUN(AtlAdvise(propscopectl,propScope,IID_IMCCComboBoxEvents,&compost));

    CComObject<SaveButtonSink> *saveProps = new CComObject<SaveButtonSink>;
    saveProps->m_dino = this;
    saveProps->m_id = PROP_WINDOW;
    RUN(AtlAdvise(propsavectl,saveProps,IID_IMCCButtonEvents,&compost));

    // layouts and their abstract graphs

    CComBSTR mainLayoutS("main_layout"),searchLayoutS("search_layout"),patternLayoutS("pattern_layout"),inputLayoutS("input_layout"),
        mainAbstractS("main_abstract"),searchAbstractS("search_abstract"),patternAbstractS("pattern_abstract"),inputAbstractS("input_abstract");
    ItemMoniker mkEngine(L"Graph Layout Engine");
	RUN(mkEngine->BindToObject(mb.bctx,PointerMoniker(main),riidret(&mainLayout)));
    RUN(CComQIPtr<IDGStringAttributes>(mainLayout)->put_Name(mainLayoutS));
	RUN(mkEngine->BindToObject(mb.bctx,PointerMoniker(search),riidret(&searchLayout)));
    RUN(CComQIPtr<IDGStringAttributes>(searchLayout)->put_Name(searchLayoutS));
	RUN(mkEngine->BindToObject(mb.bctx,PointerMoniker(pattern),riidret(&patternLayout)));
    RUN(CComQIPtr<IDGStringAttributes>(patternLayout)->put_Name(patternLayoutS));
	RUN(mkEngine->BindToObject(mb.bctx,PointerMoniker(input),riidret(&inputLayout)));
    RUN(CComQIPtr<IDGStringAttributes>(inputLayout)->put_Name(inputLayoutS));

    RUN(mainAbstract.CoCreateInstance(CLSID_DynagraphAbstractGraph));
    RUN(CComQIPtr<IDGStringAttributes>(mainAbstract)->put_Name(mainAbstractS));
    RUN(searchAbstract.CoCreateInstance(CLSID_DynagraphAbstractGraph));
    RUN(CComQIPtr<IDGStringAttributes>(searchAbstract)->put_Name(searchAbstractS));
    RUN(patternAbstract.CoCreateInstance(CLSID_DynagraphAbstractGraph));
    RUN(CComQIPtr<IDGStringAttributes>(patternAbstract)->put_Name(patternAbstractS));
    RUN(inputAbstract.CoCreateInstance(CLSID_DynagraphAbstractGraph));
    RUN(CComQIPtr<IDGStringAttributes>(inputAbstract)->put_Name(inputAbstractS));

    RUN(dinoMachine.CoCreateInstance(CLSID_DinographMachine));
    CComPtr<IDGNode> mainLayoutMN,searchLayoutMN,patternLayoutMN,inputLayoutMN,
        mainAbstractMN,searchAbstractMN,patternAbstractMN,inputAbstractMN;
    RUN(dinoMachine->get_Node(mainLayoutS,&mainLayoutMN));
    RUN(dinoMachine->get_Node(searchLayoutS,&searchLayoutMN));
    RUN(dinoMachine->get_Node(patternLayoutS,&patternLayoutMN));
    RUN(dinoMachine->get_Node(inputLayoutS,&inputLayoutMN));
    RUN(dinoMachine->get_Node(mainAbstractS,&mainAbstractMN));
    RUN(dinoMachine->get_Node(searchAbstractS,&searchAbstractMN));
    RUN(dinoMachine->get_Node(patternAbstractS,&patternAbstractMN));
    RUN(dinoMachine->get_Node(inputAbstractS,&inputAbstractMN));

    // "in" to dynagraph layout; "out" to abstract graph
    CComPtr<IDGEdge> mainMapIn,searchMapIn,patternMapIn,inputMapIn;
    RUN(dinoMachine->new_Edge(CComBSTR("main-map-in"),mainAbstractMN,mainLayoutMN,&mainMapIn));
    RUN(dinoMachine->new_Edge(CComBSTR("search-map-in"),searchAbstractMN,searchLayoutMN,&searchMapIn));
    RUN(dinoMachine->new_Edge(CComBSTR("pattern-map-in"),patternAbstractMN,patternLayoutMN,&patternMapIn));
    RUN(dinoMachine->new_Edge(CComBSTR("input-map-in"),inputAbstractMN,inputLayoutMN,&inputMapIn));

    CComPtr<IDGEdge> mainMapOut,searchMapOut,patternMapOut,inputMapOut;
    RUN(dinoMachine->new_Edge(CComBSTR("main-map-out"),mainLayoutMN,mainAbstractMN,&mainMapOut));
    RUN(dinoMachine->new_Edge(CComBSTR("search-map-out"),searchLayoutMN,searchAbstractMN,&searchMapOut));
    RUN(dinoMachine->new_Edge(CComBSTR("pattern-map-out"),patternLayoutMN,patternAbstractMN,&patternMapOut));
    RUN(dinoMachine->new_Edge(CComBSTR("input-map-out"),inputLayoutMN,inputAbstractMN,&inputMapOut));

    RUN(CComQIPtr<IDGStringAttributes>(mainMapOut)->SetAttr(CComBSTR("mapping"),CComBSTR("label2name")));
    RUN(CComQIPtr<IDGStringAttributes>(searchMapOut)->SetAttr(CComBSTR("mapping"),CComBSTR("label2name,dinosearchOut")));
    RUN(CComQIPtr<IDGStringAttributes>(searchMapOut)->SetAttr(CComBSTR("l2n_edgelabels"),CComBSTR("false")));
    RUN(CComQIPtr<IDGStringAttributes>(patternMapOut)->SetAttr(CComBSTR("mapping"),CComBSTR("label2name,dinopatternOut")));
    RUN(CComQIPtr<IDGStringAttributes>(patternMapOut)->SetAttr(CComBSTR("l2n_edgelabels"),CComBSTR("false")));
    RUN(CComQIPtr<IDGStringAttributes>(inputMapOut)->SetAttr(CComBSTR("mapping"),CComBSTR("label2name")));

    RUN(CComQIPtr<IDGStringAttributes>(mainMapIn)->SetAttr(CComBSTR("mapping"),CComBSTR("matchnames")));
    RUN(CComQIPtr<IDGStringAttributes>(searchMapIn)->SetAttr(CComBSTR("mapping"),CComBSTR("matchnames,dinosearchIn")));
    RUN(CComQIPtr<IDGStringAttributes>(patternMapIn)->SetAttr(CComBSTR("mapping"),CComBSTR("matchnames,dinopatternIn")));
    RUN(CComQIPtr<IDGStringAttributes>(inputMapIn)->SetAttr(CComBSTR("mapping"),CComBSTR("matchnames")));

    ItemMoniker mkSelection(L"Selection");
	RUN(mkSelection->BindToObject(mb.bctx,PointerMoniker(main),riidret(&mainSelection)));
	RUN(mkSelection->BindToObject(mb.bctx,PointerMoniker(search),riidret(&searchSelection)));
	RUN(mkSelection->BindToObject(mb.bctx,PointerMoniker(pattern),riidret(&patternSelection)));
	RUN(mkSelection->BindToObject(mb.bctx,PointerMoniker(input),riidret(&inputSelection)));

    CComObject<SelectionSink> *watchMainSelection = new CComObject<SelectionSink>;
    watchMainSelection->m_dino = this;
    watchMainSelection->m_id = MAIN_WINDOW;
    RUN(AtlAdvise(mainSelection,watchMainSelection,IID_IItemSetEvents,&compost));

    CComObject<SelectionSink> *watchSearchSelection = new CComObject<SelectionSink>;
    watchSearchSelection->m_dino = this;
    watchSearchSelection->m_id = DINOGRAPH_INPUT_SEARCH;
    RUN(AtlAdvise(searchSelection,watchSearchSelection,IID_IItemSetEvents,&compost));

    CComObject<SelectionSink> *watchPatternSelection = new CComObject<SelectionSink>;
    watchPatternSelection->m_dino = this;
    watchPatternSelection->m_id = DINOGRAPH_INPUT_PATTERN;
    RUN(AtlAdvise(patternSelection,watchPatternSelection,IID_IItemSetEvents,&compost));

    CComObject<SelectionSink> *watchInputSelection = new CComObject<SelectionSink>;
    watchInputSelection->m_dino = this;
    watchInputSelection->m_id = DINOGRAPH_INPUT_START;
    RUN(AtlAdvise(inputSelection,watchInputSelection,IID_IItemSetEvents,&compost));

    CComBSTR name;
	RUN(CComQIPtr<IDGStringAttributes>(mainAbstract)->get_Name(&name));
    char buf[100];
    int len = sprintf(buf,"open graph %S\n",name.m_str);
    RUN(incrClient->Write(buf,len));
    return S_OK;
}
string DinographUI::readGraph(IDGGraph *source) {
    HGLOBAL glob;
    RUN(source->SaveDotfile(&glob));
    istringstream is((char *)GlobalLock(glob));
    GlobalFree(glob);
    string text;
    is >> must("digraph") >> string();
    is >> match('{','}',text);
    return text;
}
HRESULT DinographUI::sendGraph(LPCOLESTR name,LPCOLESTR text,IDGGraph *destDiagram) {
	HGLOBAL glob = GlobalAlloc(GHND,(text?wcslen(text):0)+wcslen(name)+20);
    char *txt2 = (char*)GlobalLock(glob);
	if(text)
		sprintf(txt2,"digraph %S {%S}",name,text);
	else
		sprintf(txt2,"digraph G {}");
    GlobalUnlock(glob);
    RUN(destDiagram->LoadDotfile(glob,true,true));
    GlobalFree(glob);
    return S_OK;
}
HRESULT DinographUI::save(int id) {
    bool usetext = false;
    CComPtr<IDGGraph> sourceGraph;
    CComPtr<IMCCEdit> sourceText;
    CComBSTR name;
    switch(id) {
    case DINOGRAPH_INPUT_SEARCH:
        sourceGraph = searchAbstract;
        sourceText = searchtextctl;
        usetext = searchusetext;
        RUN(searchchoicectl->Get(&name));
        break;
    case DINOGRAPH_INPUT_PATTERN:
        sourceGraph = patternAbstract;
        sourceText = patterntextctl;
        usetext = patternusetext;
        RUN(patternchoicectl->Get(&name));
        break;
    case DINOGRAPH_INPUT_START:
        sourceGraph = inputAbstract;
        sourceText = inputtextctl;
        usetext = inputusetext;
        RUN(inputchoicectl->Get(&name));
        break;
    case PROP_WINDOW:
        return saveProperties();
    }
	if(name.Length()==0)
		return S_FALSE;
    if(usetext) {
        CComBSTR text;
        RUN(sourceText->get_Text(&text));
        RUN(m_central->ChangeInput((DINOGRAPH_INPUT_TYPES)id,name,text));
    }
    else {
        string text = readGraph(sourceGraph);
        USES_CONVERSION;
        RUN(m_central->ChangeInput((DINOGRAPH_INPUT_TYPES)id,name,A2OLE(text.c_str())));
    }
	return S_OK;
}
HRESULT DinographUI::selected(int id,BSTR val) {
    CComPtr<IDGGraph> destGraph;
    CComPtr<IMCCEdit> destText;
    bool usetext = false;
    switch(id) {
    case DINOGRAPH_INPUT_SEARCH:
        destGraph = searchAbstract;
        destText = searchtextctl;
        usetext = searchusetext;
        break;
    case DINOGRAPH_INPUT_PATTERN:
        destGraph = patternAbstract;
        destText = patterntextctl;
        usetext = patternusetext;
        break;
    case DINOGRAPH_INPUT_START:
        destGraph = inputAbstract;
        destText = inputtextctl;
        usetext = inputusetext;
        break;
    case MAIN_WINDOW: {
        LPCOLESTR chosen = val;
        if(!chosen[0])
            chosen = L"\"\"";
        CComBSTR name;
	    RUN(CComQIPtr<IDGStringAttributes>(mainAbstract)->get_Name(&name));
        char buf[200];
        int len = sprintf(buf,"modify graph %S [searchname=%S]\n",name.m_str,chosen);
        RUN(m_incrClient->Write(buf,len));
        return S_OK;
    }
    }
    LPOLESTR text=0;
	if(val[0])
		RUN(m_central->GetInput((DINOGRAPH_INPUT_TYPES)id,val,&text));

    if(usetext)
        RUN(destText->put_Text(CComBSTR(text)));
    else
        RUN(sendGraph(val,text,destGraph));
    if(text)
		CoTaskMemFree(text);
	return S_OK;
}
HRESULT DinographUI::textCheck(int id,bool val) {
    CComPtr<IDGGraph> graph;
    CComPtr<IMCCSite> diagramS,textS;
    CComPtr<IMCCEdit> text;
    CComPtr<IMCCPane> spot;
    CComPtr<IMCCComboBox> choice;
    switch(id) {
    case DINOGRAPH_INPUT_SEARCH:
        diagramS = searchS;
        spot = searchspot;
        textS = searchtextsite;
        text = searchtextctl;
        graph = searchAbstract;
        choice = searchchoicectl;
        searchusetext = val;
        break;
    case DINOGRAPH_INPUT_PATTERN:
        diagramS = patternS;
        spot = patternspot;
        textS = patterntextsite;
        text = patterntextctl;
        graph = patternAbstract;
        choice = patternchoicectl;
        patternusetext = val;
        break;
    case DINOGRAPH_INPUT_START:
        diagramS = inputS;
        spot = inputspot;
        textS = inputtextsite;
        text = inputtextctl;
        graph = inputAbstract;
        choice = inputchoicectl;
        inputusetext = val;
        break;
    }
    if(val) {
        string asText = readGraph(graph);
        RUN(text->put_Text(CComBSTR(asText.c_str())));
        RUN(diagramS->put_Intent(MCCS_HIDDEN));
        RUN(spot->put_Content(textS));
        RUN(textS->put_Intent(MCCS_ACTIVE));
    }
    else {
        CComBSTR name;
        RUN(choice->Get(&name));
        CComBSTR asText;
        RUN(text->get_Text(&asText));
        RUN(sendGraph(name,asText,graph));
        RUN(textS->put_Intent(MCCS_HIDDEN));
        RUN(spot->put_Content(diagramS));
        RUN(diagramS->put_Intent(MCCS_UIACTIVE));
    }
    return S_OK;
}
CComPtr<IMontageContainer> DinographUI::activeMC() {
    CComPtr<IOleClientSite> frameSite;
    RUN(CComQIPtr<IOleObject>(topMC)->GetClientSite(&frameSite));
    CComPtr<IOleInPlaceActiveObject> activeO;
    // I happen to know of a special interface..  um
    RUN(CComQIPtr<IMCCFrameSharer>(frameSite)->get_ActiveObject(&activeO));
    return CComQIPtr<IMontageContainer>(activeO);
}
int DinographUI::getActive() {
    CComPtr<IMontageContainer> acmc = activeMC();
    if(!acmc)
        return -1;
    CComPtr<IItemSet> sel;
    if(acmc==main)
        return MAIN_WINDOW;
    else if(acmc==search)
        return DINOGRAPH_INPUT_SEARCH;
    else if(acmc==pattern)
        return DINOGRAPH_INPUT_PATTERN;
    else if(acmc==input)
        return DINOGRAPH_INPUT_START;
    else return -1;
}
CComPtr<IItemSet> DinographUI::activeSelection() {
    CComPtr<IItemSet> sel;
    switch(getActive()) {
    case MAIN_WINDOW:
        sel = mainSelection;
        break;
    case DINOGRAPH_INPUT_SEARCH:
        sel = searchSelection;
        break;
    case DINOGRAPH_INPUT_PATTERN:
        sel = patternSelection;
        break;
    case DINOGRAPH_INPUT_START:
        sel = inputSelection;
        break;
    }
    return sel;
}
CComPtr<IDGGraph> getGraph(IUnknown *u) {
    CComPtr<IDGGraph> g;
    if(CComQIPtr<IDGNode> n = u)
        SRUN(n->get_Graph(&g));
    else if(CComQIPtr<IDGEdge> e = u)
        SRUN(e->get_Graph(&g));
    return g;
}
bool DinographUI::filterThing(IUnknown *u) {
    if(CComQIPtr<IDGNode> n = u)
        return propNodes;
    else if(CComQIPtr<IDGEdge> e = u)
        return propEdges;
    return false;
}
HRESULT DinographUI::getSelected(vector<IDGStringAttributes*> &ret) {
	CComPtr<IEnumMoniker> enumk;
	RUN(activeSelection()->Enum(&enumk));
	// don't accept none
	mk_iterator mi(enumk);
	if(mi==mk_iterator())
		return S_FALSE;
	BindCtx bctx;
	PointerMoniker mkMC(activeMC());
    ret.clear();
    CComQIPtr<IDinoEdge> xlator;
	for(; mi!=mk_iterator(); ++mi) {
		CComPtr<DIPersistor> mp;
		RUN((*mi)->BindToObject(bctx,mkMC,riidret(&mp)));
		CComPtr<IUnknown> uS;
		RUN(mp->get_Subject(&uS));
		CComPtr<IUnknown> lo;
		RUN(layoutFromSite(uS,&lo));
        if(propAbstract) {
            if(!xlator) {
                CComPtr<IDGGraph> source = getGraph(lo),dinoMachine;
                RUN(dinoMachine.CoCreateInstance(CLSID_DinographMachine));
                CComBSTR graphName;
                RUN(CComQIPtr<IDGStringAttributes>(source)->get_Name(&graphName));
                CComPtr<IDGNode> dinon;
                RUN(dinoMachine->get_Node(graphName,&dinon));
                CComPtr<IEnumDGEdges> outs;
                RUN(dinon->get_OutEdges(&outs));
                CComPtr<IDGEdge> dinoe;
                RUN(outs->Next(1,&dinoe,0));
                xlator = dinoe;
            }
            CComPtr<IEnumUnknown> eu;
            RUN(xlator->Map(lo,&eu));
            for(unk_iterator ui(eu);ui!=unk_iterator();++ui)
                if(filterThing(*ui)) {
                    ret.push_back(0);
                    RUN((*ui)->QueryInterface(&ret.back()));
                }
        }
        else {
            if(filterThing(lo)) {
                ret.push_back(0);
                RUN(lo.QueryInterface(&ret.back()));
            }
        }
    }
    return S_OK;
}
HRESULT DinographUI::refreshProperties() {
    vector<IDGStringAttributes*> sources;
    RUN(getSelected(sources));
    RUN(proppane->RemoveProperties());
    // this uses lots of OLESTRs for BSTR params but I happen to know...
    for(vector<IDGStringAttributes*>::iterator i = sources.begin(); i!=sources.end(); ++i) {
        CComPtr<IEnumString> ena;
        RUN((*i)->EnumAttrs(&ena));
        for(olestr_iterator si(ena); si!=olestr_iterator(); ++si) {
            CComBSTR val;
            RUN((*i)->GetAttr(*si,&val));
            RUN(proppane->AddValue(*si,val));
            int nv;
            RUN(proppane->CountValues(*si,&nv));
            if(nv==1)
                RUN(proppane->SetSelected(*si,val));
            else
                RUN(proppane->SetSelected(*si,0));
        }
    }
    for(vector<IDGStringAttributes*>::iterator i = sources.begin(); i!=sources.end(); ++i)
        (*i)->Release();
    return S_OK;
}
HRESULT DinographUI::saveProperties() {
    vector<IDGStringAttributes*> sources;
    RUN(getSelected(sources));
    if(!sources.size())
        return S_FALSE;
    CComPtr<IDGGraph> g = getGraph(*sources.begin());
    RUN(g->GraphLock());
    CComPtr<IEnumString> numster;
    RUN(proppane->EnumProperties(&numster));
    for(olestr_iterator si(numster); si!=olestr_iterator(); ++si) {
        CComBSTR val;
        RUN(proppane->GetSelected(*si,&val));
        if(val.Length())
            for(vector<IDGStringAttributes*>::iterator i = sources.begin(); i!=sources.end(); ++i)
                RUN((*i)->SetAttr(*si,val));

    }
    for(vector<IDGStringAttributes*>::iterator i = sources.begin(); i!=sources.end(); ++i)
        (*i)->Release();
    RUN(g->GraphUnlock());
    return S_OK;
}
HRESULT DinographUI::selectChange(int id) {
    if(id==MAIN_WINDOW) {
        if(alwaysGrab)
            RUN(grab());
    }
    if(getActive()==id)
        RUN(refreshProperties());
	return S_OK;
}
HRESULT DinographUI::grab() {
	char buf[9000], *where=buf;
	*where = 0; // in case nothing
	CComPtr<IEnumMoniker> enumk;
	RUN(mainSelection->Enum(&enumk));
	// don't accept none
	mk_iterator mi(enumk);
	if(mi==mk_iterator())
		return S_FALSE;
	BindCtx bctx;
	PointerMoniker mkMC(main);
	for(; mi!=mk_iterator(); ++mi) {
		CComPtr<DIPersistor> mp;
		RUN((*mi)->BindToObject(bctx,mkMC,riidret(&mp)));
		CComPtr<IUnknown> uS;
		RUN(mp->get_Subject(&uS));
		CComPtr<IUnknown> uE;
		RUN(layoutFromSite(uS,&uE));
		if(CComQIPtr<IDGNode> n = uE) {
			CComBSTR name;
			RUN(CComQIPtr<IDGStringAttributes>(n)->get_Name(&name));
			if(needsQuotes(name.m_str))
				where += wsprintf(where,"\"%S\"; ",name.m_str);
			else
				where += wsprintf(where,"%S; ",name.m_str);
		}
		else {
			CComQIPtr<IDGEdge> e = uE;
			CComPtr<IDGNode> t,h;
			RUN(e->get_TailNode(&t));
			RUN(e->get_HeadNode(&h));
			CComBSTR name;
			RUN(CComQIPtr<IDGStringAttributes>(t)->get_Name(&name));
			if(needsQuotes(name.m_str))
				where += wsprintf(where,"\"%S\"->",name.m_str);
			else
				where += wsprintf(where,"%S->",name.m_str);
			name.Empty();
			RUN(CComQIPtr<IDGStringAttributes>(h)->get_Name(&name));
			if(needsQuotes(name.m_str))
				where += wsprintf(where,"\"%S\"; ",name.m_str);
			else
				where += wsprintf(where,"%S; ",name.m_str);
		}
	}
    if(inputusetext)
        RUN(inputtextctl->put_Text(CComBSTR(buf)));
    else {
        CComBSTR name;
        RUN(inputchoicectl->Get(&name));
        HGLOBAL glob = GlobalAlloc(GHND,where-buf + name.Length() + 20);
        char *s = (char*)GlobalLock(glob);
        wsprintf(s,"digraph %S {%s}",name.m_str,buf);
        GlobalUnlock(glob);
        RUN(inputLayout->LoadDotfile(glob,true,true));
        GlobalFree(glob);
    }
    if(alwaysSave)
        RUN(save(DINOGRAPH_INPUT_START));
    return S_OK;
}
HRESULT DinographUI::show() {
	return S_OK;
}

HRESULT DinographUI::setType(BSTR val) {
    if(!wcscmp(val,L"nodes&edges"))
        propNodes = propEdges = true;
    else if(!wcscmp(val,L"nodes"))
        propEdges = !(propNodes = true);
    else if(!wcscmp(val,L"edges"))
        propEdges = !(propNodes = false);
    else _ASSERTE(false);
    RUN(refreshProperties());
    return S_OK;
}
HRESULT DinographUI::setScope(BSTR val) {
    if(!wcscmp(val,L"abstract"))
        propAbstract = true;
    else if(!wcscmp(val,L"layout"))
        propAbstract = false;
    else _ASSERTE(false);
    RUN(refreshProperties());
    return S_OK;
}
